//
// SDL.cpp
// Copyright Michael J Stokes, 2008.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "Application.h"
#include "SDL.h"
#include "SkeletonTranslator.h"

#ifdef WITH_SDL


////////////////////////////////////////////////////////////////////////////////
// Set the SDL viewport

static void SDLSetViewport( int width, int height )
{
	glViewport( 0, 0, width, height );
	
	vec2 fov( 90, 0 );
	fov.y = atan( (scalar)height / ( (scalar)width / tan( fov.x * M_PI / 360.0f ) ) ) * 360.0f / M_PI;
	
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( fov.y, (float)width / (float)height, 0.1f, 256.0f );
	glScalef( 1, 1, -1 );
	glMatrixMode( GL_MODELVIEW );
	
	cout << "Renderer resized to " << width << "x" << height;
	cout << std::setprecision( 3 ) << " with FOV (" << fov.x << ", " << fov.y << ")" << endl;
}

////////////////////////////////////////////////////////////////////////////////
// Initialize SDL

static bool SDLInitialize( int& width, int& height, bool fullscreen )
{
	// Initialize SDL
	
	cout << "Initializing SDL with " << width << "x" << height << " graphics..." << endl;
	
	if ( SDL_Init( SDL_INIT_VIDEO | SDL_INIT_TIMER ) < 0 )
	{
		cerr << "Error initializing SDL - " << SDL_GetError() << endl;
		return false;
	}
	
	// Query optimal BPP
	
	const SDL_VideoInfo* vi = SDL_GetVideoInfo();
	
	if ( NULL == vi )
	{
		cerr << "Error fetching SDL video information - " << SDL_GetError() << endl;
		return false;
	}
	
	// Configure GL buffer metrics
	
	SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 8 );
	SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 8 );
	SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 8 );
	SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 24 );
	SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
	
	// Check windowed dimensions
	
	if ( ! fullscreen )
	{
		if ( ( width > vi->current_w ) || ( height > vi->current_h ) )
		{
			cout << "Requested window size of " << width << "x" << height << " is too large." << endl;
			
			width	= 640;
			height	= 480;
		}
	}
	
	// Set the video mode
	
	SDL_Surface* surface = SDL_SetVideoMode( width, height,
		vi->vfmt->BitsPerPixel,
		SDL_OPENGL | ( fullscreen ? SDL_FULLSCREEN : SDL_RESIZABLE ) );
	
	if ( NULL == surface )
	{
		cerr << "Error setting " << width << "x" << height << " video mode - " << SDL_GetError() << endl;
		return false;
	}
	
	// Report details of the video mode we got
	
	vi = SDL_GetVideoInfo();
	char driver[ 64 ];
	int depth = 0;
	
	if ( NULL == SDL_VideoDriverName( driver, sizeof( driver ) ) )
	{
		strcpy( driver, "unknown" );
	}
	
	SDL_GL_GetAttribute( SDL_GL_DEPTH_SIZE, &depth );
	
	cout << "Using '" << driver << "' video driver at ";
	cout << vi->current_w << "x" << vi->current_h;
	cout << " with " << (int)vi->vfmt->BitsPerPixel << " BPP and " << depth << " depth" << endl;
	
	// Set SDL window caption
	
	SDL_WM_SetCaption( "VGS", "default" );
	
	// Setup the OpenGL viewport
	
	SDLSetViewport( width, height );
	
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// Main

int main( int argc, char** argv )
{
	int width = 1024, height = 768;
#ifdef _DEBUG
	bool fullscreen = false;
#else
	bool fullscreen = true;
#endif
	
	for ( int i = 1 ; i < argc ; i ++ )
	{
		if ( 0 == strcmp( argv[ i ], "-res" ) )
		{
			if ( argc - i < 3 )
			{
				cerr << "USAGE: -res <width> <height>" << endl;
				return 1;
			}
			
			sscanf( argv[ i + 1 ], "%d", &width );
			sscanf( argv[ i + 2 ], "%d", &height );
			i += 2;
		}
		else if ( 0 == strcmp( argv[ i ], "-fullscreen" ) )
		{
			fullscreen = true;
		}
		else if ( 0 == strcmp( argv[ i ], "-windowed" ) )
		{
			fullscreen = false;
		}
		else if ( 0 == strcmp( argv[ i ], "-st" ) )
		{
			scalar scale = 1;
			sscanf( argv[ i + 2 ], "%f", &scale );
			SkeletonTranslator::Translate( argv[ i + 1 ], scale, "" );
			return 0;
		}
		else if ( 0 == strcmp( argv[ i ], "-skm" ) )
		{
			SkeletonTranslator::SKM( argv[ i + 1 ] );
			return 0;
		}
		else
		{
			cerr << "WARNING: Unknown argument: " << argv[ i ] << endl;
		}
	}
	
	if ( ! SDLInitialize( width, height, fullscreen ) ) return false;
	
	Application* app = new Application();
	if ( ( ! app->Load( "Root" ) ) || ( ! app->PrepareContext() ) || ( ! app->Start() ) ) return 1;
	
#ifdef _WIN32
	char cd[ MAX_PATH ];
	GetCurrentDirectory( sizeof( cd ), cd );
	strcat( cd, "\\SBM" );
	ShellExecute( NULL, NULL, "SmartBody.bat", NULL, cd, SW_SHOWMINIMIZED );
#endif
	
	bool active = ( 0 != ( SDL_GetAppState() & SDL_APPINPUTFOCUS ) );
	
	if ( active )
	{
		SDL_WM_GrabInput( SDL_GRAB_ON );
		SDL_ShowCursor( SDL_DISABLE );
	}
	
	UINT32 baseTime = active ? SDL_GetTicks() : 0;
	
	quaternion heading, pitch;
	
	for ( bool run = true, mouseWait = true ; run ; )
	{
		if ( true )
		{
			quaternion orientation( heading * pitch );
			UINT actions = 0;
			
			if ( active )
			{
				BYTE mouse = SDL_GetMouseState( NULL, NULL );
				const BYTE* keys = SDL_GetKeyState( NULL );
				
				if ( ( mouse & SDL_BUTTON( 1 ) ) || keys[ SDLK_w ] || keys[ SDLK_UP ] )
					actions |= Application::ACTION_FORWARD;
				if ( ( mouse & SDL_BUTTON( 3 ) ) || keys[ SDLK_s ] || keys[ SDLK_DOWN ] )
					actions |= Application::ACTION_BACKWARD;
			}
			
			UINT32 now = SDL_GetTicks() - baseTime;
			app->Update( (scalar)now / 1000.0f, orientation.Direction(), actions );
			
			matrix4x4 rotation( orientation.Inverted() );
			glLoadMatrixf( rotation );
			glTranslatef( 0, -1.5f, 0 );
			
			glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
			app->Render();
			SDL_GL_SwapBuffers();
		}
		
		SDL_Event event;
		
		while ( run && SDL_PollEvent( &event ) )
		{
			switch ( event.type )
			{
			case SDL_QUIT:
				run = false;
				break;
				
			case SDL_VIDEORESIZE:
				width	= event.resize.w;
				height	= event.resize.h;
				SDLSetViewport( width, height );
				break;
				
			case SDL_ACTIVEEVENT:
				if ( SDL_APPINPUTFOCUS & event.active.state )
				{
					bool focused = ( 0 != event.active.gain );
					
					if ( focused != active )
					{
						if ( active = focused )
						{
							// baseTime = SDL_GetTicks() - baseTime;
							SDL_WM_GrabInput( SDL_GRAB_ON );
							SDL_ShowCursor( SDL_DISABLE );
							mouseWait = true;
						}
						else
						{
							// baseTime = SDL_GetTicks() - baseTime;
							SDL_ShowCursor( SDL_ENABLE );
							SDL_WM_GrabInput( SDL_GRAB_OFF );
						}
					}
				}
				break;
				
			case SDL_MOUSEMOTION:
				if ( mouseWait )
				{
					mouseWait = false;
				}
				else if ( active )
				{
					vec2 move;
#ifdef _WIN32
					move.x = (scalar)event.motion.xrel /  width;
					move.y = (scalar)event.motion.yrel / -height;
#else
					move.x = (scalar)( event.motion.x - width / 2 ) * 0.6f /  width;
					move.y = (scalar)( event.motion.y - height / 2 ) * 0.6f / -height;
					SDL_WarpMouse( width / 2, height / 2 );
#endif
					heading = quaternion( move.x * ( M_PI / -4 ), vec3( 0, 1, 0 ) ) * heading;
					heading.Normalize();
					pitch = quaternion( move.y * ( M_PI / 4 ), vec3( 1, 0, 0 ) ) * pitch;
					pitch.Normalize();
				}
				break;
				
			case SDL_KEYDOWN:
				if ( SDLK_ESCAPE == event.key.keysym.sym )
				{
					run = false;
				}
				break;
			}
		}
		
		SDL_Delay( 1 );
	}
	
	delete app;
	SDL_Quit();
	
	return 0;
}

#endif
