//
// 3DMath.h
// Copyright Michael J Stokes, 2007.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#pragma once


// This math module provides the following types:
//
// scalar		- scalar value
// vec2			- vector in 2D
// vec3			- vector in 3D
// plane3		- plane in 3D
// quaternion	- quaternion
// matrix3x3	- 3x3 matrix
// matrix4x4	- 4x4 matrix


// Local values for PI and EPSILON, for consistency across platforms

#undef	M_PI
#undef	M_EPSILON
#define	M_PI			3.14159265358979323846f
#define	M_EPSILON		0.1f

// Utility macros

#define sign( x )		( (x) ? ( ( (x) > 0 ) ? 1 : -1 ) : 0 )
#ifndef min
#define min( x, y )		( ( (x) < (y) ) ? (x) : (y) )
#define max( x, y )		( ( (x) > (y) ) ? (x) : (y) )
#endif

// Relationship to a plane: entirely on, entirely in front, entirely behind, spanning (both)

enum SideOfPlane
{
	SIDE_ON, SIDE_FRONT, SIDE_BACK, SIDE_BOTH
};


// Scalar value
// Selecting 32-bit here

typedef float scalar;
#define SCALAR_FLOAT


// Vector in 2D

struct vec2
{
//	Data
	union
	{
		struct { scalar x, y; };
		struct { scalar s, t; };
		struct { scalar u, v; };
		scalar n[ 2 ];
	};
	
//	Default constructor
	inline vec2() : x( 0 ), y( 0 ) {};
	
//	Copy constructor
	inline vec2( const vec2& s ) : x( s.x ), y( s.y ) {};
	
//	Component constructor
	inline vec2( scalar nx, scalar ny ) : x( nx ), y( ny ) {};
	
//	Access as pointer to scalars
	inline operator scalar * ()
	{
		return &x;
	}
	
//	Access as const pointer to scalars
	inline operator const scalar * () const
	{
		return &x;
	}
	
//	Equality operator
	inline bool operator == ( const vec2& s ) const
	{
		return	( abs( s.x - x ) <= M_EPSILON ) &&
				( abs( s.y - y ) <= M_EPSILON );
	}
	
//	Inequality operator
	inline bool operator != ( const vec2& s ) const
	{
		return	( abs( s.x - x ) > M_EPSILON ) ||
				( abs( s.y - y ) > M_EPSILON );
	}
	
//	Zero operator
	inline bool operator ! () const
	{
		return ( x == 0 ) && ( y == 0 );
	}
	
//	Assignment operator
	inline const vec2& operator = ( const vec2& s )
	{
		x = s.x; y = s.y;
		return *this;
	}
	
//	Add assignment operator
	inline const vec2& operator += ( const vec2& s )
	{
		x += s.x; y += s.y;
		return *this;
	}
	
//	Subtract assignment operator
	inline const vec2& operator -= ( const vec2& s )
	{
		x -= s.x; y -= s.y;
		return *this;
	}
	
//	Multiply by scalar assignment operator
	inline const vec2& operator *= ( scalar s )
	{
		x *= s; y *= s;
		return *this;
	}
	
//	Divide by scalar assignment operator
	inline const vec2& operator /= ( scalar s )
	{
		x /= s; y /= s;
		return *this;
	}
	
//	Negation operator
	inline vec2 operator - () const
	{
		return vec2( -x, -y );
	}

//	Addition operator
	inline vec2 operator + ( const vec2& s ) const
	{
		return vec2( x + s.x, y + s.y );
	}
	
//	Subtraction operator
	inline vec2 operator - ( const vec2& s ) const
	{
		return vec2( x - s.x, y - s.y );
	}
	
//	Multiply by scalar operator
	inline vec2 operator * ( scalar s ) const
	{
		return vec2( x * s, y * s );
	}
	
//	Multiply by scalar operator
	inline friend vec2 operator * ( scalar n, const vec2& s )
	{
		return vec2( s.x * n, s.y * n );
	}
	
//	Divide by scalar operator
	inline vec2 operator / ( scalar s ) const
	{
		return vec2( x / s, y / s );
	}
	
//	Divide scalar by vector operator
	inline friend vec2 operator / ( scalar n, const vec2& s )
	{
		return vec2( n / s.x, n / s.y );
	}
	
//	Get length
	inline scalar Length() const
	{
		return sqrt( x * x + y * y );
	}
	
//	Get length squared (faster)
	inline scalar LengthSquared() const
	{
		return x * x + y * y;
	}
	
//	Dot product (scalar product) with another vector
	inline scalar DotProduct( const vec2& s ) const
	{
		return ( x * s.x ) + ( y * s.y );
	}
	
//	Zero
	inline void Zero()
	{
		x = y = 0;
	}
	
//	Set components
	inline void Set( scalar nx, scalar ny )
	{
		x = nx; y = ny;
	}
	
//	Make this vector the minimum of itself and the argument
	inline void Min( const vec2& s )
	{
		if ( s.x < x ) x = s.x;
		if ( s.y < y ) y = s.y;
	}
	
//	Make this vector the maximum of itself and the argument
	inline void Max( const vec2& s )
	{
		if ( s.x > x ) x = s.x;
		if ( s.y > y ) y = s.y;
	}
	
//	Normalize in place
	inline void Normalize()
	{
		scalar l = Length();
		
		if ( l != 0 )
		{
			x /= l;
			y /= l;
		}
		else
		{
			Zero();
		}
	}
};

extern const vec2 vec2origin;


// Vector in 3D

struct vec3
{
//	Data
	union
	{
		struct { scalar x, y, z; };
		struct { scalar r, g, b; };
		struct { scalar pitch, yaw, roll; };
		scalar n[ 3 ];
	};
	
//	Default constructor
	inline vec3() : x( 0 ), y( 0 ), z( 0 ) {};
	
//	Copy constructor
	inline vec3( const vec3& s ) : x( s.x ), y( s.y ), z( s.z ) {};
	
//	Component constructor
	inline vec3( scalar nx, scalar ny, scalar nz ) : x( nx ), y( ny ), z( nz ) {};
	
//	Access as pointer to scalars
	inline operator scalar * ()
	{
		return &x;
	}
	
//	Access as const pointer to scalars
	inline operator const scalar * () const
	{
		return &x;
	}
	
//	Equality operator
	inline bool operator == ( const vec3& s ) const
	{
		return	( abs( s.x - x ) <= M_EPSILON ) &&
				( abs( s.y - y ) <= M_EPSILON ) &&
				( abs( s.z - z ) <= M_EPSILON );
	}
	
//	Inequality operator
	inline bool operator != ( const vec3& s ) const
	{
		return	( abs( s.x - x ) > M_EPSILON ) ||
				( abs( s.y - y ) > M_EPSILON ) ||
				( abs( s.z - z ) > M_EPSILON );
	}
	
//	Zero operator
	inline bool operator ! () const
	{
		return ( x == 0 ) && ( y == 0 ) && ( z == 0 );
	}
	
//	Assignment operator
	inline const vec3& operator = ( const vec3& s )
	{
		x = s.x; y = s.y; z = s.z;
		return *this;
	}
	
//	Add assignment operator
	inline const vec3& operator += ( const vec3& s )
	{
		x += s.x; y += s.y; z += s.z;
		return *this;
	}
	
//	Subtract assignment operator
	inline const vec3& operator -= ( const vec3& s )
	{
		x -= s.x; y -= s.y; z -= s.z;
		return *this;
	}
	
//	Multiply by scalar assignment operator
	inline const vec3& operator *= ( scalar s )
	{
		x *= s; y *= s; z *= s;
		return *this;
	}
	
//	Divide by scalar assignment operator
	inline const vec3& operator /= ( scalar s )
	{
		x /= s; y /= s; z /= s;
		return *this;
	}
	
//	Negate operator
	inline vec3 operator - () const
	{
		return vec3( -x, -y, -z );
	}
	
//	Add operator
	inline vec3 operator + ( const vec3& s ) const
	{
		return vec3( x + s.x, y + s.y, z + s.z );
	}
	
//	Subtract operator
	inline vec3 operator - ( const vec3& s ) const
	{
		return vec3( x - s.x, y - s.y, z - s.z );
	}
	
//	Multiply by scalar operator
	inline vec3 operator * ( scalar s ) const
	{
		return vec3( x * s, y * s, z * s );
	}
	
//	Multiply by scalar operator
	inline friend vec3 operator * ( scalar n, const vec3& s )
	{
		return vec3( s.x * n, s.y * n, s.z * n );
	}
	
//	Divide by scalar operator
	inline vec3 operator / ( scalar s ) const
	{
		return vec3( x / s, y / s, z / s );
	}
	
//	Divide scalar by vector operator
	inline friend vec3 operator / ( scalar n, const vec3& s )
	{
		return vec3( n / s.x, n / s.y, n / s.z );
	}
	
//	Get length
	inline scalar Length() const
	{
		return sqrt( x * x + y * y + z * z );
	}
	
//	Get length squared (faster)
	inline scalar LengthSquared() const
	{
		return x * x + y * y + z * z;
	}
	
//	Dot product (scalar product) with another vector
	inline scalar DotProduct( const vec3& s ) const
	{
		return ( x * s.x ) + ( y * s.y ) + ( z * s.z );
	}
	
//	Cross product (vector product) with another vector
//	NOTE: LEFT HANDED COORDINATE SYSTEM!
	inline vec3 CrossProduct( const vec3& s )
	{
		return vec3(
			s.z * y - s.y * z,
			s.x * z - s.z * x,
			s.y * x - s.x * y );
	}
	
//	Zero
	inline void Zero()
	{
		x = y = z = 0;
	}
	
//	Set components
	inline void Set( scalar nx, scalar ny, scalar nz )
	{
		x = nx; y = ny; z = nz;
	}
	
//	Make this vector the minimum of itself and the argument
	inline void Min( const vec3& s )
	{
		if ( s.x < x ) x = s.x;
		if ( s.y < y ) y = s.y;
		if ( s.z < z ) z = s.z;
	}
	
//	Make this vector the maximum of itself and the argument
	inline void Max( const vec3& s )
	{
		if ( s.x > x ) x = s.x;
		if ( s.y > y ) y = s.y;
		if ( s.z > z ) z = s.z;
	}
	
//	Normalize in place
	inline void Normalize()
	{
		scalar l = Length();
		
		if ( l != 0 )
		{
			x /= l;
			y /= l;
			z /= l;
		}
		else
		{
			Zero();
		}
	}
};

extern const vec3 vec3origin;


// Plane in 3D
// Using normal and distance from origin representation

struct plane3
{
//	Data
	vec3		normal;
	scalar		distance;
	
//	Default constructor
	inline plane3() : distance( 0 ) {};
	
//	Copy constructor
	inline plane3( const plane3& s ) : normal( s.normal ), distance( s.distance ) {};
	
//	Component constructor (normal and distance)
	inline plane3( const vec3& nnormal, scalar ndistance ) : normal( nnormal ), distance( ndistance ) {};
	
//	Construct from three points on the plane
	inline plane3( const vec3& p0, const vec3& p1, const vec3& p2 )
	{
		Set( p0, p1, p2 );
	}
	
//	Equality operator
	inline bool operator == ( const plane3& s ) const
	{
		return ( abs( distance - s.distance ) <= M_EPSILON ) && ( normal == s.normal );
	}
	
//	Inequality operator
	inline bool operator != ( const plane3& s ) const
	{
		return ( abs( distance - s.distance ) > M_EPSILON ) || ( normal != s.normal );
	}
	
//	Zero operator
	inline bool operator ! () const
	{
		return ! normal;
	}
	
//	Assignment operator
	inline const plane3& operator = ( const plane3& s )
	{
		normal		= s.normal;
		distance	= s.distance;
		return *this;
	}
	
//	Negation operator
	inline plane3 operator - () const
	{
		return plane3( vec3origin - normal, -distance );
	}
	
//	Zero
	inline void Zero()
	{
		normal.Zero();
		distance = 0;
	}
	
//	Set components (normal and distance)
	inline void Set( const vec3& nnormal, scalar ndistance )
	{
		normal		= nnormal;
		distance	= ndistance;
	}
	
//	Set by three points on the plane
	inline void Set( const vec3& p0, const vec3& p1, const vec3& p2 )
	{
		normal		= ( p1 - p0 ).CrossProduct( p2 - p0 );
		normal.Normalize();
		distance	= normal.DotProduct( p0 );
	}
	
//	Invert in place
	inline void Invert()
	{
		normal.x = -normal.x;
		normal.y = -normal.y;
		normal.z = -normal.z;
		distance = -distance;
	}
	
//	Return inverted
	inline plane3 Inverted() const
	{
		return plane3( vec3origin - normal, -distance );
	}
	
//	Get the distance from a 3D point to the plane
//	Positive distances are in front of the plane
	inline scalar Distance( const vec3& p ) const
	{
		return normal.DotProduct( p ) - distance;
	}
	
//	Get the SideOfPlane enumeration value for a 3D point: on, in front, behind
	inline SideOfPlane Side( const vec3& p ) const
	{
		scalar dot = normal.DotProduct( p ) - distance;
		if ( dot == 0 ) return SIDE_ON;
		return ( dot > 0 ) ? SIDE_FRONT : SIDE_BACK;
	}
};


// Quaternion

struct quaternion
{
//	Data
	scalar	w;
	vec3	v;
	
//	Default constructor
	inline quaternion() : w( 1 ) {};
	
//	Copy constructor
	inline quaternion( const quaternion& s ) : w( s.w ), v( s.v ) {};
	
//	Component constructor (w, x, y, z)
	inline quaternion( scalar nw, scalar nx, scalar ny, scalar nz ) : w( nw ), v( nx, ny, nz ) {};
	
//	Construct a rotation around a normal vector: a is the angle in radians, nv is the normal vector
	inline quaternion( scalar a, const vec3& nv )
	{
		scalar ha = -a / 2;
	    w = cos( ha );
		v = nv;
		v *= sin( ha );
	}
	
//	Equality operator
	inline bool operator == ( const quaternion& s ) const
	{
		return ( 0 == memcmp( this, &s, sizeof( quaternion ) ) );
	}
	
//	Inequality operator
	inline bool operator != ( const quaternion& s ) const
	{
		return ( 0 != memcmp( this, &s, sizeof( quaternion ) ) );
	}
	
//	Assignment operator
	inline const quaternion& operator = ( const quaternion& s )
	{
		w = s.w;
		v = s.v;
		return *this;
	}
	
//	Multiply assignment operator
	inline const quaternion& operator *= ( const quaternion& s )
	{
		*this = *this * s;
		return *this;
	}
	
//	Multiplication operator (note that order of operands is important)
	inline quaternion operator * ( const quaternion& s ) const
	{
		return quaternion(
			w * s.w 	- v.x * s.v.x 	- v.y * s.v.y 	- v.z * s.v.z,
			w * s.v.x 	+ v.x * s.w 	+ v.y * s.v.z	- v.z * s.v.y,
			w * s.v.y 	- v.x * s.v.z	+ v.y * s.w		+ v.z * s.v.x,
			w * s.v.z	+ v.x * s.v.y	- v.y * s.v.x	+ v.z * s.w );
	}
	
//	Zero
	inline void Zero()
	{
		w = 1;
		v.Zero();
	}
	
//	Get the norm (the 4D length)
	inline scalar Norm() const
	{
		return sqrt( w * w + v.LengthSquared() );
	}
	
//	Get the norm squared (the 4D length squared)
	inline scalar NormSquared() const
	{
		return w * w + v.LengthSquared();
	}
	
//	Normalize in place
	inline void Normalize()
	{
		scalar n = Norm();
		
		if ( n != 0 )
		{
			w /= n;
			v /= n;
		}
		else
		{
			Zero();
		}
	}
	
//	Invert in place
	inline void Invert()
	{
		scalar ls = NormSquared();
		
		if ( ls != 0 )
		{
			w /= ls;
			v /= -ls;
		}
	}
	
//	Return inverted
	inline quaternion Inverted() const
	{
		scalar ls = NormSquared();
		
		if ( ls != 0 )
		{
			return quaternion( w / ls, -v.x / ls, -v.y / ls, -v.z / ls );
		}
		else
		{
			return quaternion();
		}
	}
	
//	Get a normal vector in the direction the quaternion is "pointing"
	inline vec3 Direction() const;
	
//	Get the X and/or Y and/or Z axis vectors corresponding to the quaternion
	inline void Axes( vec3* x = NULL, vec3* y = NULL, vec3* z = NULL ) const;
	
//	Return a SLERP quaternion interpolation between this quaternion and o, where s is 0..1
	inline quaternion Slerp( const quaternion o, scalar s )
	{
		scalar cht = w * o.w + v.x * o.v.x + v.y * o.v.y + v.z * o.v.z;
		if ( abs( cht ) >= 1 ) return *this;
		
		quaternion io;
		const quaternion* ro = &o;
		
		if ( cht < 0 )
		{
			ro = &io;
			io.w = -o.w;
			io.v = -o.v;
			cht = -cht;
		}
		
		scalar ht	= acos( cht );
		scalar sht	= sqrt( 1.0f - cht * cht );
		
		if ( abs( sht ) < 0.001f ) return *this;
		
		scalar r0	= sin( ( 1.0f - s ) * ht ) / sht;
		scalar r1	= sin( s * ht ) / sht; 
		
		quaternion r( w * r0 + ro->w * r1,
			v.x * r0 + ro->v.x * r1,
			v.y * r0 + ro->v.y * r1,
			v.z * r0 + ro->v.z * r1 );
		r.Normalize();
		
		return r;
	}
};


// 3x3 matrix
// Column-major order

struct matrix3x3
{
//	Data
	union
	{
		scalar	n9[ 9 ];
		scalar	n33[ 3 ][ 3 ];
		
		struct
		{
			scalar row[ 3 ];
		} column[ 3 ];
	};
	
//	Default constructor
	inline matrix3x3()
	{
		memset( n9, 0, sizeof( scalar ) * 9 );
		n9[ 0 ] = n9[ 4 ] = n9[ 8 ] = 1;
	}
	
//	Copy constructor
	inline matrix3x3( const matrix3x3& s )
	{
		memcpy( n9, s.n9, sizeof( scalar ) * 9 );
	}
	
//	Construct from a pointer to 9 scalars
	inline matrix3x3( const scalar * scalars )
	{
		memcpy( n9, scalars, sizeof( scalar ) * 9 );
	}
	
//	Construct a rotation matrix from a quaternion
	inline matrix3x3( const quaternion& q )
	{
		scalar xx = q.v.x * q.v.x;
		scalar xy = q.v.x * q.v.y;
		scalar xz = q.v.x * q.v.z;
		scalar xw = q.v.x * q.w;
		scalar yy = q.v.y * q.v.y;
		scalar yz = q.v.y * q.v.z;
		scalar yw = q.v.y * q.w;
		scalar zz = q.v.z * q.v.z;
		scalar zw = q.v.z * q.w;
		
		n33[ 0 ][ 0 ] = 1.0f - 2.0f * ( yy + zz );
		n33[ 1 ][ 0 ] = 2.0f * ( xy - zw );
		n33[ 2 ][ 0 ] = 2.0f * ( xz + yw );
		
		n33[ 0 ][ 1 ] = 2.0f * ( xy + zw );
		n33[ 1 ][ 1 ] = 1.0f - 2.0f * ( xx + zz );
		n33[ 2 ][ 1 ] = 2.0f * ( yz - xw );
		
		n33[ 0 ][ 2 ] = 2.0f * ( xz - yw );
		n33[ 1 ][ 2 ] = 2.0f * ( yz + xw );
		n33[ 2 ][ 2 ] = 1.0f - 2.0f * ( xx + yy );
	}
	
//	Access as pointer to scalars
	inline operator scalar * ()
	{
		return n9;
	}
	
//	Access as const pointer to scalars
	inline operator const scalar * () const
	{
		return n9;
	}
	
//	Equality operator
	inline bool operator == ( const matrix3x3& s ) const
	{
		const scalar * pa = n9;
		const scalar * pb = s.n9;
		
		for ( int i = 9 ; i -- ; pa ++, pb ++ )
		{
			if ( abs( *pa - *pb ) > M_EPSILON ) return false;
		}
		
		return true;
	}
	
//	Inequality operator
	inline bool operator != ( const matrix3x3& s ) const
	{
		return ! ( *this == s );
	}
	
//	Assignment operator
	inline const matrix3x3& operator = ( const matrix3x3& s )
	{
		memcpy( n9, s.n9, sizeof( scalar ) * 9 );
		return *this;
	}
	
//	Add assignment operator
	inline const matrix3x3& operator += ( const matrix3x3& s )
	{
		scalar * pa = n9;
		const scalar * pb = s.n9;
		
		for ( int i = 9 ; i -- ; )
		{
			( *pa ++ ) += ( *pb ++ );
		}
		
		return *this;
	}
	
//	Subtract assignment operator
	inline const matrix3x3& operator -= ( const matrix3x3& s )
	{
		scalar * pa = n9;
		const scalar * pb = s.n9;
		
		for ( int i = 9 ; i -- ; )
		{
			( *pa ++ ) -= ( *pb ++ );
		}
		
		return *this;
	}
	
//	Multiply by scalar assignment operator
	inline const matrix3x3& operator *= ( scalar s )
	{
		scalar * pa = n9;
		
		for ( int i = 9 ; i -- ; )
		{
			( *pa ++ ) *= s;
		}
		
		return *this;
	}
	
//	Divide by scalar assignment operator
	inline const matrix3x3& operator /= ( scalar s )
	{
		scalar * pa = n9;
		
		for ( int i = 9 ; i -- ; )
		{
			( *pa ++ ) /= s;
		}
		
		return *this;
	}
	
//	Multiply by matrix assignment operator
	inline const matrix3x3& operator *= ( const matrix3x3& s )
	{
		*this = *this * s;
		return *this;
	}
	
//	Add operator
	inline matrix3x3 operator + ( const matrix3x3& s ) const
	{
		matrix3x3 r( *this );
		r += s;
		return r;
	}
	
//	Subtract operator
	inline matrix3x3 operator - ( const matrix3x3& s ) const
	{
		matrix3x3 r( *this );
		r -= s;
		return r;
	}
	
//	Multiply by scalar operator
	inline matrix3x3 operator * ( scalar s ) const
	{
		matrix3x3 r( *this );
		r *= s;
		return r;
	}
	
//	Multiply by scalar operator
	inline friend matrix3x3 operator * ( scalar n, const matrix3x3& s )
	{
		matrix3x3 r( s );
		r *= n;
		return r;
	}
	
//	Divide by scalar operator
	inline matrix3x3 operator / ( scalar s ) const
	{
		matrix3x3 r( *this );
		r /= s;
		return r;
	}
	
//	Divide scalar by matrix operator (dubious but why not)
	inline friend matrix3x3 operator / ( scalar n, const matrix3x3& s )
	{
		matrix3x3 r;
		
		const scalar * pa = s.n9;
		scalar * pb = r.n9;
		
		for ( int i = 9 ; i -- ; )
		{
			( *pb ++ ) = n / ( *pa ++ );
		}
		
		return r;
	}
	
//	Multiply matrix by matrix operator
	inline matrix3x3 operator * ( const matrix3x3& s ) const
	{
		matrix3x3 r;
		
		for ( int i = 2 ; i >= 0 ; i -- )
		{
			for ( int j = 2 ; j >= 0 ; j -- )
			{
				r.n33[ j ][ i ]	= n33[ 0 ][ i ] * s.n33[ j ][ 0 ]
								+ n33[ 1 ][ i ] * s.n33[ j ][ 1 ]
								+ n33[ 2 ][ i ] * s.n33[ j ][ 2 ];
			}
		}
		
		return r;
	}
	
//	Multiply matrix by vector and return vector operator
	inline vec3 operator * ( const vec3& v ) const
	{
		return vec3(
			v.x * n9[ 0 ] + v.y * n9[ 3 ] + v.z * n9[ 6 ],
			v.x * n9[ 1 ] + v.y * n9[ 4 ] + v.z * n9[ 7 ],
			v.x * n9[ 2 ] + v.y * n9[ 5 ] + v.z * n9[ 8 ] );
	}
	
//	Multiply rotation matrix by plane and return rotated plane operator
	inline plane3 operator * ( const plane3& p ) const
	{
		return plane3( *this * p.normal, p.distance );
	}
	
//	Set to identity matrix
	inline void Identity()
	{
		memset( n9, 0, sizeof( scalar ) * 9 );
		n9[ 0 ] = n9[ 4 ] = n9[ 8 ] = 1;
	}
	
//	Set by a pointer to 9 scalars
	inline void Set( scalar * scalars )
	{
		memcpy( n9, scalars, sizeof( scalar ) * 9 );
	}
	
//	Return the given row as a vec3
	inline vec3 Row( int i ) const
	{
		return vec3( n33[ 0 ][ i ], n33[ 1 ][ i ], n33[ 2 ][ i ] );
	}
	
//	Return the given column as a vec3
	inline vec3 Column( int i ) const
	{
		return vec3( n33[ i ][ 0 ], n33[ i ][ 1 ], n33[ i ][ 2 ] );
	}
	
//	Transpose in place
	inline void Transpose()
	{
		*this = Transposed();
	}
	
//	Return transposed
	inline matrix3x3 Transposed() const
	{
		matrix3x3 r;
		
		for ( int i = 0 ; i < 3 ; i ++ )
		{
			for ( int j = 0 ; j < 3 ; j ++ )
			{
				r.n33[ i ][ j ] = n33[ j ][ i ];
			}
		}
		
		return r;
	}
	
//	Return a vec3 in the "direction" the rotation matrix is pointing in
//	(i.e. its Z axis)
	inline vec3 Direction() const
	{
		return vec3( n33[ 2 ][ 0 ], n33[ 2 ][ 1 ], n33[ 2 ][ 2 ] );
	}
	
//	Get the X and/or Y and/or Z vectors corresponding to the rotation matrix
	inline void Axes( vec3* x = NULL, vec3* y = NULL, vec3* z = NULL ) const
	{
		if ( NULL != x )
		{
			x->Set( n33[ 0 ][ 0 ], n33[ 0 ][ 1 ], n33[ 0 ][ 2 ] );
		}
		
		if ( NULL != y )
		{
			y->Set( n33[ 1 ][ 0 ], n33[ 1 ][ 1 ], n33[ 1 ][ 2 ] );
		}
		
		if ( NULL != z )
		{
			z->Set( n33[ 2 ][ 0 ], n33[ 2 ][ 1 ], n33[ 2 ][ 2 ] );
		}
	}
};


// 4x4 matrix

struct matrix4x4
{
//	Data
	union
	{
		scalar	n16[ 16 ];
		scalar	n44[ 4 ][ 4 ];
		
		struct
		{
			scalar row[ 4 ];
		} column[ 4 ];
	};
	
//	Default constructor
	inline matrix4x4()
	{
		memset( n16, 0, sizeof( scalar ) * 16 );
		n16[ 0 ] = n16[ 5 ] = n16[ 10 ] = n16[ 15 ] = 1;
	}
	
//	Copy constructor
	inline matrix4x4( const matrix4x4& s )
	{
		memcpy( n16, s.n16, sizeof( scalar ) * 16 );
	}
	
//	Construct a rotation matrix with no translation from the given 3x3 rotation matrix
	inline matrix4x4( const matrix3x3& s )
	{
		n16[ 0 ]	= s.n9[ 0 ];
		n16[ 1 ]	= s.n9[ 1 ];
		n16[ 2 ]	= s.n9[ 2 ];
		n16[ 3 ]	= 0;
		n16[ 4 ]	= s.n9[ 3 ];
		n16[ 5 ]	= s.n9[ 4 ];
		n16[ 6 ]	= s.n9[ 5 ];
		n16[ 7 ]	= 0;
		n16[ 8 ]	= s.n9[ 6 ];
		n16[ 9 ]	= s.n9[ 7 ];
		n16[ 10 ]	= s.n9[ 8 ];
		n16[ 11 ]	= 0;
		n16[ 12 ]	= 0;
		n16[ 13 ]	= 0;
		n16[ 14 ]	= 0;
		n16[ 15 ]	= 1;
	}
	
//	Construct a translation matrix with no rotation from the given vec3 translation
	inline matrix4x4( const vec3& s )
	{
		memset( n16, 0, sizeof( scalar ) * 16 );
		n16[ 0 ] = n16[ 5 ] = n16[ 10 ] = n16[ 15 ] = 1;
		n44[ 3 ][ 0 ] = s.x;
		n44[ 3 ][ 1 ] = s.y;
		n44[ 3 ][ 2 ] = s.z;
	}
	
//	Construct a rotation matrix with no translation from the given quaternion
	inline matrix4x4( const quaternion& q )
	{
		scalar xx = q.v.x * q.v.x;
		scalar xy = q.v.x * q.v.y;
		scalar xz = q.v.x * q.v.z;
		scalar xw = q.v.x * q.w;
		scalar yy = q.v.y * q.v.y;
		scalar yz = q.v.y * q.v.z;
		scalar yw = q.v.y * q.w;
		scalar zz = q.v.z * q.v.z;
		scalar zw = q.v.z * q.w;
		
		n44[ 0 ][ 0 ] = 1.0f - 2.0f * ( yy + zz );
		n44[ 1 ][ 0 ] = 2.0f * ( xy - zw );
		n44[ 2 ][ 0 ] = 2.0f * ( xz + yw );
		n44[ 3 ][ 0 ] = 0;
		
		n44[ 0 ][ 1 ] = 2.0f * ( xy + zw );
		n44[ 1 ][ 1 ] = 1.0f - 2.0f * ( xx + zz );
		n44[ 2 ][ 1 ] = 2.0f * ( yz - xw );
		n44[ 3 ][ 1 ] = 0;
		
		n44[ 0 ][ 2 ] = 2.0f * ( xz - yw );
		n44[ 1 ][ 2 ] = 2.0f * ( yz + xw );
		n44[ 2 ][ 2 ] = 1.0f - 2.0f * ( xx + yy );
		n44[ 3 ][ 2 ] = 0;
		
		n44[ 0 ][ 3 ] = 0;
		n44[ 1 ][ 3 ] = 0;
		n44[ 2 ][ 3 ] = 0;
		n44[ 3 ][ 3 ] = 1;
	}
	
//	Construct from a pointer to 16 scalars
	inline matrix4x4( const scalar * scalars )
	{
		memcpy( n16, scalars, sizeof( scalar ) * 16 );
	}
	
//	Access as pointer to scalars
	inline operator scalar * ()
	{
		return n16;
	}
	
//	Access as const pointer to scalars
	inline operator const scalar * () const
	{
		return n16;
	}
	
//	Equality operator
	inline bool operator == ( const matrix4x4& s ) const
	{
		const scalar * pa = n16;
		const scalar * pb = s.n16;
		
		for ( int i = 16 ; i -- ; pa ++, pb ++ )
		{
			if ( abs( *pa - *pb ) > M_EPSILON ) return false;
		}
		
		return true;
	}
	
//	Inequality operator
	inline bool operator != ( const matrix4x4& s ) const
	{
		return ! ( *this == s );
	}
	
//	Assignment operator
	inline const matrix4x4& operator = ( const matrix4x4& s )
	{
		memcpy( n16, s.n16, sizeof( scalar ) * 16 );
		return *this;
	}
	
//	Add assignment operator
	inline const matrix4x4& operator += ( const matrix4x4& s )
	{
		scalar * pa = n16;
		const scalar * pb = s.n16;
		
		for ( int i = 16 ; i -- ; )
		{
			( *pa ++ ) += ( *pb ++ );
		}
		
		return *this;
	}
	
//	Subtract assignment operator
	inline const matrix4x4& operator -= ( const matrix4x4& s )
	{
		scalar * pa = n16;
		const scalar * pb = s.n16;
		
		for ( int i = 16 ; i -- ; )
		{
			( *pa ++ ) -= ( *pb ++ );
		}
		
		return *this;
	}
	
//	Multiply by scalar assignment operator
	inline const matrix4x4& operator *= ( scalar s )
	{
		scalar * pa = n16;
		
		for ( int i = 16 ; i -- ; )
		{
			( *pa ++ ) *= s;
		}
		
		return *this;
	}
	
//	Divide by scalar assignment operator
	inline const matrix4x4& operator /= ( scalar s )
	{
		scalar * pa = n16;
		
		for ( int i = 16 ; i -- ; )
		{
			( *pa ++ ) /= s;
		}
		
		return *this;
	}
	
//	Multiply by matrix assignment operator
	inline const matrix4x4& operator *= ( const matrix4x4& s )
	{
		*this = *this * s;
		return *this;
	}
	
//	Add operator
	inline matrix4x4 operator + ( const matrix4x4& s ) const
	{
		matrix4x4 r( *this );
		r += s;
		return r;
	}
	
//	Subtract operator
	inline matrix4x4 operator - ( const matrix4x4& s ) const
	{
		matrix4x4 r( *this );
		r -= s;
		return r;
	}
	
//	Multiply by scalar operator
	inline matrix4x4 operator * ( scalar s ) const
	{
		matrix4x4 r( *this );
		r *= s;
		return r;
	}
	
//	Multiply by scalar operator
	inline friend matrix4x4 operator * ( scalar n, const matrix4x4& s )
	{
		matrix4x4 r( s );
		r *= n;
		return r;
	}
	
//	Divide by scalar operator
	inline matrix4x4 operator / ( scalar s ) const
	{
		matrix4x4 r( *this );
		r /= s;
		return r;
	}
	
//	Divide scalar by matrix operator
	inline friend matrix4x4 operator / ( scalar n, const matrix4x4& s )
	{
		matrix4x4 r;
		
		const scalar * pa = s.n16;
		scalar * pb = r.n16;
		
		for ( int i = 16 ; i -- ; )
		{
			( *pb ++ ) = n / ( *pa ++ );
		}
		
		return r;
	}
	
//	Multiply matrix by matrix operator
	inline matrix4x4 operator * ( const matrix4x4& s ) const
	{
		matrix4x4 r;
		
		for ( int i = 3 ; i >= 0 ; i -- )
		{
			for ( int j = 3 ; j >= 0 ; j -- )
			{
				r.n44[ j ][ i ]	= n44[ 0 ][ i ] * s.n44[ j ][ 0 ]
								+ n44[ 1 ][ i ] * s.n44[ j ][ 1 ]
								+ n44[ 2 ][ i ] * s.n44[ j ][ 2 ]
								+ n44[ 3 ][ i ] * s.n44[ j ][ 3 ];
			}
		}
		
		return r;
	}
	
//	Multiply matrix by vector returning vector operator
	inline vec3 operator * ( const vec3& v ) const
	{
		return vec3(
			v.x * n16[ 0 ] + v.y * n16[ 4 ] + v.z * n16[ 8 ]  + n16[ 12 ],
			v.x * n16[ 1 ] + v.y * n16[ 5 ] + v.z * n16[ 9 ]  + n16[ 13 ],
			v.x * n16[ 2 ] + v.y * n16[ 6 ] + v.z * n16[ 10 ] + n16[ 14 ] );
	}
	
//	Multiply rotation part of matrix by vector returning vector
	inline vec3 Rotate( const vec3& v ) const
	{
		return vec3(
			v.x * n16[ 0 ] + v.y * n16[ 4 ] + v.z * n16[ 8 ],
			v.x * n16[ 1 ] + v.y * n16[ 5 ] + v.z * n16[ 9 ],
			v.x * n16[ 2 ] + v.y * n16[ 6 ] + v.z * n16[ 10 ] );
	}
	
//	Multiply matrix by plane and return transformed plane operator
	inline plane3 operator * ( const plane3& p ) const
	{
		vec3 n(
			p.normal.x * n16[ 0 ] + p.normal.y * n16[ 4 ] + p.normal.z * n16[ 8 ],
			p.normal.x * n16[ 1 ] + p.normal.y * n16[ 5 ] + p.normal.z * n16[ 9 ],
			p.normal.x * n16[ 2 ] + p.normal.y * n16[ 6 ] + p.normal.z * n16[ 10 ] );
		
		scalar l = n.Length();
		
		if ( l > 0 )
		{
			n /= l;
			return plane3( n, ( n.x * n16[ 12 ] + n.y * n16[ 13 ] + n.z * n16[ 14 ] ) + p.distance * l );
		}
		else
		{
			return plane3();
		}
	}
	
//	Set to identity matrix
	inline void Identity()
	{
		memset( n16, 0, sizeof( scalar ) * 16 );
		n16[ 0 ] = n16[ 5 ] = n16[ 10 ] = n16[ 15 ] = 1;
	}
	
//	Set by a pointer to 16 scalars
	inline void Set( scalar * scalars )
	{
		memcpy( n16, scalars, sizeof( scalar ) * 16 );
	}
	
//	Transpose in place
	inline void Transpose()
	{
		*this = Transposed();
	}
	
//	Return transposed
	inline matrix4x4 Transposed() const
	{
		matrix4x4 r;
		
		for ( int i = 0 ; i < 4 ; i ++ )
		{
			for ( int j = 0 ; j < 4 ; j ++ )
			{
				r.n44[ i ][ j ] = n44[ j ][ i ];
			}
		}
		
		return r;
	}
	
//	Return a vec3 in the "direction" the rotation matrix is pointing in
//	(i.e. its Z axis)
	inline vec3 Direction() const
	{
		return vec3( n44[ 2 ][ 0 ], n44[ 2 ][ 1 ], n44[ 2 ][ 2 ] );
	}
	
//	Get the X and/or Y and/or Z vectors corresponding to the rotation matrix
	inline void Axes( vec3* x = NULL, vec3* y = NULL, vec3* z = NULL ) const
	{
		if ( NULL != x )
		{
			x->Set( n44[ 0 ][ 0 ], n44[ 0 ][ 1 ], n44[ 0 ][ 2 ] );
		}
		
		if ( NULL != y )
		{
			y->Set( n44[ 1 ][ 0 ], n44[ 1 ][ 1 ], n44[ 1 ][ 2 ] );
		}
		
		if ( NULL != z )
		{
			z->Set( n44[ 2 ][ 0 ], n44[ 2 ][ 1 ], n44[ 2 ][ 2 ] );
		}
	}
};

// Get the X and/or Y and/or Z axis vectors corresponding to the quaternion
inline void quaternion::Axes( vec3* x, vec3* y, vec3* z ) const
{
	matrix3x3( *this ).Axes( x, y, z );
}

// Get a normal vector in the direction the quaternion is "pointing"
inline vec3 quaternion::Direction() const
{
	vec3 z;
	Axes( NULL, NULL, &z );
	return z;
}
