//
// Actor.cpp
// Copyright Michael J Stokes, 2009.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "Application.h"
#include "Actor.h"
#include "Character.h"
#include "Audio.h"
#include "Sound.h"
#include "Physics.h"


////////////////////////////////////////////////////////////////////////////////
// Actor - construction

Actor::Actor( const string& name, XMLElement* request )
{
	_name		= name;
	_id			= -1;
	
	_character	= NULL;
	_emitter	= NULL;
	
	_position.x	= 2;
	_position.y	= 8;
	_heading	= M_PI * 0.5f;
	_friction	= 4;
	_gravity	= -20;
	_height		= 1.2f;
	_radius		= 0.2f;
	_controller	= NULL;
	
	if ( XMLElement* position = request->GetElement( "position" ) )
	{
		if ( PCSTR target = position->GetAttributeString( "target" ) )
		{
			_position = Application::GetInstance()->GetActors()->GetPosition( target );
		}
		else
		{
			_position.x = position->GetAttributeFloat( "x" );
			_position.y = position->GetAttributeFloat( "y" );
			_position.z = position->GetAttributeFloat( "z" );
		}
	}
	
	if ( XMLElement* orientation = request->GetElement( "orientation" ) )
	{
		_heading = orientation->GetAttributeFloat( "heading" ) * M_PI / 180.0f;
	}
	
	if ( XMLElement* size = request->GetElement( "size" ) )
	{
		_height = size->GetAttributeFloat( "height", _height );
		_radius = size->GetAttributeFloat( "radius", _radius );
	}
	
	string model;
	
	if ( XMLElement* character = request->GetElement( "character" ) )
	{
		if ( PCSTR name = character->GetAttributeString( "model" ) )
		{
			model = name;
		}
	}
	
	if ( model.empty() ) model = "brad";
	
	if ( Character* cm = Application::GetInstance()->GetRenderer()->LoadCharacter( model.c_str() ) )
	{
		SetCharacter( cm );
		cm->Release();
	}
	
	NxCapsuleControllerDesc desc;
	
	desc.radius			= _radius;
	desc.height			= _height;
	desc.stepOffset		= 0.5f;
	desc.position.x		= (scalar)_position.x;
	desc.position.y		= (scalar)_position.y + _radius * 5;
	desc.position.z		= (scalar)_position.z;
	
	Physics* physics	= Application::GetInstance()->GetPhysics();
	_controller			= physics->GetControllerManager()->createController( physics->GetScene(), desc );
	
	BoneBusConnected();
}

////////////////////////////////////////////////////////////////////////////////
// Actor - destruction

Actor::~Actor()
{
	if ( NULL != _controller )
	{
		Application::GetInstance()->GetPhysics()->GetControllerManager()->releaseController( *_controller );
	}

	if ( NULL != _emitter )
	{
		Audio* audio = Application::GetInstance()->GetAudio();
		if ( NULL != audio ) audio->Delete( _emitter );
	}
	
	if ( NULL != _character )
	{
		SetCharacter( NULL );
	}
	
	if ( _id >= 0 )
	{
		GetActors()->SendSBM( "char " + _name + " remove" );
		_id = -1;
	}
}

////////////////////////////////////////////////////////////////////////////////
// Actor - send an XML request

XMLElement* Actor::Send( XMLElement* request )
{
	if ( request->IsNamed( "vgs", "position" ) )
	{
		XMLElement* position = new XMLElement( "vgs", "position" );
		position->AddAttributeFloat( "x", _position.x );
		position->AddAttributeFloat( "y", _position.y );
		position->AddAttributeFloat( "z", _position.z );
		return position;
	}
	else if ( request->IsNamed( "vgs", "walking" ) )
	{
		if ( _walkTarget.empty() )
		{
			return NULL;
		}
		else
		{
			return new XMLElement( "vgs", "walking" );
		}
	}
	else if ( request->IsNamed( "vgs", "speaking" ) )
	{
		if ( IsPlaying() )
		{
			return new XMLElement( "vgs", "speaking" );
		}
		else
		{
			return NULL;
		}
	}
	else if ( request->IsNamed( "bml" ) )
	{
		request = request->Clone();
		
		for ( XMLElementIterator it = request->GetElements() ; it ; )
		{
			XMLElement* req = *it;
			++ it;
			
			if ( req->IsNamed( "vgs", "walk" ) )
			{
				if ( PCSTR target = req->GetAttributeString( "target" ) )
				{
					if ( _walkTarget.empty() )
					{
						XMLElement update( "bml" );
						update.AddElement( "body" )->AddAttribute( "posture", "Walk" );
						SendImpl( &update );
					}
					
					_walkTarget		= target;
					_walkDistance	= req->GetAttributeFloat( "proximity", 0.2f );
				}
				else if ( ! _walkTarget.empty() )
				{
					_walkTarget.clear();
					
					XMLElement update( "bml" );
					update.AddElement( "body" )->AddAttribute( "posture", _posture.c_str() );
					SendImpl( &update );
				}
				
				req->Delete();
			}
			else if ( req->IsNamed( "vgs", "speak" ) )
			{
				Silence();
				
				if ( PCSTR file = req->GetAttributeString( "file" ) )
				{
					Play( file );
				}
				
				req->Delete();
			}
			else if ( req->IsNamed( "body" ) )
			{
				if ( XMLAttribute* posture = req->GetAttribute( "posture" ) )
				{
					if ( _posture == posture->GetValue() )
					{
						req->Delete();
					}
					else
					{
						_posture = posture->GetValue();
						if ( ! _walkTarget.empty() ) req->Delete();
					}
				}
			}
			else if ( req->IsNamed( "gaze" ) )
			{
				if ( XMLAttribute* target = req->GetAttribute( "target" ) )
				{
					_gazeTarget = target->GetValue();
					_gazeLast = _gazeTarget;
				}
				
				_gazeWander	= req->GetAttributeFloat( "wander" );
				_gazeDelay	= (scalar)( 2 + ( rand() % 3 ) );
			}
		}
		
		if ( request->GetElements() ) SendImpl( request );
		
		request->Delete();
		
		return NULL;
	}
	else
	{
		SendImpl( request );
		return NULL;
	}
}

////////////////////////////////////////////////////////////////////////////////
// Actor - bone bus connection

void Actor::BoneBusConnected()
{
	SetID( -1 );
	GetActors()->SendSBM( "char " + _name + " init common.sk CLASSNAME" );
	GetActors()->SendSBM( "set char " + _name + " voice audiofile speech" );
	
	_posture.clear();
	_gazeTarget.clear();
	_walkTarget.clear();
}

////////////////////////////////////////////////////////////////////////////////
// Actor - bone bus rotations

void Actor::BoneBusRotations( int count, const Actors::BulkBoneRotation* bone )
{
	if ( NULL == _character ) return;
	
	for ( int i = 0 ; i < count ; i ++, bone ++ )
	{
		PCSTR boneName	= Actors::BONE_NAMES[ bone->bone ];
		int boneIndex	= _character->FindBone( boneName );

		if ( boneIndex >= 0 )
		{
			_orientations[ boneIndex ] = quaternion(
				-bone->w,
				bone->x,
				bone->y,
				-bone->z );
		}
		else cerr << "Missing bone " << boneName << endl;
	}
}

////////////////////////////////////////////////////////////////////////////////
// Actor - bone bus positions

void Actor::BoneBusPositions( int count, const Actors::BulkBonePosition* bone )
{
	if ( NULL == _character ) return;
	
	for ( int i = 0 ; i < count ; i ++, bone ++ )
	{
		PCSTR boneName	= Actors::BONE_NAMES[ bone->bone ];
		int boneIndex	= _character->FindBone( boneName );
		
		if ( boneIndex >= 0 )
		{
			_positions[ boneIndex ] = vec3(
				-bone->x,
				bone->y,
				bone->z );
		}
		else cerr << "Missing bone " << boneName << endl;
	}
}

////////////////////////////////////////////////////////////////////////////////
// Actor - update

void Actor::Update()
{
	UpdateWalk();
	UpdatePhysics();
	
	if ( NULL != _character )
	{
		_character->Calculate( &_positions.front(), &_orientations.front(), &_transforms.front() );
	}
	
	if ( NULL != _emitter )
	{
		_emitter->SetPosition( _position );
	}
	
	GetActors()->RegisterPosition( _name.c_str(), _position, _heading );
}

////////////////////////////////////////////////////////////////////////////////
// Actor - render

void Actor::Render()
{
	if ( NULL == _character ) return;
	
	glPushMatrix();
	glTranslatef( _position.x, _position.y, _position.z );
	glRotatef( _heading * 180.0f / M_PI, 0, 1, 0 );
	glRotatef( -12, 1, 0, 0 );
	glTranslatef( 0, 1.4f, 0.75f );
	
	_character->Render( &_transforms.front() );
	
	glPopMatrix();
}

////////////////////////////////////////////////////////////////////////////////
// Actor - get the Actors instance

Actors* Actor::GetActors()
{
	return Application::GetInstance()->GetActors();
}

////////////////////////////////////////////////////////////////////////////////
// Actor - send an XML request (implementation)

void Actor::SendImpl( XMLElement* request )
{
	GetActors()->SendSBM( "test bml char " + _name + " " + request->ToString( false, false ) );
}

////////////////////////////////////////////////////////////////////////////////
// Actor - set the character

void Actor::SetCharacter( Character* character )
{
	if ( _character == character ) return;
	
	_character->Release();
	
	UINT bones = 0;
	
	if ( _character = character )
	{
		_character->AddRef();
		bones = _character->GetBoneCount();
	}
	
	_positions.clear();
	_orientations.clear();
	_positions.resize( bones );
	_orientations.resize( bones );
	_transforms.resize( bones );
}

////////////////////////////////////////////////////////////////////////////////
// Actor - play a sound

bool Actor::Play( PCSTR name, int channel, bool loop, scalar volume )
{
	Audio* audio = Application::GetInstance()->GetAudio();
	if ( NULL == audio ) return false;
	
	string local;
	
	if ( NULL == strchr( name, '/' ) )
	{
		local = "characters/" + _character->GetName() + "/sounds/" + name;
	}
	else
	{
		local = name;
	}
	
	Sound* sound = audio->LoadSound( local.c_str() );
	
	if ( NULL == sound ) sound = audio->LoadSound( name );
	if ( NULL == sound ) return false;
	
	if ( NULL == _emitter ) _emitter = audio->Register();
	
	_emitter->Play( sound, channel, loop, volume );
	sound->Release();
	
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// Actor - silence one or more sound channels

void Actor::Silence( int channels )
{
	if ( NULL != _emitter )
	{
		_emitter->Silence( channels );
	}
}

////////////////////////////////////////////////////////////////////////////////
// Actor - check for sound on one or more sound channels

bool Actor::IsPlaying( int channel )
{
	return ( NULL != _emitter ) ? _emitter->IsPlaying( channel ) : false;
}

////////////////////////////////////////////////////////////////////////////////
// Actor - set the position

void Actor::SetPosition( const vec3& position )
{
	_position = position;
	
	if ( NULL != _controller )
	{
		NxExtendedVec3 pos( _position.x, _position.y + _radius * 5, _position.z );
		_controller->setPosition( pos );
	}
}

////////////////////////////////////////////////////////////////////////////////
// Actor - update the physics

void Actor::UpdatePhysics()
{
	if ( NULL == _controller ) return;
	
	scalar delta = Application::GetInstance()->GetDelta();
	if ( delta <= 0 ) return;
	
	_velocity.y += _gravity * delta;
	
	vec3 scaled( _velocity * delta );
	
	NxU32 flags = 0;
	NxVec3 nxScaled( scaled.x, scaled.y, scaled.z );
	_controller->move( nxScaled, 0xFFFFFFFF, 0, flags );
	
	const NxExtendedVec3& nxPos( _controller->getDebugPosition() );
	vec3 pos( (scalar)nxPos.x, (scalar)nxPos.y - _radius * 5, (scalar)nxPos.z );
	
	scaled = ( ( pos - _position ) / delta );
	if ( scaled.y > _velocity.y ) scaled.y = _velocity.y;
	_velocity = scaled;
	_position = pos;
	
	scaled = _velocity;
	scaled.y = 0;
	_velocity -= scaled * _friction * delta;
}

////////////////////////////////////////////////////////////////////////////////
// Actor - update the walking

void Actor::UpdateWalk()
{
	string gazeTarget = _walkTarget;
	if ( gazeTarget.empty() ) gazeTarget = _gazeTarget;
	if ( gazeTarget.empty() ) return;
	
	vec3 toward( GetActors()->GetPosition( gazeTarget.c_str() ) - _position );
	toward.y = 0;
	scalar distance = toward.Length();
	
	if ( distance < 0.01f )
	{
		if ( ! _walkTarget.empty() )
		{
			_walkTarget.clear();
			
			XMLElement update( "bml" );
			update.AddElement( "body" )->AddAttribute( "posture", _posture.c_str() );
			SendImpl( &update );
		}
		
		return;
	}
	
	toward.Normalize();
	scalar heading = GetHeading( toward );
	
	scalar diff = heading - _heading;
	
	if ( abs( diff ) >= M_PI * ( _walkTarget.empty() ? 0.2 : 0.1f ) )
	{
		while ( diff < -M_PI ) diff += M_PI * 2;
		while ( diff > M_PI ) diff -= M_PI * 2;
		
		scalar turn = Application::GetInstance()->GetDelta() * M_PI * 2.0f;
		if ( 0 == turn ) return;
		
		if ( distance > 0.3f )
		{
			if ( fabs( diff ) > turn ) diff = ( diff > 0 ) ? turn : -turn;
		}
		
		_heading += diff;
		
		while ( _heading < 0 ) _heading += M_PI * 2;
		while ( _heading >= M_PI * 2 ) _heading -= M_PI * 2;
	}
	
	if ( _gazeWander > 0 )
	{
		_gazeDelay -= Application::GetInstance()->GetDelta();
		if ( _gazeDelay <= 0 ) UpdateGaze();
	}
	
	if ( _walkTarget.empty() ) return;
	
	if ( distance < _walkDistance )
	{
		_walkTarget.clear();
		
		XMLElement update( "bml" );
		update.AddElement( "body" )->AddAttribute( "posture", _posture.c_str() );
		SendImpl( &update );
	}
	else
	{
		quaternion quat( _heading, vec3( 0, -1, 0 ) );
		scalar speed = ( _name == "guide" ) ? ( ( distance > 5.0f ) ? 15.0f : 5.0f ) : 8.0f;
		_velocity += quat.Direction() * speed * Application::GetInstance()->GetDelta();
	}
}

////////////////////////////////////////////////////////////////////////////////
// Actor - get a 2D heading from a direction vector

scalar Actor::GetHeading( const vec3& toward )
{
	if ( toward.x > 0 )
	{
		if ( toward.z > 0 )
			return atanf( toward.x / toward.z );
		else if ( toward.z < 0 )
			return M_PI * 0.5f + atanf( -toward.z / toward.x );
		else
			return M_PI * 0.5f;
	}
	else if ( toward.x < 0 )
	{
		if ( toward.z > 0 )
			return M_PI * 2 - atanf( -toward.x / toward.z );
		else if ( toward.z < 0 )
			return M_PI + atanf( toward.x / toward.z );
		else
			return M_PI * 1.5f;
	}
	else
	{
		return ( toward.z > 0 ) ? 0 : M_PI;
	}
}

////////////////////////////////////////////////////////////////////////////////
// Actor - update the wandering gaze

void Actor::UpdateGaze()
{
	quaternion quat( _heading, vec3( 0, -1, 0 ) );
	vec3 forward( quat.Direction() );
	
	scalar limit = cosf( _gazeWander * M_PI / 180.0f );
	
	const map<string,vec3>& pawns = GetActors()->GetPawns();
	vector<string> candidates;
	
	for ( map<string,vec3>::const_iterator it = pawns.begin() ; it != pawns.end() ; ++ it )
	{
		if ( it->first == _gazeLast ) continue;
		
		vec3 delta( it->second - _position );
		delta.y = 0;
		
		scalar distance = delta.Length();
		if ( distance < 0.25f ) continue;
		delta /= distance;
		
		if ( forward.DotProduct( delta ) < limit ) continue;
		
		int count = 1;
		
		if ( it->first.find( "_" ) == string::npos ) count = 10;
		else if ( it->first.find( "_l" ) == string::npos ) count = 0;
		
		while ( count -- > 0 )
		{
			candidates.push_back( it->first );
		}
	}
	
	_gazeLast = _gazeTarget;
	
	if ( ! candidates.empty() )
	{
		int index = rand() % (int)candidates.size();
		_gazeLast = candidates[ index ];
	}
	
	_gazeDelay = (scalar)( 2 + ( rand() % 2 ) );
	
	XMLElement bml( "bml" );
	bml.AddElement( "gaze" )->AddAttribute( "target", _gazeLast.c_str() );
	SendImpl( &bml );
	
	cout << "Actor " << _name << " switching gaze to " << _gazeLast << endl;
}
