//
// Actors.cpp
// Copyright Michael J Stokes, 2009.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "Application.h"
#include "Renderer.h"
#include "Actors.h"
#include "Actor.h"
#include "Ego.h"
#include "Features.h"

#ifdef IS_EDITOR
#include "../Editor/BMLFrame.h"
#endif


////////////////////////////////////////////////////////////////////////////////
// Actors - construction

Actors::Actors()
{
	WSADATA data;
	WSAStartup( 0x0101, &data );
	
	_tcp		= socket( AF_INET, SOCK_STREAM, IPPROTO_TCP );
	_udp		= socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP );
	_client		= INVALID_SOCKET;
	
	u_long one = 1;
	ioctlsocket( _tcp, FIONBIO, &one );
	ioctlsocket( _udp, FIONBIO, &one );
	
	sockaddr_in sa;
	memset( &sa, 0, sizeof( sa ) );
	
	sa.sin_family	= AF_INET;
	sa.sin_port		= htons( 15102 );
	
	if ( 0 != bind( _tcp, (sockaddr*)&sa, sizeof( sa ) ) )
	{
		cerr << "Unable to listen on TCP port " << ntohs( sa.sin_port ) << "." << endl;
	}
	
	sa.sin_port		= htons( 15100 );
	
	if ( 0 != bind( _udp, (sockaddr*)&sa, sizeof( sa ) ) )
	{
		cerr << "Unable to listen on UDP port " << ntohs( sa.sin_port ) << "." << endl;
	}
	
	listen( _tcp, 3 );
	
	cout << "Listening for BoneBus connections..." << endl;
}

////////////////////////////////////////////////////////////////////////////////
// Actors - destruction

Actors::~Actors()
{
	if ( INVALID_SOCKET != _client ) closesocket( _client );
	closesocket( _udp );
	closesocket( _tcp );
	
	while ( ! _list.empty() )
	{
		delete _list.back();
		_list.pop_back();
	}
}

////////////////////////////////////////////////////////////////////////////////
// Actors - update

void Actors::Update()
{
	UpdateBoneBus();
	UpdatePositions();
	
	for ( list<Actor*>::iterator it = _list.begin() ; it != _list.end() ; ++ it )
	{
		( *it )->Update();
	}
}

////////////////////////////////////////////////////////////////////////////////
// Actors - render

void Actors::Render()
{
	for ( list<Actor*>::iterator it = _list.begin() ; it != _list.end() ; ++ it )
	{
		( *it )->Render();
	}
}

////////////////////////////////////////////////////////////////////////////////
// Actors - send an XML request

XMLElement* Actors::Send( PCSTR name, XMLElement* request )
{
#ifdef IS_EDITOR
	if ( BMLFrame* frame = BMLFrame::GetInstance() ) frame->AddXML( name, request );
#endif
	
	if ( request->IsNamed( "vgs", "create" ) )
	{
		if ( Actor* actor = FindActor( name ) )
		{
			delete actor;
			_list.remove( actor );
		}
		
		Actor* actor = new Actor( name, request );
		_list.push_back( actor );
	}
	else if ( request->IsNamed( "vgs", "destroy" ) )
	{
		if ( Actor* actor = FindActor( name ) )
		{
			delete actor;
			_list.remove( actor );
		}
	}
	else
	{
		if ( Actor* actor = FindActor( name ) )
		{
			return actor->Send( request );
		}
	}
	
	return NULL;
}

////////////////////////////////////////////////////////////////////////////////
// Actors - register a named position

void Actors::RegisterPosition( PCSTR name, const vec3& position, scalar heading )
{
	vec3 pos( position * 100.0f );
	char command[ 1024 ];
	
	if ( heading < 0 )
	{
		if ( _pawns.end() == _pawns.find( name ) )
		{
			sprintf( command, "pawn %s init", name );
			SendSBM( command );
		}
	}
	
	_pawns[ name ] = position;
	
	if ( heading < 0 )
	{
		sprintf( command, "set pawn %s world_offset x %f y %f z %f", name, -pos.x, pos.y, pos.z );
	}
	else
	{
		sprintf( command, "set char %s world_offset x %f y %f z %f h %f", name, -pos.x, pos.y, pos.z, 360.0f - heading * 180.0f / M_PI );
	}
	
	SendSBM( command );
}

////////////////////////////////////////////////////////////////////////////////
// Actors - send a raw smartbody request

void Actors::SendSBM( const string& command )
{
	if ( INVALID_SOCKET != _client )
	{
		string encoded( command + "\r\r\r" );
		_transmit.Add( encoded.c_str(), (UINT)encoded.length() );
		_transmit.Transmit( _client );
	}
}

////////////////////////////////////////////////////////////////////////////////
// Actors - find an actor by name

Actor* Actors::FindActor( PCSTR name )
{
	for ( list<Actor*>::iterator it = _list.begin() ; it != _list.end() ; ++ it )
	{
		if ( ( *it )->GetName() == name ) return *it;
	}
	
	return NULL;
}

////////////////////////////////////////////////////////////////////////////////
// Actors - find an actor by ID

Actor* Actors::FindActor( int id )
{
	for ( list<Actor*>::iterator it = _list.begin() ; it != _list.end() ; ++ it )
	{
		if ( ( *it )->GetID() == id ) return *it;
	}
	
	return NULL;
}

////////////////////////////////////////////////////////////////////////////////
// Actors - update the BoneBus connection

void Actors::UpdateBoneBus()
{
	for ( ; ; )
	{
		sockaddr_in sa;
		int len = sizeof( sa );
		
		SOCKET client = accept( _tcp, (sockaddr*)&sa, &len );
		if ( INVALID_SOCKET == client ) break;
		
		cout << "SmartBody process connected from " << inet_ntoa( sa.sin_addr ) << "." << endl;
		
		_client = client;
		_transmit.Empty();
		_receive.Empty();
		_pawns.clear();
		
		BoneBusConnected();
	}
	
	if ( INVALID_SOCKET != _client )
	{
		_transmit.Transmit( _client );
		
		if ( _receive.Receive( _client ) )
		{
			while ( _receive.GetLength() > 0 )
			{
				PSTR ptr = (PSTR)_receive.GetPointer();
				UINT len = _receive.GetLength();
				UINT pak = 0;
				
				for ( ; pak < len ; pak ++ )
				{
					if ( ptr[ pak ] == ';' ) break;
				}
				
				if ( pak >= len ) break;
				
				vector<string> args;
				UINT last = 0;
				
				for ( len = 0 ; len <= pak ; len ++ )
				{
					if ( ( ptr[ len ] == '|' ) || ( ptr[ len ] == ';' ) )
					{
						args.push_back( string( ptr + last, len - last ) );
						last = len + 1;
					}
				}
				
				if ( ! args.empty() ) HandleTCP( args );
				
				_receive.Remove( pak + 1 );
			}
		}
		else
		{
			cout << "SmartBody connection lost." << endl;
			
			closesocket( _client );
			_client = INVALID_SOCKET;
			
			for ( list<Actor*>::iterator it = _list.begin() ; it != _list.end() ; ++ it )
			{
				( *it )->SetID( -1 );
			}
		}
	}
	
	for ( ; ; )
	{
		sockaddr_in sa;
		int sal = sizeof( sa );
		
		static BulkBoneRotations buffer;
		
		int len = recvfrom( _udp, (char*)&buffer, sizeof( buffer ), 0, (sockaddr*)&sa, &sal );
		if ( len < 0 ) break;
		
		if ( INVALID_SOCKET == _client ) continue;
		
		if ( len < sizeof( int ) * 3 ) continue;
		len -= sizeof( int ) * 3;
		
		if ( 0x10 == buffer.packet )
		{
			len = ( len - sizeof( int ) ) / sizeof( BulkBoneRotation );
			
			if ( buffer.count <= 0 ) continue;
			if ( buffer.count > len ) continue;
			
			if ( Actor* actor = FindActor( buffer.character ) )
			{
				actor->BoneBusRotations( buffer.count, buffer.bones );
			}
		}
		else if ( 0x11 == buffer.packet )
		{
			len = len / sizeof( BulkBonePosition );
			
			if ( buffer.count <= 0 ) continue;
			if ( buffer.count > len ) continue;
			
			if ( Actor* actor = FindActor( buffer.character ) )
			{
				BulkBonePositions* positions = (BulkBonePositions*)&buffer;
				actor->BoneBusPositions( positions->count, positions->bones );
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
// Actors - handle BoneBus connection

void Actors::BoneBusConnected()
{
	for ( list<Actor*>::iterator it = _list.begin() ; it != _list.end() ; ++ it )
	{
		( *it )->BoneBusConnected();
	}
	
	UpdatePositions();
}

////////////////////////////////////////////////////////////////////////////////
// Actors - handle a TCP command vector

void Actors::HandleTCP( vector<string>& args )
{
	if ( args[ 0 ] == "CreateActor" )
	{
		if ( args.size() >= 5 )
		{
			if ( Actor* actor = FindActor( args[ 3 ].c_str() ) )
			{
				actor->SetID( atoi( args[ 1 ].c_str() ) );
			}
			else
			{
				// TODO: Something else?
			}
		}
	}
	else if ( args[ 0 ] == "DeleteActor" )
	{
		if ( args.size() >= 2 )
		{
			if ( Actor* actor = FindActor( atoi( args[ 1 ].c_str() ) ) )
			{
				actor->SetID( -1 );
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
// Actors - update positions

void Actors::UpdatePositions()
{
	if ( Ego* ego = Application::GetInstance()->GetEgo() )
	{
		vec3 pos( ego->GetPosition() );
		pos.y += 0.75f;
		RegisterPosition( "ego", pos );
	}
	
	Application::GetInstance()->GetFeatures()->RegisterPositions();
}

////////////////////////////////////////////////////////////////////////////////
// Actors - bone names

const PCSTR Actors::BONE_NAMES[] =
{
	"",
	"",
	"base",
	"l_hip",
	"l_knee",
	"l_ankle",
	"l_forefoot",
	"l_toe",
	"r_hip",
	"r_knee",
	"r_ankle",
	"r_forefoot",
	"r_toe",
	"spine1",
	"spine2",
	"spine3",
	"spine4",
	"spine5",
	"skullbase",
	"face_top_parent",
	"brow_parent_left",
	"brow01_left",
	"brow02_left",
	"brow03_left",
	"brow04_left",
	"brow_parent_right",
	"brow01_right",
	"brow02_right",
	"brow03_right",
	"brow05_right",
	"ear_left",
	"eyeball_left",
	"upper_nose_left",
	"lower_nose_left",
	"upper_nose_right",
	"lower_nose_right",
	"lower_eyelid_right",
	"upper_eyelid_right",
	"eyeball_right",
	"ear_right",
	"lower_eyelid_left",
	"upper_eyelid_left",
	"joint18",
	"face_bottom_parent",
	"Jaw",
	"Jaw_back",
	"Jaw_front",
	"Lip_bttm_mid",
	"Lip_bttm_right",
	"Lip_bttm_left",
	"Tongue_back",
	"Tongue_mid",
	"Tongue_front",
	"Lip_top_left",
	"Lip_top_right",
	"Cheek_low_right",
	"Cheek_up_right",
	"cheek_low_left",
	"Cheek_up_left",
	"Lip_out_left",
	"Lip_out_right",
	"Lip_top_mid",
	"l_sternoclavicular",
	"l_acromioclavicular",
	"l_shoulder",
	"l_elbow",
	"l_forearm",
	"l_wrist",
	"l_pinky1",
	"l_pinky2",
	"l_pinky3",
	"l_pinky4",
	"l_ring1",
	"l_ring2",
	"l_ring3",
	"l_ring4",
	"l_middle1",
	"l_middle2",
	"l_middle3",
	"l_middle4",
	"l_index1",
	"l_index2",
	"l_index3",
	"l_index4",
	"l_thumb1",
	"l_thumb2",
	"l_thumb3",
	"l_thumb4",
	"r_sternoclavicular",
	"r_acromioclavicular",
	"r_shoulder",
	"r_elbow",
	"r_forearm",
	"r_wrist",
	"r_pinky1",
	"r_pinky2",
	"r_pinky3",
	"r_pinky4",
	"r_ring1",
	"r_ring2",
	"r_ring3",
	"r_ring4",
	"r_middle1",
	"r_middle2",
	"r_middle3",
	"r_middle4",
	"r_index1",
	"r_index2",
	"r_index3",
	"r_index4",
	"r_thumb1",
	"r_thumb2",
	"r_thumb3",
	"r_thumb4",
	NULL
};
