//
// Application.cpp
// Copyright Michael J Stokes, 2008.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "Application.h"
#include "Renderer.h"
#include "Audio.h"
#include "Physics.h"
#include "Environment.h"
#include "Features.h"
#include "Actors.h"
#include "PhysicsEgo.h"
#include "Guide.h"
#include "FileSearch.h"

#if defined( WITH_SDL )
#include "../App/SDLAudio.h"
#elif defined( _WIN32 )
#include "WinAudio.h"
#endif

Application* Application::_instance = NULL;


////////////////////////////////////////////////////////////////////////////////
// Application - construction

Application::Application()
{
	_instance		= this;
	_root			= NULL;
	
	_renderer		= new Renderer();
#if defined( WITH_SDL )
	_audio			= new SDLAudio();
#elif defined( _WIN32 )
	_audio			= new WinAudio();
#else
	_audio			= NULL;
#endif
	_physics		= new Physics();
	_environment	= new Environment();
	_features		= new Features();
	_actors			= new Actors();
	_ego			= NULL;
	_guide			= NULL;
	_time			= 0;
	_queryPixel.x	= -1;
}

////////////////////////////////////////////////////////////////////////////////
// Application - destruction

Application::~Application()
{
	if ( NULL != _guide ) delete _guide;
	if ( NULL != _ego ) delete _ego;
	delete _actors;
	delete _features;
	delete _environment;
	delete _physics;
	if ( NULL != _audio ) delete _audio;
	delete _renderer;
	
	_root->Delete();
	_instance = NULL;
}

////////////////////////////////////////////////////////////////////////////////
// Application - create a new project

bool Application::Create( PCSTR path )
{
#ifdef IS_EDITOR
	FileSearch search( path );
	string file;
	
	if ( search.GetNext( file ) )
	{
		wxMessageBox( wxString( T("Project directory '") ) + path + T("' is not empty, so a new project cannot be created there.") );
		return false;
	}
	
	XMLElement root( "root" );
	root.AddAttribute( "xmlns", "urn:no.ntnu.idi.vgs:root:1.1" );
	
	if ( ! root.ToFile( ( string( path ) + "/root.xml" ).c_str() ) )
	{
		wxMessageBox( wxString( T("Unable to create project XML file '") ) + string( path ) + T("/root.xml'.") );
		return false;
	}
	
	wxMkdir( ( string( path ) + "/characters" ).c_str() );
	wxMkdir( ( string( path ) + "/environment" ).c_str() );
	wxMkdir( ( string( path ) + "/features" ).c_str() );
	wxMkdir( ( string( path ) + "/system" ).c_str() );
	
	root.SetName( "environment" );
	root.AddAttribute( "xmlns", "urn:no.ntnu.idi.vgs:environment:1.0" );
	root.ToFile( ( string( path ) + "/environment/environment.xml" ).c_str() );
	
	return Load( path );
#else
	return false;
#endif
}

////////////////////////////////////////////////////////////////////////////////
// Application - load a project

bool Application::Load( PCSTR path )
{
	cout << "Loading root configuration file '" << path << "/root.xml'..." << endl;
	
	_path = path;
	_root = XMLElement::FromFile( ( _path + "/root.xml" ).c_str() );
	
	if ( ! _root->IsNamed( "root" ) )
	{
		_root->Delete();
		_root = NULL;
		cerr << "Unable to open root configuration file '" << path << "/root.xml'" << endl;
		return false;
	}
	
	if ( ! _renderer->Initialize() ) return false;
	if ( ( NULL != _audio ) && ( ! _audio->Initialize() ) ) return false;
	
	if ( ! _environment->Load() ) return false;
	if ( ! _features->Load() ) return false;
	
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// Application - save a project

bool Application::Save()
{
#ifdef IS_EDITOR
	if ( NULL == _root ) return false;
	
	cout << "Saving root configuration file '" << _path << "/root.xml'..." << endl;
	
	if ( ! _root->ToFile( ( _path + "/root.xml" ).c_str() ) )
	{
		wxMessageBox( wxString( T("Unable to create project XML file '") ) + _path + T("/root.xml'.") );
		return false;
	}
	
	_environment->Save();
	
	return true;
#else
	return false;
#endif
}

////////////////////////////////////////////////////////////////////////////////
// Application - start simulation

bool Application::Start()
{
	Stop();
	
	if ( NULL == _root ) return false;
	
	_ego	= new PhysicsEgo();
	_guide	= new Guide();
	
	if ( _ego->Load( _root ) && _guide->Load( _root ) ) return true;
	
	Stop();
	
	return false;
}

////////////////////////////////////////////////////////////////////////////////
// Application - stop simulation

void Application::Stop()
{
	if ( NULL != _guide )
	{
		delete _guide;
		_guide = NULL;
	}
	
	if ( NULL != _ego )
	{
		delete _ego;
		_ego = NULL;
	}
}

////////////////////////////////////////////////////////////////////////////////
// Application - prepare an OpenGL context

bool Application::PrepareContext()
{
	return _renderer->PrepareContext();
}

////////////////////////////////////////////////////////////////////////////////
// Application - update

void Application::Update( scalar time, const vec3& direction, UINT actions )
{
	_delta	= ( _time > 0 ) ? ( time - _time ) : 0;
	_time	= time;
	
	_actors->Update();
	
	_physics->Start( _delta );
	_physics->Finish();
	
	if ( NULL != _ego )
	{
        _ego->Update( direction, actions );
		if ( NULL != _audio ) _audio->SetListener( _ego->GetPosition(), direction, vec3( 0, 1, 0 ).CrossProduct( direction ) );
	}
	
	if ( NULL != _guide ) _guide->Update();
}

////////////////////////////////////////////////////////////////////////////////
// Application render

void Application::Render()
{
	if ( NULL != _ego ) _ego->PreRender();
	
	_environment->Render();
	
	if ( _queryPixel.x >= 0 )
	{
		scalar depth = 0;
		glReadPixels( (int)_queryPixel.x, (int)_queryPixel.y, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &depth );
		
		if ( depth < 1 )
		{
			GLint viewport[ 4 ];
			glGetIntegerv( GL_VIEWPORT, viewport );
			
			GLdouble modelView[ 16 ], projection[ 16 ], output[ 3 ];
			glGetDoublev( GL_MODELVIEW_MATRIX, modelView );
			glGetDoublev( GL_PROJECTION_MATRIX, projection );	
			
			gluUnProject(	(GLdouble)_queryPixel.x, (GLdouble)_queryPixel.y, (GLdouble)depth,
							modelView, projection, viewport, output + 0, output + 1, output + 2 );
			
			_queryResult.x = (scalar)output[ 0 ];
			_queryResult.y = (scalar)output[ 1 ];
			_queryResult.z = (scalar)output[ 2 ];
		}
	}
	
	_features->Render();
	_actors->Render();
}
