//
// Audio.cpp
// Copyright (c) Michael J Stokes, 2007.  All rights reserved.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "Application.h"
#include "Audio.h"
#include "Sound.h"


////////////////////////////////////////////////////////////////////////////////
// Audio - construction

Audio::Audio()
{
	_active = false;
}

////////////////////////////////////////////////////////////////////////////////
// Audio - destruction

Audio::~Audio()
{
	Terminate();
	
	while ( ! _sounds.empty() )
	{
		_sounds.front()->Release();
		_sounds.pop_front();
	}
}

////////////////////////////////////////////////////////////////////////////////
// Audio - initialize

bool Audio::Initialize()
{
	_active = true;
	Pause( false );
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// Audio - terminate

void Audio::Terminate()
{
	while ( ! _cleanup.empty() )
	{
		delete _cleanup.back();
		_cleanup.pop_back();
	}
}

////////////////////////////////////////////////////////////////////////////////
// Audio - make sure everything is synchronized

void Audio::Synchronize()
{
	Lock();
	
	for ( list<Source*>::iterator it = _sources.begin() ; it != _sources.end() ; )
	{
		Source* source = *it ++;
		
		if ( NULL == source->emitter )
		{
			_sources.remove( source );
			delete source;
		}
	}
	
	while ( ! _cleanup.empty() )
	{
		delete _cleanup.back();
		_cleanup.pop_back();
	}
	
	Unlock();
}

////////////////////////////////////////////////////////////////////////////////
// Audio - register and return a new audio emitter

Emitter* Audio::Register()
{
	Lock();
	
	Emitter* emitter = new Emitter( this );
	_emitters.push_back( emitter );
	
	Unlock();
	
	return emitter;
}

////////////////////////////////////////////////////////////////////////////////
// Audio - delete an audio emitter

void Audio::Delete( Emitter* emitter )
{
	Lock();	
	
	_emitters.remove( emitter );
	delete emitter;
	
	Unlock();
}

////////////////////////////////////////////////////////////////////////////////
// Audio - set the listener

void Audio::SetListener( const vec3& position, const vec3& forward, const vec3& right )
{
	if ( _active )
	{
		Lock();
		
		while ( ! _cleanup.empty() )
		{
			delete _cleanup.back();
			_cleanup.pop_back();
		}
		
		for ( list<Emitter*>::iterator it = _emitters.begin() ; it != _emitters.end() ; ++ it )
		{
			Emitter* emitter = *it;
			
			vec3 toward( emitter->_position - position );
			scalar distance = toward.Length();
			scalar volume;
			
			if ( distance < 1 )
			{
				volume = 127;
			}
			else if ( distance < 30 )
			{
				volume = distance / 30;
				volume = 127 - volume * 127;
				if ( volume < 0 ) volume = 0;
			}
			else
			{
				volume = 0;
			}
			
			toward /= distance;
			
			scalar dot = ( toward.DotProduct( right ) + 1.0f ) / 2.0f;
			
			if ( dot < 0 ) dot = 0;
			else if ( dot > 1 ) dot = 1;
			
			emitter->_left	= (UINT)( ( 1.0f - dot ) * volume );
			emitter->_right	= (UINT)( dot * volume );
		}
		
		Unlock();
	}
}

////////////////////////////////////////////////////////////////////////////////
// Audio - process

void Audio::Process( int* output, UINT samples )
{
	memset( output, 0, samples * 8 );
	
	for ( list<Source*>::iterator it = _sources.begin() ; it != _sources.end() ; )
	{
		Source* source = *it ++;
		
		if ( NULL == source->emitter )
		{
			_sources.remove( source );
			_cleanup.push_back( source );
			continue;
		}
		
		int left	= ( source->volume * source->emitter->_left ) >> 7;
		int right	= ( source->volume * source->emitter->_right ) >> 7;
		
		Sound::Instance* instance = (Sound::Instance*)source->sound;
		
		if ( ! instance->Mix( output, samples, left, right, source->loop ) )
		{
			_sources.remove( source );
			_cleanup.push_back( source );
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
// Audio - load and cache a sound

Sound* Audio::LoadSound( PCSTR name )
{
	string lname( Application::GetInstance()->GetPath() + "/" + name );
	
	for ( list<Sound*>::iterator it = _sounds.begin() ; it != _sounds.end() ; ++ it )
	{
		if ( ( * it )->IsNamed( lname.c_str() ) )
		{
			( *it )->AddRef();
			return *it;
		}
	}
	
	Sound* sound = new Sound( lname.c_str() );
	
	if ( sound->Load() )
	{
		_sounds.push_back( sound );
		sound->AddRef();
		return sound;
	}
	else
	{
		sound->Release();
		return NULL;
	}
}

////////////////////////////////////////////////////////////////////////////////
// Emitter - play a sound on a channel

void Emitter::Play( Sound* sound, UINT channel, bool loop, scalar volume )
{
	if ( _audio->IsActive() && ( 0 != channel ) )
	{
		Sound::Instance* instance	= sound->CreateInstance();
		Audio::Source* source		= new Audio::Source;
		
		source->emitter	= this;
		source->sound	= instance;
		source->channel	= channel;
		source->loop	= loop;
		source->volume	= (int)( volume * 127 );
		
		_audio->Lock();
		_sources.push_back( source );
		_audio->_sources.push_back( source );
		_audio->Unlock();
	}
}

////////////////////////////////////////////////////////////////////////////////
// Emitter - silence one or more channels

void Emitter::Silence( UINT channels )
{
	_audio->Lock();
	
	for ( list<Audio::Source*>::iterator it = _sources.begin() ; it != _sources.end() ; )
	{
		Audio::Source* source = *it ++;
		
		if ( source->channel & channels )
		{
			source->emitter = NULL;
			_sources.remove( source );
		}
	}
	
	_audio->Unlock();
}

////////////////////////////////////////////////////////////////////////////////
// Emitter - check if one or more channels are active

bool Emitter::IsPlaying( UINT channels )
{
	_audio->Lock();
	
	for ( list<Audio::Source*>::iterator it = _sources.begin() ; it != _sources.end() ; )
	{
		Audio::Source* source = *it ++;
		
		if ( source->channel & channels )
		{
			_audio->Unlock();
			return true;
		}
	}
	
	_audio->Unlock();
	return false;
}

////////////////////////////////////////////////////////////////////////////////
// Audio::Source - destruction

Audio::Source::~Source()
{
	delete (Sound::Instance*)sound;
	
	if ( NULL != emitter )
	{
		emitter->_sources.remove( this );
	}
}
