//
// Character.cpp
// Copyright Michael J Stokes, 2008.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "Application.h"
#include "Character.h"
#include "Material.h"
#include "MeshCompiler.h"
#include "Shader.h"


////////////////////////////////////////////////////////////////////////////////
// Character - construction

Character::Character( PCSTR name ) : Resource( name )
{
	_scale	= 0.003f;
	_vbuf	= 0;
	_ibuf	= 0;
	
	ZeroMemory( &Bones, sizeof( Bones ) );
}

////////////////////////////////////////////////////////////////////////////////
// Character - destruction

Character::~Character()
{
	for ( vector<Group>::iterator it = _groups.begin() ; it != _groups.end() ; ++ it )
	{
		it->material->Release();
	}
}

///////////////////////////////////////////////////////////////////////////////
// Character - load

bool Character::Load()
{
	cout << "Loading character '" << _name << "'..." << endl;
	return ( LoadSkeleton() && CompileMesh() && LoadMesh() );
}

///////////////////////////////////////////////////////////////////////////////
// Character - load the skeleton

bool Character::LoadSkeleton()
{
	string path( Application::GetInstance()->GetPath() + "/characters/" + _name + "/skeleton.xml" );
	XMLElement* root = XMLElement::FromFile( path.c_str() );
	XMLAutoDelete ad( root );
	
	if ( ! root->IsNamed( "skeleton" ) )
	{
		cerr << "Character skeleton '" << path << "' not found." << endl;
		return false;
	}
	
	_scale = root->GetAttributeFloat( "scale", 1 );
	
	if ( XMLElement* bones = root->GetElement( "bones" ) )
	{
		for ( XMLElementIterator it = bones->GetElements() ; it ; ++ it )
		{
			if ( it->IsNamed( "bone" ) )
			{
				_bones.resize( _bones.size() + 1 );
				Bone& bone = _bones[ _bones.size() - 1 ];
				
				bone.name		= it->GetAttributeString( "name" );
				bone.parent		= NULL;
				
				PCSTR parent	= it->GetAttributeString( "parent" );
				bone.parentName	= parent ? parent : "";
				
				if ( XMLElement* position = it->GetElement( "position" ) )
				{
					bone.position.x = position->GetAttributeFloat( "x" );
					bone.position.y = position->GetAttributeFloat( "y" );
					bone.position.z = position->GetAttributeFloat( "z" );
				}
				
				if ( XMLElement* orientation = it->GetElement( "orientation" ) )
				{
					vec3 axis;
					axis.x = orientation->GetAttributeFloat( "x" );
					axis.y = orientation->GetAttributeFloat( "y" );
					axis.z = orientation->GetAttributeFloat( "z" );
					axis.Normalize();
					bone.orientation = quaternion( orientation->GetAttributeFloat( "angle" ), axis );
				}

				if ( bone.name == "Neck" )
					Bones.Neck = (int)_bones.size();
				else if ( bone.name == "Head" )
					Bones.Head = (int)_bones.size();
				else if ( bone.name == "Jaw" )
					Bones.Jaw = (int)_bones.size();
			}
		}
	}
	
	if ( _bones.empty() )
	{
		cerr << "Character skeleton '" << _name << "' has no bones." << endl;
		return false;
	}
	
	for ( UINT index = 0 ; index < _bones.size() ; index ++ )
	{
		Bone& b = _bones[ index ];
		
		if ( b.parentName.empty() )
		{
			if ( 0 != index )
			{
				Bone& a = _bones[ 0 ];
				a.name.swap( b.name );
				a.parentName.swap( b.parentName );
				std::swap( a.position, b.position );
				std::swap( a.orientation, b.orientation );
			}
			
			break;
		}
	}
	
	for ( UINT index = 0 ; index < _bones.size() ; index ++ )
	{
		Bone& b = _bones[ index ];
		b.id = index;
		
		if ( ! b.parentName.empty() )
		{
			for ( vector<Bone>::iterator it = _bones.begin() ; it != _bones.end() ; ++ it )
			{
				if ( it->name == b.parentName )
				{
					b.parent = &*it;
					it->children.push_back( &b );
					break;
				}
			}
		}
	}
	
	_bones[ 0 ].Prepare( matrix4x4() );
	
	return true;
}

///////////////////////////////////////////////////////////////////////////////
// Character::Bone - prepare the bone

void Character::Bone::Prepare( const matrix4x4& parentTransform )
{
	transform = matrix4x4( orientation.Inverted() ) * matrix4x4( -position ) * parentTransform;
	
	for ( vector<Bone*>::iterator it = children.begin() ; it != children.end() ; ++ it )
	{
		( *it )->Prepare( transform );
	}
}

///////////////////////////////////////////////////////////////////////////////
// Character - load the compiled mesh

bool Character::LoadMesh()
{
	string path( Application::GetInstance()->GetPath() + "/characters/" + _name + "/mesh.bin" );
	std::ifstream stream( path.c_str(), std::ios_base::binary );
	
	if ( ! stream.is_open() )
	{
		cerr << "Compiled character mesh '" << path << "' not found." << endl;
		return false;
	}
	
	char id[ 8 ];
	memset( id, 0, 8 );
	stream.read( id, 8 );
	
	if ( 0 != memcmp( id, "VGS-C-1", 8 ) )
	{
		cerr << "Compiled mesh '" << path << "' is invalid." << endl;
		return false;
	}
	
	UINT32 count;
	
	stream.read( (PSTR)&count, sizeof( count ) );
	_vertices.resize( count );
	_skinned.resize( count );
	stream.read( (PSTR)&_vertices.front(), sizeof( Vertex ) * count );
	
	stream.read( (PSTR)&count, sizeof( count ) );
	_indices.resize( count );
	stream.read( (PSTR)&_indices.front(), sizeof( UINT ) * count );
	
	stream.read( (PSTR)&count, sizeof( count ) );
	_groups.resize( count );

	Renderer* renderer = Application::GetInstance()->GetRenderer();
	
	for ( vector<Group>::iterator it = _groups.begin() ; it != _groups.end() ; ++ it )
	{
		UINT32 sl;
		stream.read( (PSTR)&sl, sizeof( sl ) );
		char* sp = new char[ sl ];
		stream.read( sp, sl );
		string material( sp, sl );
		delete [] sp;
		
		stream.read( (PSTR)&it->firstIndex, sizeof( UINT32 ) );
		stream.read( (PSTR)&it->indexCount, sizeof( UINT32 ) );
		
		string path( Application::GetInstance()->GetPath() + "/characters/" );
		path += _name;
		path += "/textures/";
		path += material;
		
		it->material = renderer->LoadMaterial( path.c_str() );
	}
	
	return true;
}

///////////////////////////////////////////////////////////////////////////////
// Character - compile the mesh if necessary

bool Character::CompileMesh()
{
	string skeleton( Application::GetInstance()->GetPath() + "/characters/" + _name + "/skeleton.xml" );
	string mesh( Application::GetInstance()->GetPath() + "/characters/" + _name + "/mesh.xml" );
	string compiled( Application::GetInstance()->GetPath() + "/characters/" + _name + "/mesh.bin" );
	
	if ( MeshCompiler::IsFileNewer( compiled.c_str(), skeleton.c_str() ) )
	{
		if ( MeshCompiler::IsFileNewer( compiled.c_str(), mesh.c_str() ) ) return true;
	}
	
	cerr << "Character mesh '" << _name << "' needs to be recompiled." << endl;
	
	vector<string> boneNames;
	
	for ( vector<Bone>::iterator it = _bones.begin() ; it != _bones.end() ; ++ it )
	{
		boneNames.push_back( it->name );
	}
	
	return MeshCompiler::CompileSkeleton( mesh.c_str(), compiled.c_str(), boneNames );
}

///////////////////////////////////////////////////////////////////////////////
// Character - upload to graphics hardware

bool Character::Upload()
{
	if ( 0 == _vbuf ) glGenBuffers( 1, &_vbuf );
	if ( 0 == _vbuf ) return false;
	glBindBuffer( GL_ARRAY_BUFFER, _vbuf );
	
	if ( _bones.size() <= 32 )
	{
		glBufferData( GL_ARRAY_BUFFER, _vertices.size() * sizeof( Vertex ), &_vertices.front(), GL_STATIC_DRAW );
	}
	else
	{
		glBufferData( GL_ARRAY_BUFFER, _vertices.size() * sizeof( CPUVertex ), NULL, GL_STREAM_DRAW );
	}
	
	glBindBuffer( GL_ARRAY_BUFFER, 0 );
	
	if ( 0 == _ibuf ) glGenBuffers( 1, &_ibuf );
	if ( 0 == _ibuf ) return false;
	glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, _ibuf );
	glBufferData( GL_ELEMENT_ARRAY_BUFFER, _indices.size() * sizeof( UINT32 ), &_indices.front(), GL_STATIC_DRAW );
	glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, 0 );
	
	for ( vector<Group>::iterator it = _groups.begin() ; it != _groups.end() ; ++ it )
	{
		if ( it->material ) it->material->Upload();
	}
	
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// Character - erase from graphics hardware

void Character::Erase()
{
	if ( 0 != _ibuf ) glDeleteBuffers( 1, &_ibuf );
	if ( 0 != _vbuf ) glDeleteBuffers( 1, &_vbuf );
}

////////////////////////////////////////////////////////////////////////////////
// Character - calculate bone matrices

void Character::Calculate( const vec3* positions, const quaternion* orientations, matrix4x4* transforms )
{
	_bones[ 0 ].Calculate( this, positions, orientations, matrix4x4(), transforms );
}

////////////////////////////////////////////////////////////////////////////////
// Character::Bone - calculate bone matrices

void Character::Bone::Calculate( Character* character, const vec3* bonePosition, const quaternion* boneOrientation, const matrix4x4& parentTransform, matrix4x4* outputTransform )
{
	vec3 finalPosition( bonePosition[ id ] * character->GetScale() + position );
	quaternion finalOrientation( boneOrientation[ id ] * orientation );
	
	matrix4x4 boneTransform( parentTransform * matrix4x4( finalPosition ) * matrix4x4( finalOrientation ) );
	outputTransform[ id ] = boneTransform * transform;
	
	for ( vector<Bone*>::iterator it = children.begin() ; it != children.end() ; ++ it )
	{
		( *it )->Calculate( character, bonePosition, boneOrientation, boneTransform, outputTransform );
	}
}

////////////////////////////////////////////////////////////////////////////////
// Character - render

void Character::Render( const matrix4x4* transforms )
{
	if ( 0 == _vbuf ) Upload();
	
	glBindBuffer( GL_ARRAY_BUFFER, _vbuf );
	
	if ( _bones.size() <= 32 )
	{
		glVertexPointer( 3, GL_FLOAT, sizeof( Vertex ), (PSTR)NULL + offsetof( Vertex, position ) );
		glNormalPointer( GL_FLOAT, sizeof( Vertex ), (PSTR)NULL + offsetof( Vertex, normal ) );
		glTexCoordPointer( 2, GL_FLOAT, sizeof( Vertex ), (PSTR)NULL + offsetof( Vertex, material ) );
		glVertexAttribPointer( Shader::BONE_INDEX_ATTRIB, 4, GL_FLOAT, GL_FALSE, sizeof( Vertex ), (PSTR)NULL + offsetof( Vertex, bone ) );
		glVertexAttribPointer( Shader::BONE_WEIGHT_ATTRIB, 4, GL_FLOAT, GL_FALSE, sizeof( Vertex ), (PSTR)NULL + offsetof( Vertex, weight ) );
		
		glEnableVertexAttribArray( Shader::BONE_INDEX_ATTRIB );
		glEnableVertexAttribArray( Shader::BONE_WEIGHT_ATTRIB );
		
		Shader* shader = Application::GetInstance()->GetRenderer()->GetSkeletonShader();
		shader->Begin();
		shader->SetBoneMatrices( transforms, (int)_bones.size() );
	}
	else
	{
        const Vertex* in	= &_vertices.front();
		CPUVertex* out		= &_skinned.front();
		
		for ( size_t count = _vertices.size() ; count -- ; in ++, out ++ )
		{
			const matrix4x4 m0( transforms[ (int)floor( in->bone[ 0 ] + 0.5f ) ] );
			const matrix4x4 m1( transforms[ (int)floor( in->bone[ 1 ] + 0.5f ) ] );
			const matrix4x4 m2( transforms[ (int)floor( in->bone[ 2 ] + 0.5f ) ] );
			const matrix4x4 m3( transforms[ (int)floor( in->bone[ 3 ] + 0.5f ) ] );
			
			out->material	= in->material;
			out->position	= in->weight[ 0 ] * ( m0 * in->position )
							+ in->weight[ 1 ] * ( m1 * in->position )
							+ in->weight[ 2 ] * ( m2 * in->position )
							+ in->weight[ 3 ] * ( m3 * in->position );
			out->normal		= in->weight[ 0 ] * m0.Rotate( in->normal )
							+ in->weight[ 1 ] * m1.Rotate( in->normal )
							+ in->weight[ 2 ] * m2.Rotate( in->normal )
							+ in->weight[ 3 ] * m3.Rotate( in->normal );
		}
		
		glBufferData( GL_ARRAY_BUFFER, _skinned.size() * sizeof( CPUVertex ), &_skinned.front(), GL_STREAM_DRAW );
		
		glVertexPointer( 3, GL_FLOAT, sizeof( CPUVertex ), (PSTR)NULL + offsetof( CPUVertex, position ) );
		glNormalPointer( GL_FLOAT, sizeof( CPUVertex ), (PSTR)NULL + offsetof( CPUVertex, normal ) );
		glTexCoordPointer( 2, GL_FLOAT, sizeof( CPUVertex ), (PSTR)NULL + offsetof( CPUVertex, material ) );
	}
	
	glEnable( GL_VERTEX_ARRAY );
	glEnable( GL_NORMAL_ARRAY );
	glEnable( GL_TEXTURE_COORD_ARRAY );
	
	glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, _ibuf );
	
	scalar v[ 4 ] = { 1, 1, 1, 1 };
	glLightfv( GL_LIGHT0, GL_DIFFUSE, v );
	
	for ( vector<Group>::iterator it = _groups.begin() ; it != _groups.end() ; ++ it )
	{
		it->material->Begin();
		glDrawElements( GL_TRIANGLES, it->indexCount, GL_UNSIGNED_INT, (int*)NULL + it->firstIndex );
		it->material->End();
	}
	
	glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, 0 );
	glBindBuffer( GL_ARRAY_BUFFER, 0 );
	
	glDisable( GL_TEXTURE_COORD_ARRAY );
	glDisable( GL_NORMAL_ARRAY );
	glDisable( GL_VERTEX_ARRAY );
	
	if ( _bones.size() <= 32 )
	{
		glUseProgram( 0 );
		glDisableVertexAttribArray( Shader::BONE_INDEX_ATTRIB );
		glDisableVertexAttribArray( Shader::BONE_WEIGHT_ATTRIB );
	}
}
