//
// Character.h
// Copyright Michael J Stokes, 2008.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#pragma once

#include "Renderer.h"

class Material;


class Character : public Resource
{
// Construction
public:
	Character( PCSTR name );
	virtual ~Character();
	
// Inlines
public:
	inline int FindBone( PCSTR name ) const
	{
		for ( int i = 0 ; i < (int)_bones.size() ; i ++ )
		{
			if ( _bones[ i ].name == name ) return i;
		}
		
		return -1;
	}
	
	inline scalar GetScale() const
	{
		return _scale;
	}
	
// Operations
public:
	virtual bool		Load();
	virtual bool		Upload();
	virtual void		Erase();
public:
	void				Calculate( const vec3* positions, const quaternion* orientations, matrix4x4* transforms );
	virtual void		Render( const matrix4x4* transforms );
	
// Vertex
public:
	struct Vertex
	{
		vec3		position;
		vec3		normal;
		vec2		material;
		scalar		bone[ 4 ];
		scalar		weight[ 4 ];
	};
	
// CPUVertex
public:
	struct CPUVertex
	{
		vec3		position;
		vec3		normal;
		vec2		material;
	};
	
// Bone
public:
	struct Bone
	{
		UINT			id;
		string			name;
		Bone*			parent;
		string			parentName;
		vector<Bone*>	children;
		vec3			position;
		quaternion		orientation;
		matrix4x4		transform;
		
		void			Prepare( const matrix4x4& parentTransform );
		void			Calculate( Character* character, const vec3* bonePosition, const quaternion* boneOrientation, const matrix4x4& parentTransform, matrix4x4* outputTransform );
	};
	
// Group
protected:
	struct Group
	{
		Material*	material;
		UINT32		firstIndex;
		UINT32		indexCount;
	};
	
// Inlines
public:
	inline UINT GetBoneCount() const
	{
		return (UINT)_bones.size();
	}
	
	inline const Bone & GetBone( UINT index ) const
	{
		return _bones[ index ];
	}
	
// Implementation
protected:
	bool			LoadSkeleton();
	bool			LoadMesh();
	bool			CompileMesh();
	
// Data
public:
	struct
	{
		int			Neck;
		int			Head;
		int			Jaw;
	} Bones;
protected:
	vector<Bone>		_bones;
	vector<Vertex>		_vertices;
	vector<CPUVertex>	_skinned;
	vector<UINT32>		_indices;
	vector<Group>		_groups;
	scalar				_scale;
	GLuint				_vbuf;
	GLuint				_ibuf;
};
