//
// Environment.cpp
// Copyright Michael J Stokes, 2008.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "Application.h"
#include "Environment.h"
#include "EnvironmentMesh.h"
#include "Renderer.h"
#include "Mesh.h"


////////////////////////////////////////////////////////////////////////////////
// Environment - construction

Environment::Environment()
{
}

////////////////////////////////////////////////////////////////////////////////
// Environment - destruction

Environment::~Environment()
{
	Empty();
}

////////////////////////////////////////////////////////////////////////////////
// Environment - empty

void Environment::Empty()
{
	while ( ! _meshes.empty() )
	{
		delete _meshes.front();
		_meshes.pop_front();
	}
}

////////////////////////////////////////////////////////////////////////////////
// Environment - load

bool Environment::Load()
{
	cout << "Loading environment configuration file '" << Application::GetInstance()->GetPath() << "/environment/environment.xml'..." << endl;
	
	XMLElement* root = XMLElement::FromFile( ( Application::GetInstance()->GetPath() + "/environment/environment.xml" ).c_str() );
	XMLAutoDelete ad( root );
	
	if ( ! root->IsNamed( "environment" ) )
	{
		cerr << "Unable to open environment configuration file '" << Application::GetInstance()->GetPath() << "/environment/environment.xml'" << endl;
		return false;
	}
	
	if ( XMLElement* meshes = root->GetElement( "meshes" ) )
	{
		for ( XMLElementIterator it = meshes->GetElements() ; it ; ++ it )
		{
			if ( it->IsNamed( "mesh" ) )
			{
				EnvironmentMesh* mesh = new EnvironmentMesh();
				
				if ( mesh->Load( *it ) )
				{
					if ( EnvironmentMesh::DEPTH == ( mesh->GetFlags() & ( EnvironmentMesh::RENDER | EnvironmentMesh::DEPTH ) ) )
					{
						_meshes.push_back( mesh );
					}
					else
					{
						_meshes.push_front( mesh );
					}
				}
				else
				{
					delete mesh;
					return false;
				}
			}
		}
	}
	
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// Environment - save

bool Environment::Save()
{
	cout << "Saving environment configuration file '" << Application::GetInstance()->GetPath() << "/environment/environment.xml'..." << endl;
	
	XMLElement root( "environment" );
	root.AddAttribute( "xmlns", "urn:no.ntnu.idi.vgs:environment:1.0" );
	
	XMLElement* meshes = root.AddElement( "meshes" );
	
	for ( list<EnvironmentMesh*>::iterator it = _meshes.begin() ; it != _meshes.end() ; ++ it )
	{
		meshes->Add( ( *it )->Save() );
	}
	
	return root.ToFile( ( Application::GetInstance()->GetPath() + "/environment/environment.xml" ).c_str() );
}

////////////////////////////////////////////////////////////////////////////////
// Environment - render

void Environment::Render()
{
	for ( list<EnvironmentMesh*>::iterator it = _meshes.begin() ; it != _meshes.end() ; ++ it )
	{
		( *it )->Render();
	}
}

////////////////////////////////////////////////////////////////////////////////
// Environment - add a mesh

void Environment::AddMesh( PCSTR source, int flags )
{
	XMLElement xml( "mesh" );
	
	xml.AddAttribute( "source", source );
	if ( flags & EnvironmentMesh::CLIP ) xml.AddElement( "clip" );
	if ( flags & EnvironmentMesh::RENDER ) xml.AddElement( "render" );
	if ( flags & EnvironmentMesh::DEPTH ) xml.AddElement( "depth" );
	
	EnvironmentMesh* mesh = new EnvironmentMesh();
	
	if ( mesh->Load( &xml ) )
	{
		if ( EnvironmentMesh::DEPTH == ( mesh->GetFlags() & ( EnvironmentMesh::RENDER | EnvironmentMesh::DEPTH ) ) )
		{
			_meshes.push_back( mesh );
		}
		else
		{
			_meshes.push_front( mesh );
		}
	}
	else
	{
		delete mesh;
#ifdef IS_EDITOR
		wxMessageBox( T("Unable to load environment mesh.") );
#endif
	}
}
