//
// Feature.cpp
// Copyright Michael J Stokes, 2008.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "Application.h"
#include "Renderer.h"
#include "Physics.h"
#include "Features.h"
#include "Feature.h"
#include "Actors.h"
#include "Mesh.h"
#include "Ego.h"
#include "FileSearch.h"


////////////////////////////////////////////////////////////////////////////////
// Feature - construction

Feature::Feature( PCSTR name )
{
	_name		= name;
	_script		= NULL;
	_presented	= 0;
	_suggested	= 0;
}

////////////////////////////////////////////////////////////////////////////////
// Feature - destruction

Feature::~Feature()
{
	_script->Delete();
}

////////////////////////////////////////////////////////////////////////////////
// Feature - load

bool Feature::Load()
{
	string path( Application::GetInstance()->GetPath() + "/features/" + _name + "/feature.xml" );
	XMLElement* root = XMLElement::FromFile( path.c_str() );
	XMLAutoDelete ad( root );
	
	if ( ! root->IsNamed( "feature" ) )
	{
		cerr << "Unable to load feature '" << path << "'." << endl;
		return false;
	}
	
	if ( XMLElement* locations = root->GetElement( "locations" ) )
	{
		for ( XMLElementIterator it = locations->GetElements() ; it ; ++ it )
		{
			vec3 position;
			
			if ( XMLElement* pos = it->GetElement( "position" ) )
			{
				position.x = pos->GetAttributeFloat( "x" );
				position.y = pos->GetAttributeFloat( "y" );
				position.z = pos->GetAttributeFloat( "z" );
			}
			
			if ( 0 == position.y )
			{
				position.y = Application::GetInstance()->GetPhysics()->Drop( position );
			}
			
			if ( it->IsNamed( "activation-point" ) )
			{
				ActivationPoint ap;
				
				ap.position	= position;
				ap.inner	= 1.5f;
				ap.outer	= 5;
				
				if ( XMLElement* radius = it->GetElement( "radius" ) )
				{
					ap.inner = radius->GetAttributeFloat( "inner", ap.inner );
					ap.outer = radius->GetAttributeFloat( "outer", ap.outer );
				}
				
				_activations.push_back( ap );
			}
			else if ( it->IsNamed( "location-point" ) )
			{
				_locations.push_back( position );
			}
			else if ( it->IsNamed( "guide-point" ) )
			{
				_guides.push_back( position );
			}
		}
	}
	
	if ( _activations.empty() )
	{
		cerr << "Feature '" << _name << "' has no activation points." << endl;
		return false;
	}
	
	if ( _script = root->GetElement( "script" ) )
	{
		_script->Detach();
	}
	
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// Feature - save

bool Feature::Save()
{
#ifdef IS_EDITOR
	XMLElement root( "feature" );
	root.AddAttribute( "xmlns", "urn:no.ntnu.idi.vgs:feature:1.1" );
	
	XMLElement* locations = root.AddElement( "locations" );
	
	for ( vector<ActivationPoint>::iterator it = _activations.begin() ; it != _activations.end() ; ++ it )
	{
		XMLElement* location	= locations->AddElement( "activation-point" );
		XMLElement* position	= location->AddElement( "position" );
		XMLElement* radius		= location->AddElement( "radius" );
		
		position->AddAttributeFloat( "x", it->position.x );
		position->AddAttributeFloat( "y", it->position.y );
		position->AddAttributeFloat( "z", it->position.z );
		
		radius->AddAttributeFloat( "inner", it->inner );
		radius->AddAttributeFloat( "outer", it->outer );
	}
	
	for ( vector<vec3>::iterator it = _locations.begin() ; it != _locations.end() ; ++ it )
	{
		XMLElement* location	= locations->AddElement( "location-point" );
		XMLElement* position	= location->AddElement( "position" );
		
		position->AddAttributeFloat( "x", it->x );
		position->AddAttributeFloat( "y", it->y );
		position->AddAttributeFloat( "z", it->z );
	}
	
	for ( vector<vec3>::iterator it = _guides.begin() ; it != _guides.end() ; ++ it )
	{
		XMLElement* location	= locations->AddElement( "guide-point" );
		XMLElement* position	= location->AddElement( "position" );
		
		position->AddAttributeFloat( "x", it->x );
		position->AddAttributeFloat( "y", it->y );
		position->AddAttributeFloat( "z", it->z );
	}
	
	if ( NULL != _script )
	{
		root.Add( _script->Clone() );
	}
	
	string path( Application::GetInstance()->GetPath() + "/features/" + _name );
	
	if ( ! wxDirExists( path.c_str() ) ) wxMkdir( path.c_str() );
	
	return root.ToFile( ( path + "/feature.xml" ).c_str() );
#else
	return false;
#endif
}

////////////////////////////////////////////////////////////////////////////////
// Feature - rename

bool Feature::Rename( PCSTR name )
{
#ifdef IS_EDITOR
	if ( _name == name ) return true;
	
	string from( Application::GetInstance()->GetPath() + "/features/" + _name );
	string to( Application::GetInstance()->GetPath() + "/features/" + string( name ) );
	
	if ( wxRenameFile( from, to, false ) )
	{
		_name = name;
		return true;
	}
#endif
	
	return false;
}

////////////////////////////////////////////////////////////////////////////////
// Feature - delete

bool Feature::Delete( PCSTR path )
{
#ifdef IS_EDITOR
	if ( NULL == path )
	{
		string root( Application::GetInstance()->GetPath() + "/features/" + _name );
		return Delete( root.c_str() );
	}
	
	FileSearch fs( path );
	string name;
	bool folder;
	
	while ( fs.GetNext( name, &folder ) )
	{
		string sub( path );
		sub += PATH_SLASH_STR;
		sub += name;
		
		if ( folder )
		{
			Delete( sub.c_str() );
		}
		else
		{
			if ( wxYES == wxMessageBox( "Delete file \"" + sub + "\"?", "Delete Feature", wxYES_NO ) ) wxRemoveFile( sub.c_str() );
		}
	}
	
	return wxRmdir( path );
#else
	return false;
#endif
}

////////////////////////////////////////////////////////////////////////////////
// Feature - register positions

void Feature::RegisterPositions()
{
	Actors* actors = Application::GetInstance()->GetActors();
	char num[ 32 ];
	int index = 1;
	
	for ( vector<ActivationPoint>::iterator it = _activations.begin() ; it != _activations.end() ; ++ it, index ++ )
	{
		string name( _name );
		sprintf( num, "_a%i", index );
		name += num;
		actors->RegisterPosition( name.c_str(), it->position );
	}
	
	index = 1;
	
	for ( vector<vec3>::iterator it = _locations.begin() ; it != _locations.end() ; ++ it, index ++ )
	{
		string name( _name );
		sprintf( num, "_l%i", index );
		name += num;
		actors->RegisterPosition( name.c_str(), *it );
	}
	
	index = 1;
	
	for ( vector<vec3>::iterator it = _guides.begin() ; it != _guides.end() ; ++ it, index ++ )
	{
		string name( _name );
		sprintf( num, "_g%i", index );
		name += num;
		actors->RegisterPosition( name.c_str(), *it );
	}
}

////////////////////////////////////////////////////////////////////////////////
// Feature - check if the position is within an activation point

bool Feature::IsInsideActivationPoint( const vec3& position )
{
	for ( vector<ActivationPoint>::iterator it = _activations.begin() ; it != _activations.end() ; ++ it )
	{
		vec3 toward( it->position - position );
		toward.y = 0;
		if ( toward.Length() <= it->inner ) return true;
	}
	
	return false;
}

////////////////////////////////////////////////////////////////////////////////
// Feature - check if the position is outside all activation points

bool Feature::IsOutsideActivationPoints( const vec3& position )
{
	for ( vector<ActivationPoint>::iterator it = _activations.begin() ; it != _activations.end() ; ++ it )
	{
		vec3 toward( it->position - position );
		toward.y = 0;
		if ( toward.Length() <= it->outer ) return false;
	}
	
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// Feature - get the closest guide point

string Feature::GetClosestGuidePoint( const vec3& position )
{
	scalar bestDistance = 9999999;
	int index = 1, bestPoint = 0;
	
	for ( vector<vec3>::iterator it = _guides.begin() ; it != _guides.end() ; ++ it, index ++ )
	{
		vec3 toward( *it - position );
		toward.y = 0;
		scalar distance = toward.Length();
		
		if ( distance < bestDistance )
		{
			bestPoint		= index;
			bestDistance	= distance;
		}
	}
	
	string name( _name );
	
	if ( bestPoint > 0 )
	{
		char num[ 16 ];
		sprintf( num, "_g%i", bestPoint );
		name += num;
	}
	else
	{
		name += "_a1";
	}
	
	return name;
}

////////////////////////////////////////////////////////////////////////////////
// Feature - get the closest location point

string Feature::GetClosestLocationPoint( const vec3& position )
{
	scalar bestDistance = 9999999;
	int index = 1, bestPoint = 0;
	
	for ( vector<vec3>::iterator it = _locations.begin() ; it != _locations.end() ; ++ it, index ++ )
	{
		vec3 toward( *it - position );
		toward.y = 0;
		scalar distance = toward.Length();
		
		if ( distance < bestDistance )
		{
			bestPoint		= index;
			bestDistance	= distance;
		}
	}
	
	string name( _name );
	
	if ( bestPoint > 0 )
	{
		char num[ 16 ];
		sprintf( num, "_l%i", bestPoint );
		name += num;
	}
	else
	{
		name += "_a1";
	}
	
	return name;
}

////////////////////////////////////////////////////////////////////////////////
// Feature - mark the feature as presented

XMLElement* Feature::Present()
{
	_presented = Application::GetInstance()->GetTime();
	return _script ? _script->Clone() : NULL;
}

////////////////////////////////////////////////////////////////////////////////
// Feature - mark the feature as suggested

void Feature::Suggest()
{
	_suggested = Application::GetInstance()->GetTime();
}

////////////////////////////////////////////////////////////////////////////////
// Feature - add a point

int Feature::AddPoint( int type, const vec3& position )
{
	switch ( type )
	{
	case 0:
		{
			ActivationPoint ap;
			ap.position	= position;
			ap.inner	= 1;
			ap.outer	= 3;
			_activations.push_back( ap );
			Save();
			return (int)_activations.size() - 1;
		}
	case 1:
		_locations.push_back( position );
		Save();
		return (int)_locations.size() - 1;
	case 2:
		_guides.push_back( position );
		Save();
		return (int)_guides.size() - 1;
	}
	
	return -1;
}

////////////////////////////////////////////////////////////////////////////////
// Feature - remove a point

bool Feature::RemovePoint( int type, int index )
{
	if ( index < 0 ) return true;
	
	switch ( type )
	{
	case 0:
		if ( index >= (int)_activations.size() ) return true;
		if ( 1 == _activations.size() ) return false;
		_activations.erase( _activations.begin() + index );
		Save();
		return true;
	case 1:
		if ( index >= (int)_locations.size() ) return true;
		_locations.erase( _locations.begin() + index );
		Save();
		return true;
	case 2:
		if ( index >= (int)_guides.size() ) return true;
		_guides.erase( _guides.begin() + index );
		Save();
		return true;
	}
	
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// Feature - move a point

bool Feature::MovePoint( int type, int index, const vec3& position )
{
	if ( index < 0 ) return false;
	
	switch ( type )
	{
	case 0:
		if ( index >= (int)_activations.size() ) return false;
		_activations[ index ].position = position;
		return true;
	case 1:
		if ( index >= (int)_locations.size() ) return false;
		_locations[ index ] = position;
		return true;
	case 2:
		if ( index >= (int)_guides.size() ) return false;
		_guides[ index ] = position;
		return true;
	}
	
	return false;
}

////////////////////////////////////////////////////////////////////////////////
// Feature - check that a point actually exists

bool Feature::CheckPoint( int type, int index )
{
	if ( index < 0 ) return false;
	
	switch ( type )
	{
	case 0:
		return ( index < (int)_activations.size() );
	case 1:
		return ( index < (int)_locations.size() );
	case 2:
		return ( index < (int)_guides.size() );
	}
	
	return false;
}

////////////////////////////////////////////////////////////////////////////////
// Feature - get the radii of an activation point

bool Feature::GetActivationRadii( int index, scalar& inner, scalar& outer, vec3& position )
{
	if ( ( index >= 0 ) && ( index < (int)_activations.size() ) )
	{
		inner = _activations[ index ].inner;
		outer = _activations[ index ].outer;
		position = _activations[ index ].position;
		return true;
	}
	
	return false;
}

////////////////////////////////////////////////////////////////////////////////
// Feature - set the radii of an activation point

bool Feature::SetActivationRadii( int index, scalar inner, scalar outer )
{
	if ( ( index >= 0 ) && ( index < (int)_activations.size() ) )
	{
		_activations[ index ].inner = inner;
		_activations[ index ].outer = outer;
		return true;
	}
	
	return false;
}

////////////////////////////////////////////////////////////////////////////////
// Feature - set the feature script

void Feature::SetScript( XMLElement* script )
{
	_script->Delete();
	_script = script;
	Save();
}

////////////////////////////////////////////////////////////////////////////////
// Feature - render

void Feature::Render()
{
	Application* application = Application::GetInstance();
	Features::Selection& sel = application->GetFeatures()->GetSelection();
	
	if ( ( NULL == application->GetEgo() ) && ( this == sel.feature ) )
	{
		UINT index = 0;
		
		for ( vector<ActivationPoint>::const_iterator it = _activations.begin() ; it != _activations.end() ; ++ it, index ++ )
		{
			RenderFloatingIcon( *it, ( sel.type == 0 ) && ( sel.index == index ) );
			RenderBox( it->position, vec3( 0, 1, 0 ), ( sel.type == 0 ) && ( sel.index == index ) );
		}
		
		index = 0;
		
		for ( vector<vec3>::const_iterator it = _locations.begin() ; it != _locations.end() ; ++ it, ++ index )
		{
			RenderBox( *it, vec3( 0, 0, 1 ), ( sel.type == 1 ) && ( sel.index == index ) );
		}
		
		index = 0;
		
		for ( vector<vec3>::const_iterator it = _guides.begin() ; it != _guides.end() ; ++ it, ++ index )
		{
			RenderBox( *it, vec3( 1, 0, 0 ), ( sel.type == 2 ) && ( sel.index == index ) );
		}
		
		glDisable( GL_LIGHTING );
		glColor3f( 0.2f, 0.2f, 0.2f );
		glBegin( GL_LINES );
		
		for ( vector<ActivationPoint>::const_iterator it1 = _activations.begin() ; it1 != _activations.end() ; ++ it1 )
		{
			for ( vector<ActivationPoint>::const_iterator it2 = _activations.begin() ; it2 != _activations.end() ; ++ it2 )
			{
				if ( it1 != it2 )
				{
					glVertex3fv( it1->position );
					glVertex3fv( it2->position );
				}
			}
		}
		
		if ( ! _activations.empty() )
		{
			const ActivationPoint & ap = _activations.front();
			
			for ( vector<vec3>::const_iterator it = _locations.begin() ; it != _locations.end() ; ++ it )
			{
				glVertex3fv( ap.position );
				glVertex3fv( *it );
			}
			
			for ( vector<vec3>::const_iterator it = _guides.begin() ; it != _guides.end() ; ++ it )
			{
				glVertex3fv( ap.position );
				glVertex3fv( *it );
			}
		}
		
		glEnd();
		glColor3f( 1, 1, 1 );
		glEnable( GL_LIGHTING );
	}
	else if ( ! _activations.empty() )
	{
		const ActivationPoint & ap = _activations.front();
		RenderFloatingIcon( ap, false );
		
		if ( NULL == application->GetEgo() )
		{
			RenderBox( ap.position, vec3( 0, 1, 0 ), false );
		}
	}
}

///////////////////////////////////////////////////////////////////////////////
// Feature - render a box

void Feature::RenderBox( const vec3& center, const vec3& colour, bool selected, scalar size )
{
	glPushMatrix();
	glTranslatef( center.x, center.y, center.z );
	
	if ( selected )
	{
		size *= 2.0f;
		glRotatef( -fmod( Application::GetInstance()->GetTime() * 100, 360 ), 0, 1, 0 );
	}
	
	scalar v[ 4 ] = { colour.r, colour.g, colour.b, 1 };
	glLightfv( GL_LIGHT0, GL_DIFFUSE, v );
	
	glBegin( GL_QUADS );
	
	glNormal3f( 1, 0, 0 );
	glVertex3f( size, size, -size );
	glVertex3f( size, size, size );
	glVertex3f( size, -size, size );
	glVertex3f( size, -size, -size );
	glNormal3f( -1, 0, 0 );
	glVertex3f( -size, size, size );
	glVertex3f( -size, size, -size );
	glVertex3f( -size, -size, -size );
	glVertex3f( -size, -size, size );
	
	glNormal3f( 0, 1, 0 );
	glVertex3f( -size, size, size );
	glVertex3f( size, size, size );
	glVertex3f( size, size, -size );
	glVertex3f( -size, size, -size );
	glNormal3f( 0, -1, 0 );
	glVertex3f( -size, -size, -size );
	glVertex3f( size, -size, -size );
	glVertex3f( size, -size, size );
	glVertex3f( -size, -size, size );
	
	glNormal3f( 0, 0, -1 );
	glVertex3f( -size, size, -size );
	glVertex3f( size, size, -size );
	glVertex3f( size, -size, -size );
	glVertex3f( -size, -size, -size );
	glNormal3f( 0, 0, 1 );
	glVertex3f( size, size, size );
	glVertex3f( -size, size, size );
	glVertex3f( -size, -size, size );
	glVertex3f( size, -size, size );
	
	glEnd();
	glPopMatrix();
}

///////////////////////////////////////////////////////////////////////////////
// Feature - render a floating icon

void Feature::RenderFloatingIcon( const ActivationPoint& ap, bool selected )
{
	Application* application = Application::GetInstance();
	scalar time = application->GetTime();
	
	glPushMatrix();
	glTranslatef( ap.position.x, ap.position.y, ap.position.z );
	
	scalar effect = 1;
	
	if ( Ego* ego = application->GetEgo() )
	{
		effect = ( ego->GetPosition() - ap.position ).Length();
	}
	
	effect = ( effect < ap.outer ) ? ( effect / ap.outer ) : 1;
	
	glTranslatef( 0, sinf( time * 6 ) * 0.1f * effect, 0 );
	glRotatef( fmod( time * 100, 360 ), 0, 1, 0 );
	glScalef( 0.5f, 0.5f, 0.5f );
	
	scalar v[ 4 ] = { 1, 1, 1, effect * 0.5f };
	
	glLightfv( GL_LIGHT0, GL_DIFFUSE, v );
	if ( ! selected ) glEnable( GL_BLEND );
	
	application->GetRenderer()->GetInfoMesh()->Render();
	
	glDisable( GL_BLEND );
	glPopMatrix();
}

////////////////////////////////////////////////////////////////////////////////
// Feature - hit test

bool Feature::HitTest( const vec3& origin, const vec3& direction, scalar& distance, int& type, int& index, bool selected ) const
{
	type	= 0;
	index	= 0;
	
	for ( vector<ActivationPoint>::const_iterator it = _activations.begin() ; it != _activations.end() ; ++ it, ++ index )
	{
		if ( RayTest( origin, direction, distance, it->position ) ) return true;
		if ( ! selected ) return false;
	}
	
	type	= 1;
	index	= 0;
	
	for ( vector<vec3>::const_iterator it = _locations.begin() ; it != _locations.end() ; ++ it, ++ index )
	{
		if ( RayTest( origin, direction, distance, *it ) ) return true;
	}
	
	type	= 2;
	index	= 0;
	
	for ( vector<vec3>::const_iterator it = _guides.begin() ; it != _guides.end() ; ++ it, ++ index )
	{
		if ( RayTest( origin, direction, distance, *it ) ) return true;
	}
	
	return false;
}

///////////////////////////////////////////////////////////////////////////////
// Feature - do a ray test with a box

bool Feature::RayTest( const vec3& origin, const vec3& direction, scalar& distance, const vec3& center, scalar size )
{
	plane3 side[ 6 ];
	
	side[ 0 ].Set( vec3( 1, 0, 0 ), size );
	side[ 1 ].Set( vec3( -1, 0, 0 ), size );
	side[ 2 ].Set( vec3( 0, 1, 0 ), size );
	side[ 3 ].Set( vec3( 0, -1, 0 ), size );
	side[ 4 ].Set( vec3( 0, 0, 1 ), size );
	side[ 5 ].Set( vec3( 0, 0, -1 ), size );
	
	vec3 orig( origin - center );
	
	bool front = false;
	
	for ( int i = 0 ; i < 6 ; i ++ )
	{
		if ( SIDE_FRONT == side[ i ].Side( orig ) )
		{
			front = true;
			break;
		}
	}
	
	if ( ! front ) return false;
	
	vec3 start( orig );
	vec3 end( orig + direction * ( ( distance < 16384 ) ? distance : 16384 ) );
	
	for ( int i = 0 ; i < 6 ; i ++ )
	{
		scalar dist1 = side[ i ].Distance( start );
		scalar dist2 = side[ i ].Distance( end );
		
		if ( ( dist1 >= 0 ) && ( dist2 >= 0 ) )
		{
			return false;
		}
		else if ( ( dist1 > 0 ) || ( dist2 > 0 ) )
		{
			scalar frac = dist1 / ( dist1 - dist2 );
			
			if ( dist1 > 0 )
			{
				start = start + ( end - start ) * frac;
			}
			else
			{
				end = start + ( end - start ) * frac;
			}
		}
	}
	
	scalar dist = direction.DotProduct( start - orig );
	
	if ( ( dist >= 0 ) && ( dist < distance ) )
	{
		distance = dist;
		return true;
	}
	
	return false;
}
