//
// Features.cpp
// Copyright Michael J Stokes, 2008.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "Application.h"
#include "Features.h"
#include "Feature.h"
#include "FileSearch.h"
#include "Actors.h"


////////////////////////////////////////////////////////////////////////////////
// Features - construction

Features::Features()
{
}

////////////////////////////////////////////////////////////////////////////////
// Features - destruction

Features::~Features()
{
	while ( ! _features.empty() )
	{
		delete _features.front();
		_features.pop_front();
	}
}

////////////////////////////////////////////////////////////////////////////////
// Features - load

bool Features::Load()
{
	FileSearch search( ( Application::GetInstance()->GetPath() + "/features" ).c_str() );
	string name;
	bool dir;
	
	while( search.GetNext( name, &dir ) )
	{
		if ( ! dir ) continue;
		
		Feature* feature = new Feature( name.c_str() );
		
		if ( feature->Load() )
		{
			_features.push_back( feature );
		}
		else
		{
			delete feature;
			return false;
		}
	}
	
	cout << "Loaded " << (int)_features.size() << " features." << endl;
	
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// Features - render

void Features::Render()
{
	for ( list<Feature*>::iterator it = _features.begin() ; it != _features.end() ; ++ it )
	{
		( *it )->Render();
	}
}

////////////////////////////////////////////////////////////////////////////////
// Features - register positions

void Features::RegisterPositions()
{
	for ( list<Feature*>::iterator it = _features.begin() ; it != _features.end() ; ++ it )
	{
		( *it )->RegisterPositions();
	}
}

////////////////////////////////////////////////////////////////////////////////
// Features - get a feature that is activated by the given position

Feature* Features::GetActivatedFeature( const vec3& position )
{
	for ( list<Feature*>::iterator it = _features.begin() ; it != _features.end() ; ++ it )
	{
		Feature* feature = *it;
		if ( feature->IsInsideActivationPoint( position ) ) return feature;
	}
	
	return NULL;
}

////////////////////////////////////////////////////////////////////////////////
// Features - select the best feature to present

Feature* Features::SelectFeature( const vec3& position )
{
	Actors* actors = Application::GetInstance()->GetActors();
	scalar now = Application::GetInstance()->GetTime();
	
	scalar bestDistance		= 999999;
	Feature* bestFeature	= NULL;
	
	for ( list<Feature*>::iterator it = _features.begin() ; it != _features.end() ; ++ it )
	{
		Feature* feature = *it;
		
		if ( 0 != feature->GetPresented() ) continue;
		if ( ( 0 != feature->GetSuggested() ) && ( ( now - feature->GetSuggested() ) < 60 ) ) continue;
		
		vec3 toward( actors->GetPosition( feature->GetClosestGuidePoint( position ).c_str() ) - position );
		toward.y = 0;
		scalar distance = toward.Length();
		
		if ( distance < bestDistance )
		{
			bestDistance	= distance;
			bestFeature		= feature;
		}
	}
	
	return bestFeature;
}

////////////////////////////////////////////////////////////////////////////////
// Features - hit test the features returning a point selection

Features::Selection Features::HitTest( const vec3& origin, const vec3& ray, scalar& distance )
{
	Selection sel;
	
	for ( list<Feature*>::iterator it = _features.begin() ; it != _features.end() ; ++ it )
	{
		Feature* feature = *it;
		scalar dist = distance;
		int type, index;
		
		if ( feature->HitTest( origin, ray, dist, type, index, ( _selection.feature == feature ) ) )
		{
			if ( dist < distance )
			{
				sel.feature	= feature;
				sel.type	= type;
				sel.index	= index;
				distance	= dist;
			}
		}
	}
	
	return sel;
}

////////////////////////////////////////////////////////////////////////////////
// Features - add a new feature

Feature* Features::AddFeature()
{
	char name[ 9 ];
	name[ 8 ] = 0;
	
	for ( int i = 0 ; i < 8 ; i ++ )
	{
		switch ( rand() % 3 )
		{
		case 0:
			name[ i ] = 'A' + ( rand() % 26 );
			break;
		case 1:
			name[ i ] = 'a' + ( rand() % 26 );
			break;
		case 2:
			name[ i ] = '0' + ( rand() % 10 );
			break;
		}
	}
	
	_selection.feature	= new Feature( name );
	_selection.type		= 0;
	_selection.index	= 0;
	
	_features.push_back( _selection.feature );
	_selection.feature->Save();
	
	return _selection.feature;
}

////////////////////////////////////////////////////////////////////////////////
// Features - delete a feature

void Features::DeleteFeature( Feature* feature )
{
	_features.remove( feature );
	feature->Delete();
	delete feature;
	_selection.feature = NULL;
}
