
// Guide.cpp
// Copyright Michael J Stokes, 2008.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "Application.h"
#include "Actors.h"
#include "Guide.h"
#include "Ego.h"
#include "Features.h"
#include "Feature.h"


////////////////////////////////////////////////////////////////////////////////
// Guide - construction

Guide::Guide()
{
	_state		= GS_WAITING_TO_FOLLOW;
	_transition	= 0;
	_target		= NULL;
	_ignore		= NULL;
	_script		= NULL;
}

////////////////////////////////////////////////////////////////////////////////
// Guide - destruction

Guide::~Guide()
{
	XMLElement destroy( "vgs", "destroy" );
	Send( &destroy );
}

////////////////////////////////////////////////////////////////////////////////
// Guide - load

bool Guide::Load( XMLElement* root )
{
	XMLElement create( "vgs", "create" );
	
	if ( XMLElement* guide = root->GetElement( "guide" ) )
	{
		if ( XMLElement* position = guide->GetElement( "position" ) )
		{
			create.Add( position->Clone() );
		}
		
		if ( XMLElement* orientation = guide->GetElement( "orientation" ) )
		{
			create.Add( orientation->Clone() );
		}
		
		if ( XMLElement* character = guide->GetElement( "character" ) )
		{
			create.Add( character->Clone() );
		}
	}
	
	Send( &create );
	StartWaitingToFollow();
	
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// Guide - update

void Guide::Update()
{
	// Make sure the actor system is up and running
	
	Actors* actors = Application::GetInstance()->GetActors();
	if ( NULL == actors ) return;
	if ( ! actors->IsConnected() ) return;
	
	// Query the position of the guide
	
	XMLElement positionQuery( "vgs", "position" );
	
	if ( XMLElement* position = Send( &positionQuery ) )
	{
		_position.x = position->GetAttributeFloat( "x" );
		_position.y = position->GetAttributeFloat( "y" );
		_position.z = position->GetAttributeFloat( "z" );
	}
	
	// Cache some often-used data about the guide's relationship with the ego
	
	_now		= Application::GetInstance()->GetTime();
	_ego		= Application::GetInstance()->GetEgo();
	_toward		= ( _ego->GetPosition() - _position );
	_toward.y	= 0;
	_distance	= _toward.Length();
	_toward		/= _distance;
	
	// Update the ignored feature
	
	if (  NULL != _ignore )
	{
		if ( _ignore->IsOutsideActivationPoints( _ego->GetPosition() ) ) _ignore = NULL;
	}
	
	// Switch based on the current state
	
	switch ( _state )
	{
	case GS_FOLLOWING:
		RunFollowing();
		break;
	case GS_WAITING_TO_FOLLOW:
		RunWaitingToFollow();
		break;
	case GS_LEADING:
		RunLeading();
		break;
	case GS_WAITING_TO_LEAD:
		RunWaitingToLeadOrPresent( true );
		break;
	case GS_WAITING_TO_PRESENT:
		RunWaitingToLeadOrPresent( false );
		break;
	case GS_SUGGESTING:
		RunSuggesting();
		break;
	case GS_WALKING_TO_PRESENT:
		RunWalkingToPresent();
		break;
	case GS_PRESENTING:
		RunPresenting();
		break;
	}
}

////////////////////////////////////////////////////////////////////////////////
// Guide - start - FOLLOWING

void Guide::StartFollowing()
{
	// Set the state
	
	SetState( GS_FOLLOWING );
	
	// Gaze at the ego with wandering
	
	XMLElement bml( "bml" );
	XMLElement* g = bml.AddElement( "gaze" );
	g->AddAttribute( "target", "ego" );
	g->AddAttribute( "wander", "45" );
	
	// Walk to the ego
	
	XMLElement* w = bml.AddElement( "vgs", "walk" );
	w->AddAttribute( "target", "ego" );
	w->AddAttribute( "proximity", "2" );
	
	// Send requests
	
	Send( &bml );
}

////////////////////////////////////////////////////////////////////////////////
// Guide - run - FOLLOWING

void Guide::RunFollowing()
{
	// Check if a feature has been activated
	
	if ( CheckActivation() ) return;
	
	// Check if the walk is still in progress
	
	XMLElement query( "vgs", "walking" );
	
	if ( XMLElement* result = Send( &query ) )
	{
		result->Delete();
		return;
	}
	
	// Once we reach the ego, transition to WAITING_TO_FOLLOW
	
	StartWaitingToFollow();
}

////////////////////////////////////////////////////////////////////////////////
// Guide - start - WAITING_TO_FOLLOW

void Guide::StartWaitingToFollow( scalar delay )
{
	// Set the state
	
	SetState( GS_WAITING_TO_FOLLOW );
	
	// Gaze at the ego with wandering
	
	XMLElement bml( "bml" );
	XMLElement* g = bml.AddElement( "gaze" );
	g->AddAttribute( "target", "ego" );
	g->AddAttribute( "wander", "45" );
	
	// Stand idle
	
	bml.AddElement( "body" )->AddAttribute( "posture", "LHandOnHip_Motex" );
	
	// Send requests
	
	Send( &bml );
	
	// Save the delay
	
	_delay = delay;
}

////////////////////////////////////////////////////////////////////////////////
// Guide - run - WAITING_TO_FOLLOW

void Guide::RunWaitingToFollow()
{
	// Check if a feature has been activated
	
	if ( CheckActivation() ) return;
	
	// Check if the ego has started moving away from us
	
	if ( _distance >= 4 )
	{
		// Transition to FOLLOWING
		
		StartFollowing();
		return;
	}
	
	// Detect inactivity
	
	if ( ( _now - _transition ) > _delay )
	{
		// Suggest a new target
		
		StartSuggesting();
	}
}

////////////////////////////////////////////////////////////////////////////////
// Guide - start - LEADING

void Guide::StartLeading()
{
	// Set the state
	
	SetState( GS_LEADING );
	
	// Walk toward the target feature's guide point
	
	XMLElement bml( "bml" );
	bml.AddElement( "gaze" )->AddAttribute( "target", _target->GetClosestLocationPoint( _position ).c_str() );
	bml.AddElement( "vgs", "walk" )->AddAttribute( "target", _target->GetClosestGuidePoint( _position ).c_str() );
	Send( &bml );
}

////////////////////////////////////////////////////////////////////////////////
// Guide - run - LEADING

void Guide::RunLeading()
{
	// Check if a feature has been activated
	
	if ( CheckActivation() ) return;
	
	// Check if the ego is lagging too far behind
	
	if ( _distance > 9 )
	{
		// Transition to WAITING_TO_LEAD
		
		StartWaitingToLeadOrPresent( true );
		return;
	}
	
	// Check if the walk is still in progress
	
	XMLElement query( "vgs", "walking" );
	
	if ( XMLElement* result = Send( &query ) )
	{
		result->Delete();
		return;
	}
	
	// Transition to WAITING_TO_PRESENT
	
	StartWaitingToLeadOrPresent( false );
}

////////////////////////////////////////////////////////////////////////////////
// Guide - start - WAITING_TO_LEAD and WAITING_TO_PRESENT

void Guide::StartWaitingToLeadOrPresent( bool lead )
{
	// Set the state
	
	SetState( lead ? GS_WAITING_TO_LEAD : GS_WAITING_TO_PRESENT );
	
	// Save the base ego position
	
	_base = _ego->GetPosition();	
	
	// Gaze at the ego
	
	XMLElement bml( "bml" );
	XMLElement* g = bml.AddElement( "gaze" );
	g->AddAttribute( "target", "ego" );
	g->AddAttribute( "wander", "45" );
	
	// Stand idle
	
	bml.AddElement( "body" )->AddAttribute( "posture", "HandsAtSide_Motex" );
	
	// Send requests
	
	Send( &bml );
}

////////////////////////////////////////////////////////////////////////////////
// Guide - run - WAITING_TO_LEAD and WAITING_TO_PRESENT

void Guide::RunWaitingToLeadOrPresent( bool lead )
{
	// Check if a feature has been activated
	
	if ( CheckActivation() ) return;
	
	// Check if the ego has reached us if we are still leading
	
	if ( lead && ( _distance < 5 ) )
	{
		// Transition to LEADING
		
		StartLeading();
		return;
	}
	
	// Check if the ego is moving away from their base position
	
	vec3 direction( _ego->GetPosition() - _base );
	direction.y = 0;
	scalar distance = direction.Length();
	if ( distance > 0 ) direction /= distance;
	
	if ( ( distance > 3 ) && ( direction.DotProduct( -_toward ) < 0.2f ) )
	{
		// Abandon the target and transition to FOLLOWING
		
		_target = NULL;
		StartFollowing();
		return;
	}
	
	// Detect inactivity
	
	if ( ( _now - _transition ) > 15 )
	{
		// Abandon the target and transition to FOLLOWING
		
		_target = NULL;
		StartFollowing();
	}
}

////////////////////////////////////////////////////////////////////////////////
// Guide - start - SUGGESTING

void Guide::StartSuggesting()
{
	// Set the state
	
	SetState( GS_SUGGESTING );
	
	// Select a target
	
	if ( _target = Application::GetInstance()->GetFeatures()->SelectFeature( _ego->GetPosition() ) )
	{
		// Mark it as suggested
		
		_target->Suggest();
	}
	else
	{
		// If none was found, go back to waiting
		
		StartWaitingToFollow();
		return;
	}
	
	// Ask the user to follow
	
	XMLElement bml( "bml" );
	bml.AddElement( "vgs", "speak" )->AddAttribute( "file", "follow.ogg" );
	
	// Gaze at the target
	
	XMLElement* gaze1 = bml.AddElement( "gaze" );
	gaze1->AddAttribute( "target", _target->GetClosestLocationPoint( _position ).c_str() );
	
	// Wave the arm

	XMLElement* wave = bml.AddElement( "sbm", "animation" );
	wave->AddAttribute( "id", "wave" );
	wave->AddAttribute( "name", "LHandOnHip_RArm_GestureOffer" );
	
	// Shift gaze back to the ego
	
	XMLElement* gaze2 = bml.AddElement( "gaze" );
	gaze2->AddAttribute( "target", "ego" );
	gaze2->AddAttribute( "start", "wave:end" );
	
	// Send requests
	
	Send( &bml );
}

////////////////////////////////////////////////////////////////////////////////
// Guide - run - SUGGESTING

void Guide::RunSuggesting()
{
	// Check if the speech is still in progress
	
	XMLElement query( "vgs", "speaking" );
	
	if ( XMLElement* result = Send( &query ) )
	{
		result->Delete();
		return;
	}
	
	// Transition to LEADING
	
	StartLeading();
}

////////////////////////////////////////////////////////////////////////////////
// Guide - start - WALKING_TO_PRESENT

void Guide::StartWalkingToPresent()
{
	// Set the state
	
	SetState( GS_WALKING_TO_PRESENT );
	
	// Walk toward the target feature's guide point
	
	XMLElement bml( "bml" );
	bml.AddElement( "gaze" )->AddAttribute( "target", _target->GetClosestLocationPoint( _position ).c_str() );
	bml.AddElement( "vgs", "walk" )->AddAttribute( "target", _target->GetClosestGuidePoint( _position ).c_str() );
	Send( &bml );
}

////////////////////////////////////////////////////////////////////////////////
// Guide - run - WALKING_TO_PRESENT

void Guide::RunWalkingToPresent()
{
	// Check if the ego has left the feature
	
	if ( _target->IsOutsideActivationPoints( _ego->GetPosition() ) )
	{
		// Drop the target and transition to FOLLOWING
		
		_target = NULL;
		StartFollowing();
		return;
	}
	
	// Check if the walk is still in progress
	
	XMLElement query( "vgs", "walking" );
	
	if ( XMLElement* result = Send( &query ) )
	{
		result->Delete();
		return;
	}
	
	// Transition to PRESENTING
	
	StartPresenting();
}

////////////////////////////////////////////////////////////////////////////////
// Guide - start - PRESENTING

void Guide::StartPresenting()
{
	// Set the state
	
	SetState( GS_PRESENTING );
	
	// Mark the target as presented
	
	_script->Delete();
	_script = _target->Present();
	
	// Create a default script
	
	if ( NULL == _script )
	{
		// Create the script with one block
		
		_script = new XMLElement( "script" );
		XMLElement* block = _script->AddElement( "vgs", "block" );
		
		block->AddAttribute( "actor", "guide" );
		block->AddAttribute( "time", "0.0" );
		
		// Start talking
		
		block->AddElement( "vgs", "speak" )->AddAttribute( "file", ( "features/" + _target->GetName() + "/voice.ogg" ).c_str() );
		
		// Gesture
		
		XMLElement* body = block->AddElement( "body" );
		body->AddAttribute( "id", "posture" );
		body->AddAttribute( "posture", "HandsAtSide_Arms_Sweep" );
		
		// Gaze at the target
		
		XMLElement* gaze = block->AddElement( "gaze" );
		gaze->AddAttribute( "target", _target->GetClosestLocationPoint( _position ).c_str() );
		
		// Gaze back to the ego
		
		block = _script->AddElement( "vgs", "block" );
		
		block->AddAttribute( "actor", "guide" );
		block->AddAttribute( "time", "5.0" );
		
		gaze = block->AddElement( "gaze" );
		gaze->AddAttribute( "target", "ego" );
		gaze->AddAttribute( "wander", "45" );
	}
}

////////////////////////////////////////////////////////////////////////////////
// Guide - state - PRESENTING

void Guide::RunPresenting()
{
	// Check if the ego has left the feature
	
	if ( ( NULL == _script ) || _target->IsOutsideActivationPoints( _ego->GetPosition() ) )
	{
		// Drop the target
		
		_target = NULL;
		
		// Destroy any characters that have been created in script
		
		if ( NULL != _script )
		{
			Actors* actors = Application::GetInstance()->GetActors();
			
			for ( XMLElementIterator it1 = _script->GetElements() ; it1 ; ++ it1 )
			{
				for ( XMLElementIterator it2 = it1->GetElements() ; it2 ; ++ it2 )
				{
					if ( it2->IsNamed( "vgs", "destroy" ) )
					{
						if ( PCSTR actor = it1->GetAttributeString( "actor" ) )
						{
							actors->Send( actor, *it2 );
						}
					}
				}
			}
			
			_script->Delete();
			_script = NULL;
		}
		
		// Transition to FOLLOWING
		
		StartFollowing();
		
		return;
	}
	
	// Dispatch new blocks from the script
	
	Actors* actors = Application::GetInstance()->GetActors();
	scalar now = Application::GetInstance()->GetTime() - _transition;
	
	for ( XMLElementIterator it = _script->GetElements() ; it ; )
	{
		XMLElement* block = *it;
		++ it;
		
		if ( block->GetAttributeFloat( "time" ) <= now )
		{
			string actor( "guide" );
			
			block->SetNamespace( "" );
			block->SetName( "bml" );
			
			if ( XMLAttribute* att = block->GetAttribute( "time" ) )
			{
				att->Delete();
			}
			
			if ( XMLAttribute* att = block->GetAttribute( "actor" ) )
			{
				actor = att->GetValue();
				att->Delete();
			}
			
			for ( XMLElementIterator itp = block->GetElements() ; itp ; )
			{
				XMLElement* child = *itp;
				++ itp;
				
				if ( child->IsNamed( "vgs", "create" ) || child->IsNamed( "vgs", "destroy" ) )
				{
					actors->Send( actor.c_str(), child );
					child->Delete();
				}
			}
			
			actors->Send( actor.c_str(), block );
			block->Delete();
		}
	}
	
	// Check if the speech is still in progress
	
	XMLElement query( "vgs", "speaking" );
	
	if ( XMLElement* result = Send( &query ) )
	{
		result->Delete();
		return;
	}
	
	// Check if the script is not empty
	
	if ( _script->GetElements() ) return;
	
	// Transition to WAITING_TO_FOLLOW
	
	_target = NULL;
	StartWaitingToFollow( 2.0f );
}

////////////////////////////////////////////////////////////////////////////////
// Guide - send a behaviour request to the guide

XMLElement* Guide::Send( XMLElement* request )
{
	Actors* actors = Application::GetInstance()->GetActors();
	return actors ? actors->Send( "guide", request ) : NULL;
}

////////////////////////////////////////////////////////////////////////////////
// Guide - set a new state

void Guide::SetState( GuideState state )
{
	static PCSTR states[] =
	{
		"FOLLOWING", "WAITING_TO_FOLLOW",
		"LEADING", "WAITING_TO_LEAD", "WAITING_TO_PRESENT",
		"SUGGESTING", "WALKING_TO_PRESENT", "PRESENTING"
	};
	
	// Update the state
	
	_state = (GuideState)state;
	
	// Set the transition time
	
	_transition = _now;
	
	// Stop walking and speaking
	
	XMLElement bml( "bml" );
	bml.AddElement( "vgs", "walk" );
	bml.AddElement( "vgs", "speak" );
	Send( &bml );
	
	// Debug
	
	cout << "Guide state is now " << states[ _state ] << endl;
}

////////////////////////////////////////////////////////////////////////////////
// Guide - check if the ego has activated a feature

bool Guide::CheckActivation()
{
	// Check if the ego is within the outer radius of an activation point
	
	Feature* feature = Application::GetInstance()->GetFeatures()->GetActivatedFeature( _ego->GetPosition() );
	if ( NULL == feature ) return false;
	
	// Check if the feature is to be ignored
	
	if ( _ignore == feature ) return false;
	
	// Adopt the feature as our target
	
	_target = _ignore = feature;
	
	// Transition to WALKING_TO_PRESENT
	
	StartWalkingToPresent();
	
	return true;
}
