//
// Material.cpp
// Copyright Michael J Stokes, 2008.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "Material.h"
#include "stb_image.h"


///////////////////////////////////////////////////////////////////////////////
// Material - construction

Material::Material( PCSTR name ) : Resource( name )
{
	_opacity	= 1;
	_shininess	= 10;
	_flat		= false;
	_pad1		= 1;
	_pad2		= 1;
	_alpha		= 0.5f;
	
	_image		= NULL;
	_width		= 0;
	_height		= 0;
	_texture	= 0;
	
	_ambient.Set( 1, 1, 1 );
	_diffuse.Set( 1, 1, 1 );
	_specular.Set( 1, 1, 1 );
}

///////////////////////////////////////////////////////////////////////////////
// Material - destruction

Material::~Material()
{
	if ( NULL != _image ) free( _image );
}

///////////////////////////////////////////////////////////////////////////////
// Material - load

bool Material::Load()
{
	std::ifstream stream( _name.c_str(), std::ios_base::binary );
	if ( ! stream.is_open() ) return false;
	
	cout << "Loading material texture '" << _name << "'..." << endl;
	
	INT32 components = 0;
	_image = stbi_load_from_file( &stream, &_width, &_height, &components, 3 );
	
	return ( NULL != _image );
}

///////////////////////////////////////////////////////////////////////////////
// Material - upload to graphics hardware

bool Material::Upload()
{
	if ( 0 == _texture ) glGenTextures( 1, &_texture );
	
	glBindTexture( GL_TEXTURE_2D, _texture );
	
	gluBuild2DMipmaps( GL_TEXTURE_2D,
		GL_RGB, _width, _height, GL_RGB, GL_UNSIGNED_BYTE, _image );
	
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );
	
	return true;
}

///////////////////////////////////////////////////////////////////////////////
// Material - erase from graphics hardware

void Material::Erase()
{
	if ( 0 != _texture ) glDeleteTextures( 1, &_texture );
}

///////////////////////////////////////////////////////////////////////////////
// Material - begin rendering

void Material::Begin()
{
	if ( NULL == this ) return;
	if ( 0 == _texture ) Upload();
	
	glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT, _ambient );
	glMaterialfv( GL_FRONT_AND_BACK, GL_DIFFUSE, _diffuse );
	glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR, _specular );
	glMaterialf( GL_FRONT_AND_BACK, GL_SHININESS, _flat ? 0 : _shininess );
	
	if ( 0 != _texture )
	{
		glBindTexture( GL_TEXTURE_2D, _texture );
		glEnable( GL_TEXTURE_2D );
	}
}

///////////////////////////////////////////////////////////////////////////////
// Material - end rendering

void Material::End()
{
	if ( NULL == this ) return;
	if ( 0 != _texture ) glDisable( GL_TEXTURE_2D );
}
