//
// Mesh.cpp
// Copyright Michael J Stokes, 2008.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "Application.h"
#include "Mesh.h"
#include "MeshCompiler.h"
#include "Renderer.h"
#include "Material.h"


////////////////////////////////////////////////////////////////////////////////
// Mesh - construction

Mesh::Mesh( PCSTR name, PCSTR materialPath, scalar scale ) : Resource( name )
{
	_materials	= materialPath;
	_scale		= scale;
	_vbuf		= 0;
	_ibuf		= 0;
}

////////////////////////////////////////////////////////////////////////////////
// Mesh - destruction

Mesh::~Mesh()
{
	for ( vector<Group>::iterator it = _groups.begin() ; it != _groups.end() ; ++ it )
	{
		it->material->Release();
	}
}

///////////////////////////////////////////////////////////////////////////////
// Mesh - load

bool Mesh::Load()
{
	cout << "Loading mesh '" << _name << "'..." << endl;
	
	string compiled( _name + ".bin" );
	
	if ( MeshCompiler::IsFileNewer( _name.c_str(), compiled.c_str() ) )
	{
		cerr << "Mesh '" << _name << "' needs to be recompiled." << endl;
		if ( ! MeshCompiler::CompileOBJ( _name.c_str(), compiled.c_str(), _scale, _materials.c_str() ) ) return false;
	}
	
	std::ifstream stream( compiled.c_str(), std::ios_base::binary );
	
	if ( ! stream.is_open() )
	{
		cerr << "Compiled mesh '" << compiled << "' not found." << endl;
		return false;
	}
	
	char id[ 8 ];
	memset( id, 0, 8 );
	stream.read( id, 8 );
	
	if ( 0 != memcmp( id, "VGS-M-1", 8 ) )
	{
		cerr << "Compiled mesh '" << compiled << "' is invalid." << endl;
		return false;
	}
	
	UINT32 count;
	
	stream.read( (PSTR)&count, sizeof( count ) );
	_vertices.resize( count );
	stream.read( (PSTR)&_vertices.front(), sizeof( Vertex ) * count );
	
	stream.read( (PSTR)&count, sizeof( count ) );
	_indices.resize( count );
	stream.read( (PSTR)&_indices.front(), sizeof( UINT ) * count );
	
	stream.read( (PSTR)&count, sizeof( count ) );
	_groups.resize( count );

	Renderer* renderer = Application::GetInstance()->GetRenderer();
	
	for ( vector<Group>::iterator it = _groups.begin() ; it != _groups.end() ; ++ it )
	{
		UINT32 sl;
		stream.read( (PSTR)&sl, sizeof( sl ) );
		char* sp = new char[ sl ];
		stream.read( sp, sl );
		string material( sp, sl );
		delete [] sp;
		
		stream.read( (PSTR)&it->firstIndex, sizeof( UINT32 ) );
		stream.read( (PSTR)&it->indexCount, sizeof( UINT32 ) );
		
		string path( _materials + "/" + material );
		it->material = renderer->LoadMaterial( path.c_str() );
	}
	
	return true;
}

///////////////////////////////////////////////////////////////////////////////
// Mesh - upload to graphics hardware

bool Mesh::Upload()
{
	if ( 0 == _vbuf ) glGenBuffers( 1, &_vbuf );
	if ( 0 == _vbuf ) return false;
	glBindBuffer( GL_ARRAY_BUFFER, _vbuf );
	glBufferData( GL_ARRAY_BUFFER, _vertices.size() * sizeof( Vertex ), &_vertices.front(), GL_STATIC_DRAW );
	glBindBuffer( GL_ARRAY_BUFFER, 0 );
	
	if ( 0 == _ibuf ) glGenBuffers( 1, &_ibuf );
	if ( 0 == _ibuf ) return false;
	glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, _ibuf );
	glBufferData( GL_ELEMENT_ARRAY_BUFFER, _indices.size() * sizeof( UINT32 ), &_indices.front(), GL_STATIC_DRAW );
	glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, 0 );
	
	return true;
}

///////////////////////////////////////////////////////////////////////////////
// Mesh - erase from graphics hardware

void Mesh::Erase()
{
	if ( 0 != _ibuf ) glDeleteBuffers( 1, &_ibuf );
	if ( 0 != _vbuf ) glDeleteBuffers( 1, &_vbuf );
}

////////////////////////////////////////////////////////////////////////////////
// Mesh - render

void Mesh::Render()
{
	// Upload
	
	if ( 0 == _vbuf ) Upload();
	
	// Select geometry
	
	glBindBuffer( GL_ARRAY_BUFFER, _vbuf );
	glVertexPointer( 3, GL_FLOAT, sizeof( Vertex ), NULL );
	glNormalPointer( GL_FLOAT, sizeof( Vertex ), (vec3*)NULL + 1 );
	glTexCoordPointer( 2, GL_FLOAT, sizeof( Vertex ), (vec3*)NULL + 2 );
	glBindBuffer( GL_ARRAY_BUFFER, 0 );
	glEnable( GL_VERTEX_ARRAY );
	glEnable( GL_NORMAL_ARRAY );
	glEnable( GL_TEXTURE_COORD_ARRAY );
	
	// Select indices
	
	glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, _ibuf );
	
	// Render
	
	for ( vector<Group>::iterator it = _groups.begin() ; it != _groups.end() ; ++ it )
	{
		it->material->Begin();
		glDrawElements( GL_TRIANGLES, it->indexCount, GL_UNSIGNED_INT, (int*)NULL + it->firstIndex );
		it->material->End();
	}
	
	// Deselect indices
	
	glDisable( GL_VERTEX_ARRAY );
	glDisable( GL_NORMAL_ARRAY );
	glDisable( GL_TEXTURE_COORD_ARRAY );
	glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, 0 );
}
