//
// MeshCompiler.cpp
// Copyright Michael J Stokes, 2008.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "MeshCompiler.h"
#include "OBJParser.h"
#include "Mesh.h"
#include "Character.h"


////////////////////////////////////////////////////////////////////////////////
// MeshCompiler - get the modified time for a file

UINT32 MeshCompiler::GetModifiedTime( PCSTR path )
{
#ifdef _WIN32
	HANDLE handle = CreateFile( path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
	
	if ( INVALID_HANDLE_VALUE != handle )
	{
		unsigned __int64 ft;
		GetFileTime( handle, NULL, NULL, (FILETIME*)&ft );
		CloseHandle( handle );
		return (UINT32)( ft >> 24 );
	}
#else
	struct stat sb;
	if ( 0 == stat( path, &sb ) ) return (UINT32)sb.st_mtime;
#endif
	
	return 0;
}

////////////////////////////////////////////////////////////////////////////////
// MeshCompiler - check if one file is newer than another

bool MeshCompiler::IsFileNewer( PCSTR newer, PCSTR older )
{
	return GetModifiedTime( newer ) > GetModifiedTime( older );
}

////////////////////////////////////////////////////////////////////////////////
// MeshCompiler - compile an OBJ mesh

bool MeshCompiler::CompileOBJ( PCSTR input, PCSTR output, scalar scale, PCSTR materials )
{
	MeshCompiler compiler;
	return compiler.CompileOBJImpl( input, output, scale, materials );
}

////////////////////////////////////////////////////////////////////////////////
// MeshCompiler - compile an OBJ mesh (implementation)

bool MeshCompiler::CompileOBJImpl( PCSTR input, PCSTR output, scalar scale, PCSTR materials )
{
	std::ifstream istream( input );
	
	cout << "Compiling mesh '" << input << "'..." << endl;
	
	if ( ! istream.is_open() )
	{
		cerr << "Unable to open OBJ input '" << input << "'." << endl;
		return false;
	}
	
	OBJParser parser( istream );
	
	vector<vec3>			positions;
	vector<vec3>			normals;
	vector<vec2>			uvs;
	vector<Group*>			groups;
	Group*					group = NULL;
	
	for ( ; ; )
	{
		while ( parser.GetNewline() );
		
		if ( parser.IsEOF() )
		{
			break;
		}
		else if ( parser.GetMatch( "f" ) )
		{
			int ip[ 3 ] = { 0, 0, 0 };
			int in[ 3 ] = { 0, 0, 0 };
			int iu[ 3 ] = { 0, 0, 0 };

			for ( int pt = 0 ; pt < 3 ; pt ++ )
			{
				if ( ! parser.GetInt( ip[ pt ] ) ) return false;
				if ( ! parser.GetMatch( "/" ) ) return false;
				parser.GetInt( iu[ pt ] );
				if ( ! parser.GetMatch( "/" ) ) return false;
				if ( ! parser.GetInt( in[ pt ] ) ) return false;
				
				if ( ip[ pt ] < 0 ) ip[ pt ] += (int)positions.size();
				if ( in[ pt ] < 0 ) in[ pt ] += (int)normals.size();
				if ( iu[ pt ] < 0 ) iu[ pt ] += (int)uvs.size();
			}

			AddVertex( ip[ 0 ], in[ 0 ], iu[ 0 ] );
			AddVertex( ip[ 2 ], in[ 2 ], iu[ 2 ] );
			AddVertex( ip[ 1 ], in[ 1 ], iu[ 1 ] );
			
			if ( NULL != group ) group->count += 3;
		}
		else if ( parser.GetMatch( "v" ) )
		{
			vec3 v;
			
			if ( ! parser.GetFloat( v.x ) ) return false;
			if ( ! parser.GetFloat( v.y ) ) return false;
			if ( ! parser.GetFloat( v.z ) ) return false;
			
			v.z *= -1;
			v *= scale;
			positions.push_back( v );
		}
		else if ( parser.GetMatch( "vn" ) )
		{
			vec3 v;
			
			if ( ! parser.GetFloat( v.x ) ) return false;
			if ( ! parser.GetFloat( v.y ) ) return false;
			if ( ! parser.GetFloat( v.z ) ) return false;
			
			v.z *= -1;
			v.Normalize();
			normals.push_back( v );
		}
		else if ( parser.GetMatch( "vt" ) )
		{
			scalar w;
			vec2 v;
			
			if ( ! parser.GetFloat( v.u ) ) return false;
			if ( ! parser.GetFloat( v.v ) ) return false;
			parser.GetFloat( w );
			
			v.v = 1.0f - v.v;
			uvs.push_back( v );
		}
		else if ( parser.GetMatch( "g" ) || parser.GetMatch( "o" ) )
		{
			string name = parser.GetNameString();
			if ( name.empty() ) return false;
			continue;
		}
		else if ( parser.GetMatch( "usemtl" ) )
		{
			string name = parser.GetNameString();
			if ( name.empty() ) return false;
			
			string::size_type pos = name.find( ".jpg" );
			if ( string::npos != pos ) name.resize( pos + 4 );
			else if ( string::npos != name.find( ".jp" ) ) name.append( "g" );
			
			if ( NULL != materials )
			{
				for ( ; ; )
				{
					string test( materials );
					test += "/";
					test += name;
					
					if ( 0 != GetModifiedTime( test.c_str() ) ) break;
					
					pos = name.find( '_' );
					if ( string::npos == pos ) break;
					name = name.substr( pos + 1 );
				}
			}
			
			group			= new Group;
			group->name		= name;
			group->first	= (int)_indices.size();
			group->count	= 0;
			
			groups.push_back( group );
			
			continue;
		}
		else if ( parser.GetMatch( "mtllib" ) )
		{
			string path = parser.GetNameString();
			if ( path.empty() ) return false;
			continue;
		}
		else if ( parser.GetMatch( "s" ) )
		{
			parser.GetToken();
		}
		else
		{
			cerr << "Unknown obj statement: '" << parser.GetToken() << "'" << endl;
		}
		
		if ( ! parser.GetNewline() ) return false;
	}
	
	std::ofstream ostream( output, std::ios_base::binary | std::ios_base::out | std::ios_base::trunc );
	
	if ( ! ostream.is_open() )
	{
		cerr << "Unable to open OBJ.BIN output '" << output << "'." << endl;
		for ( vector<Group*>::iterator it = groups.begin() ; it != groups.end() ; ++ it ) delete *it;
		return false;
	}
	
	ostream.write( "VGS-M-1", 8 );
	
	UINT32 count = (UINT32)_vertices.size();
	ostream.write( (PCSTR)&count, sizeof( count ) );
	
	for ( vector<Triplet>::iterator it = _vertices.begin() ; it != _vertices.end() ; ++ it )
	{
		ostream.write( (PCSTR)&positions[ it->position - 1 ], sizeof( vec3 ) );
		
		if ( it->normal > 0 )
			ostream.write( (PCSTR)&normals[ it->normal - 1 ], sizeof( vec3 ) );
		else
			ostream.write( (PCSTR)&vec3origin, sizeof( vec3 ) );
		
		if ( it->uv > 0 )
			ostream.write( (PCSTR)&uvs[ it->uv - 1 ], sizeof( vec2 ) );
		else
			ostream.write( (PCSTR)&vec2origin, sizeof( vec2origin ) );
	}
	
	count = (UINT32)_indices.size();
	ostream.write( (PCSTR)&count, sizeof( count ) );
	ostream.write( (PCSTR)&_indices.front(), sizeof( UINT32 ) * count );
	
	count = (UINT32)groups.size();
	ostream.write( (PCSTR)&count, sizeof( count ) );
	
	for ( vector<Group*>::iterator it = groups.begin() ; it != groups.end() ; ++ it )
	{
		group = *it;
		count = (UINT32)group->name.length();
		ostream.write( (PCSTR)&count, sizeof( count ) );
		ostream.write( group->name.c_str(), count );
		ostream.write( (PCSTR)&group->first, sizeof( group->first ) );
		ostream.write( (PCSTR)&group->count, sizeof( group->count ) );
		delete group;
	}
	
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// MeshCompiler - add a vertex to the vertex map

void MeshCompiler::AddVertex( UINT32 ip, UINT32 in, UINT32 iu )
{
	UINT hash = ( ip + in + iu ) & 255;
	vector<UINT32>& mp = _map[ hash ];
	
	for ( vector<UINT32>::iterator it = mp.begin() ; it != mp.end() ; ++ it )
	{
		const Triplet& v = _vertices[ *it ];
		
		if ( ( v.position == ip ) && ( v.normal == in ) && ( v.uv == iu ) )
		{
			_indices.push_back( *it );
			return;
		}
	}
	
	Triplet v;
	v.position = ip;
	v.normal = in;
	v.uv = iu;
	_vertices.push_back( v );
	
	mp.push_back( (UINT32)_vertices.size() - 1 );
	_indices.push_back( (UINT32)_vertices.size() - 1 );
}

////////////////////////////////////////////////////////////////////////////////
// MeshCompiler - compile a skeleton mesh

bool MeshCompiler::CompileSkeleton( PCSTR input, PCSTR output, vector<string> & bones )
{
	cout << "Compiling character mesh '" << input << "'..." << endl;
	
	XMLElement* root = XMLElement::FromFile( input );
	XMLAutoDelete ad( root );
	
	if ( ! root->IsNamed( "mesh" ) )
	{
		cerr << "Unable to open character mesh input '" << input << "'." << endl;
		return false;
	}
	
	vector<Character::Vertex> vertices;
	vector<UINT32> indices;
	vector<Group*> groups;
	vector<UINT> boneMap;
	
	if ( XMLElement* boneTable = root->GetElement( "bones" ) )
	{
		for ( XMLElementIterator it = boneTable->GetElements() ; it ; ++ it )
		{
			if ( it->IsNamed( "bone" ) )
			{
				PCSTR name = it->GetAttributeString( "name" );
				if ( NULL == name ) return false;
				
				bool found = false;
				
				for ( UINT search = 0 ; search < bones.size() ; search ++ )
				{
					if ( bones[ search ] == name )
					{
						boneMap.push_back( search );
						found = true;
						break;
					}
				}
				
				if ( ! found )
				{
					cerr << "Character mesh '" << input << "' references unknown bone '" << name << "'." << endl;
					return false;
				}
			}
		}
	}
	
	if ( XMLElement* verticesContainer = root->GetElement( "vertices" ) )
	{
		for ( XMLElementIterator vertexIt = verticesContainer->GetElements() ; vertexIt ; ++ vertexIt )
		{
			if ( vertexIt->IsNamed( "vertex" ) )
			{
				vertices.resize( vertices.size() + 1 );
				Character::Vertex& vertex = vertices[ vertices.size() - 1 ];
				
				vertex.bone[ 0 ] = vertex.bone[ 1 ] = vertex.bone[ 2 ] = vertex.bone[ 3 ] = 0;
				vertex.weight[ 0 ] = vertex.weight[ 1 ] = vertex.weight[ 2 ] = vertex.weight[ 3 ] = 0;
				
				int bone = 0;
				scalar sum = 0;
				
				for ( XMLElementIterator it = vertexIt->GetElements() ; it ; ++ it )
				{
					if ( it->IsNamed( "position" ) )
					{
						vertex.position.x = it->GetAttributeFloat( "x" );
						vertex.position.y = it->GetAttributeFloat( "y" );
						vertex.position.z = it->GetAttributeFloat( "z" );
					}
					else if ( it->IsNamed( "normal" ) )
					{
						vertex.normal.x = it->GetAttributeFloat( "x" );
						vertex.normal.y = it->GetAttributeFloat( "y" );
						vertex.normal.z = it->GetAttributeFloat( "z" );
						vertex.normal.Normalize();
					}
					else if ( it->IsNamed( "material" ) )
					{
						vertex.material.s = it->GetAttributeFloat( "s" );
						vertex.material.t = it->GetAttributeFloat( "t" );
					}
					else if ( it->IsNamed( "bone" ) )
					{
						int index		= boneMap[ it->GetAttributeInt( "id" ) ];
						scalar weight	= it->GetAttributeFloat( "weight" );
						
						if ( bone < 4 )
						{
							vertex.bone[ bone ]		= (scalar)index;
							vertex.weight[ bone ]	= weight;
							sum += weight;
							bone ++;
						}
						else
						{
							scalar minWeight = 10;
							int minIndex = 0;
							
							for ( bone = 0 ; bone < 4 ; bone ++ )
							{
								if ( vertex.weight[ bone ] < minWeight )
								{
									minIndex	= bone;
									minWeight	= vertex.weight[ bone ];
								}
							}
							
							if ( minWeight < weight )
							{
								sum -= minWeight;
								vertex.bone[ minIndex ]		= (scalar)index;
								vertex.weight[ minIndex ]	= weight;
								sum += weight;
							}
						}
					}
				}
				
				if ( sum > 0 )
				{
					for ( int index = 0 ; index < 4 ; index ++ )
					{
						vertex.weight[ index ] /= sum;
					}
				}
			}
		}
	}
	
	if ( XMLElement* groupContainer = root->GetElement( "groups" ) )
	{
		for ( XMLElementIterator groupIt = groupContainer->GetElements() ; groupIt ; ++ groupIt )
		{
			if ( groupIt->IsNamed( "group" ) )
			{
				Group* group = new Group();
				groups.push_back( group );
				
				if ( PCSTR material = groupIt->GetAttributeString( "material" ) )
				{
					group->name = material;
				}
				
				group->first = (UINT32)indices.size();
				
				for ( XMLElementIterator faceIt = groupIt->GetElements() ; faceIt ; ++ faceIt )
				{
					if ( faceIt->IsNamed( "face" ) )
					{
						indices.push_back( faceIt->GetAttributeInt( "v1" ) );
						indices.push_back( faceIt->GetAttributeInt( "v2" ) );
						indices.push_back( faceIt->GetAttributeInt( "v3" ) );
					}
				}
				
				group->count = (UINT32)indices.size() - group->first;
			}
		}
	}
	
	std::ofstream ostream( output, std::ios_base::binary | std::ios_base::out | std::ios_base::trunc );
	
	if ( ! ostream.is_open() )
	{
		cerr << "Unable to open OBJ.BIN output '" << output << "'." << endl;
		for ( vector<Group*>::iterator it = groups.begin() ; it != groups.end() ; ++ it ) delete *it;
		return false;
	}
	
	ostream.write( "VGS-C-1", 8 );
	
	UINT32 count = (UINT32)vertices.size();
	ostream.write( (PCSTR)&count, sizeof( count ) );
	ostream.write( (PCSTR)&vertices.front(), sizeof( Character::Vertex ) * count );
	
	count = (UINT32)indices.size();
	ostream.write( (PCSTR)&count, sizeof( count ) );
	ostream.write( (PCSTR)&indices.front(), sizeof( UINT32 ) * count );
	
	count = (UINT32)groups.size();
	ostream.write( (PCSTR)&count, sizeof( count ) );
	
	for ( vector<Group*>::iterator it = groups.begin() ; it != groups.end() ; ++ it )
	{
		Group* group = *it;
		count = (UINT32)group->name.length();
		ostream.write( (PCSTR)&count, sizeof( count ) );
		ostream.write( group->name.c_str(), count );
		ostream.write( (PCSTR)&group->first, sizeof( group->first ) );
		ostream.write( (PCSTR)&group->count, sizeof( group->count ) );
		delete group;
	}
	
	return true;
}
