//
// OBJParser.h
// Copyright Michael J Stokes, 2008.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#pragma once


// The maximum length of a token

#ifndef OBJ_PARSER_MAX_TOKEN
#define OBJ_PARSER_MAX_TOKEN 256
#endif


class OBJParser
{
// Construction
public:
	
//	Construct an OBJParser, reading from a C++ istream
	inline OBJParser( std::istream& stream ) : _stream( stream )
	{
		_buf[ 0 ][ 0 ] = _buf[ 1 ][ 0 ] = 0;
		_cur	= 0;
		_line	= 1;
		Read();
	}
	
// Methods
public:
	
//	Get the current line number
	inline unsigned int GetLineNumber() const
	{
		return _line;
	}
	
//	Check if the parser has reached EOF
	inline bool IsEOF() const
	{
		return ( 0 == _buf[ _cur ][ 0 ] );
	}
	
//	Consume and return a raw token
	inline const char * GetToken()
	{
		if ( IsEOF() ) return NULL;
		const char * ret = _buf[ _cur ];
		Read();
		return ret;
	}
	
//	Get and consume an exact match, or do nothing
	inline bool GetMatch( const char * match )
	{
		if ( 0 != strcmp( _buf[ _cur ], match ) ) return false;
		Read();
		return true;
	}
	
//	Get and consume a newline
	inline bool GetNewline()
	{
		return GetMatch( "\n" );
	}
	
//	Get and consume an integer number
	inline bool GetInt( int& dest )
	{
		if ( IsEOF() ) return false;
		if ( 1 != sscanf( _buf[ _cur ], "%i", &dest ) ) return false;
		Read();
		return true;
	}
	
//	Get and consume a floating point number
	inline bool GetFloat( float& dest )
	{
		if ( IsEOF() ) return false;
		if ( 1 != sscanf( _buf[ _cur ], "%f", &dest ) ) return false;
		Read();
		return true;
	}
	
//	Consume and return a name token, i.e. not an operator
	inline const char * GetName()
	{
		if ( IsEOF() ) return NULL;
		if ( strchr( "#/\\\n", _buf[ _cur ][ 0 ] ) ) return NULL;
		const char * ret = _buf[ _cur ];
		Read();
		return ret;
	}
	
//	Consume and return a name string, possibly several tokens long
	inline string GetNameString()
	{
		string ret;
		
		if ( const char * token = GetName() )
		{
			ret += token;
			
			while ( ! GetNewline() )
			{
				const char * token = GetToken();
				if ( NULL == token ) break;
				ret.append( " " );
				ret.append( token );
			}
		}
		
		return ret;
	}
	
// Implementation
protected:
	
	inline void Read()
	{
		_cur ^= 1;
		
		while ( ReadRaw() )
		{
			if ( 0 == strcmp( _buf[ _cur ], "\\" ) )
			{
				while ( ReadRaw() )
				{
					if ( 0 == strcmp( _buf[ _cur ], "\n" ) ) break;
				}
			}
			else if ( 0 == strcmp( _buf[ _cur ], "#" ) )
			{
				while ( ReadRaw() )
				{
					if ( 0 == strcmp( _buf[ _cur ], "\n" ) ) break;
				}
				
				break;
			}
			else
			{
				break;
			}
		}
	}
	
	inline bool ReadRaw()
	{
		char* s = _buf[ _cur ];
		char* e = s + OBJ_PARSER_MAX_TOKEN - 1;
		
		*s = *e = 0;
		
		char c;
		
		do
		{
			c = ReadChar();
			if ( ! c ) return false;
		}
		while ( strchr( " \t\r", c ) );
		
		if ( strchr( "/#\\\n", c ) )
		{
			*s ++ = c;
			*s = 0;
			if ( '\n' == c ) _line ++;
			return true;
		}
		
		while ( s < e )
		{
			*s ++ = c;
			c = ReadChar();
			
			if ( ! c || strchr( " /#\\\t\r\n", c ) )
			{
				UnreadChar();
				break;
			}
		}
		
		*s = 0;
		return true;
	}
	
	inline char ReadChar()
	{
		int i = _stream.get();
		return ( i > 0 ) ? (char)i : 0;
	}
	
	inline void UnreadChar()
	{
		_stream.unget();
	}
	
// Data
protected:
	std::istream&	_stream;
	char			_buf[ 2 ][ OBJ_PARSER_MAX_TOKEN ];
	int				_cur;
	unsigned int	_line;
};
