//
// Physics.cpp
// Copyright Michael J Stokes, 2008.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "Physics.h"
#include "Mesh.h"


///////////////////////////////////////////////////////////////////////////////
// Physics construction

Physics::Physics()
{
	_sdk = NxCreatePhysicsSDK( NX_PHYSICS_SDK_VERSION, NULL, NULL );
	if ( NULL == _sdk ) return;
	
	NxSceneDesc scene;
	_scene = _sdk->createScene( scene );
	
	_manager = NxCreateControllerManager( new NxUserAllocatorDefault() );
}

///////////////////////////////////////////////////////////////////////////////
// Physics destruction

Physics::~Physics()
{
	if ( NULL != _manager ) NxReleaseControllerManager( _manager );
	if ( NULL != _sdk ) _sdk->release();
}

///////////////////////////////////////////////////////////////////////////////
// Physics start simulation

void Physics::Start( scalar delta )
{
	_scene->simulate( delta );
	_scene->flushStream();
}

///////////////////////////////////////////////////////////////////////////////
// Physics finish simulation

void Physics::Finish()
{
	_scene->fetchResults( NX_RIGID_BODY_FINISHED, true );
}

///////////////////////////////////////////////////////////////////////////////
// Physics add a mesh

void* Physics::AddMesh( Mesh* mesh )
{
	NxInitCooking();
	
	NxTriangleMeshDesc tmd;
	
	tmd.numVertices			= (NxU32)mesh->_vertices.size();
	tmd.numTriangles		= (NxU32)mesh->_indices.size() / 3;
	tmd.pointStrideBytes	= sizeof( Mesh::Vertex );
	tmd.triangleStrideBytes	= sizeof( UINT ) * 3;
	tmd.points				= &mesh->_vertices.front().position;
	tmd.triangles			= &mesh->_indices.front();
	
	WriteStream stream;
	NxCookTriangleMesh( tmd, stream );
	NxCloseCooking();
	
	NxTriangleMesh* tm = _sdk->createTriangleMesh( ReadStream( stream ) );
	if ( NULL == tm ) return NULL;
	
	NxTriangleMeshShapeDesc tmsd;
	tmsd.meshData = tm;
	
	NxActorDesc ad;
	ad.shapes.pushBack( &tmsd );
	
	return _scene->createActor( ad );
}

///////////////////////////////////////////////////////////////////////////////
// Physics remove a mesh

void Physics::RemoveMesh( void* handle )
{
	NxActor* actor = (NxActor*)handle;
	NxShape * const* shapes = actor->getShapes();
	
	list<NxTriangleMesh*> release;
	
	for ( int i = 0, n = actor->getNbShapes() ; i < n ; i ++, shapes ++ )
	{
		if ( NxTriangleMeshShape* tms = ( *shapes )->isTriangleMesh() )
		{
			release.push_back( &tms->getTriangleMesh() );
		}
	}
	
	_scene->releaseActor( *actor );
	
	while ( ! release.empty() )
	{
		_sdk->releaseTriangleMesh( *release.back() );
		release.pop_back();
	}
}

///////////////////////////////////////////////////////////////////////////////
// Physics - drop an object

scalar Physics::Drop( const vec3& position )
{
	NxCapsuleControllerDesc desc;
	
	desc.radius			= 0.15f;
	desc.height			= 1.2f;
	desc.stepOffset		= 0.5f;
	desc.position.x		= (scalar)position.x;
	desc.position.y		= (scalar)40;
	desc.position.z		= (scalar)position.z;
	
	NxController* controller = GetControllerManager()->createController( GetScene(), desc );
	
	NxU32 flags = 0;
	NxVec3 nxScaled( 0, -80, 0 );
	controller->move( nxScaled, 0xFFFFFFFF, 0, flags );
	const NxExtendedVec3& nxPos( controller->getDebugPosition() );
	scalar y = (scalar)nxPos.y - desc.radius * 5;
	
	GetControllerManager()->releaseController( *controller );
	
	return y;
}


///////////////////////////////////////////////////////////////////////////////
// Physics::WriteStream

Physics::WriteStream::WriteStream()
{
	_data	= NULL;
	_length	= 0;
}

Physics::WriteStream::~WriteStream()
{
	free( _data );
}

NxStream& Physics::WriteStream::storeByte( NxU8 b )
{
	return storeBuffer( &b, sizeof( b ) );
}

NxStream& Physics::WriteStream::storeWord( NxU16 w )
{
	return storeBuffer( &w, sizeof( w ) );
}

NxStream& Physics::WriteStream::storeDword( NxU32 d )
{
	return storeBuffer( &d, sizeof( d ) );
}

NxStream& Physics::WriteStream::storeFloat( NxReal f )
{
	return storeBuffer( &f, sizeof( f ) );
}

NxStream& Physics::WriteStream::storeDouble( NxF64 f )
{
	return storeBuffer( &f, sizeof( f ) );
}

NxStream& Physics::WriteStream::storeBuffer( const void* buffer, NxU32 size )
{
	UINT allocated = ( _length + 1023 ) & ~1023;
	
	if ( size > ( allocated - _length ) )
	{
		allocated = ( _length + size + 1023 ) & ~1023;
		_data = (BYTE*)realloc( _data, allocated );
	}
	
	memcpy( _data + _length, buffer, size );
	_length += size;
	
	return *this;
}


///////////////////////////////////////////////////////////////////////////////
// Physics::ReadStream

Physics::ReadStream::ReadStream( WriteStream& source )
{
	_internal = new Internal( source._data, source._length );
}

Physics::ReadStream::ReadStream( const void* source, UINT length )
{
	_internal = new Internal( source, length );
}

Physics::ReadStream::ReadStream( const ReadStream& copy )
{
	_internal = copy._internal;
	_internal->ref();
}

Physics::ReadStream::~ReadStream()
{
	_internal->unref();
}

NxU8 Physics::ReadStream::readByte() const
{
	NxU8 v;
	readBuffer( &v, sizeof( v ) );
	return v;
}

NxU16 Physics::ReadStream::readWord() const
{
	NxU16 v;
	readBuffer( &v, sizeof( v ) );
	return v;
}

NxU32 Physics::ReadStream::readDword() const
{
	NxU32 v;
	readBuffer( &v, sizeof( v ) );
	return v;
}

float Physics::ReadStream::readFloat() const
{
	float v;
	readBuffer( &v, sizeof( v ) );
	return v;
}

double Physics::ReadStream::readDouble() const
{
	double v;
	readBuffer( &v, sizeof( v ) );
	return v;
}

void Physics::ReadStream::readBuffer( void* buffer, NxU32 size ) const
{
	if ( size > ( _internal->_length - _internal->_pos ) ) size = _internal->_length - _internal->_pos;
	memcpy( buffer, _internal->_data + _internal->_pos, size );
	_internal->_pos += size;
}

Physics::ReadStream::Internal::Internal( const void* source, UINT length )
{
	_data	= (const BYTE*)source;
	_length	= length;
	_pos	= 0;
	_refs	= 1;
}

void Physics::ReadStream::Internal::ref()
{
	_refs ++;
}

void Physics::ReadStream::Internal::unref()
{
	if ( 0 == -- _refs ) delete this;
}
