//
// RCO.h
// Copyright Michael J Stokes, 2008.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#pragma once


//
// RCO: A reference counted object
//

class RCO
{
// Construction
public:
	inline RCO( bool floating = false ) : _refs( floating ? 0 : 1 ) {};
	virtual ~RCO() {};
	
// Public Operations
public:
	
//	Add a reference
	inline int AddRef()
	{
		if ( NULL == this ) return 0;
		return ++ _refs;
	}
	
//	Release a reference, deleting if no references remain
	inline int Release()
	{
		if ( NULL == this ) return 0;
		int refs = -- _refs;
		if ( refs < 1 ) delete this;
		return refs;
	}
	
// Data
private:
	int	_refs;
};


//
// ScopedReference: Hold a reference on an RCO while in scope
//

class ScopedReference
{
// Construction
public:
	inline ScopedReference( RCO* target ) : _target( target )
	{
		_target->AddRef();
	}
	
	inline ~ScopedReference()
	{
		_target->Release();
	}
	
// Data
protected:
	RCO*	_target;
};


//
// NamedObject: An RCO that has a name
//

class NamedObject : public RCO
{
// Construction
public:
	inline NamedObject( PCSTR name, bool floating = false ) : RCO( floating )
	{
		if ( NULL != name ) _name = name;
	}
	
// Operations
public:
	
//	Get the object's name
	inline const string & GetName() const
	{
		return _name;
	}
	
//	Check if the object has a given name
	inline bool IsNamed( PCSTR name ) const
	{
		return ( 0 == _name.compare( name ) );
	}
	
//	Check if the object has a given name
	inline bool IsNamed( const string& name ) const
	{
		return ( 0 == _name.compare( name ) );
	}
	
// Data
protected:
	string	_name;
};
