//
// Renderer.cpp
// Copyright Michael J Stokes, 2008.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "Renderer.h"
#include "Shader.h"
#include "Material.h"
#include "Mesh.h"
#include "Character.h"


////////////////////////////////////////////////////////////////////////////////
// OpenGL extensions

PFNGLGENBUFFERSPROC					glGenBuffers;
PFNGLDELETEBUFFERSPROC				glDeleteBuffers;
PFNGLBINDBUFFERPROC					glBindBuffer;
PFNGLBUFFERDATAPROC					glBufferData;

PFNGLCREATESHADERPROC				glCreateShader;
PFNGLSHADERSOURCEPROC				glShaderSource;
PFNGLCOMPILESHADERPROC				glCompileShader;
PFNGLGETSHADERIVPROC				glGetShaderiv;
PFNGLGETSHADERINFOLOGPROC			glGetShaderInfoLog;
PFNGLATTACHSHADERPROC				glAttachShader;
PFNGLDETACHSHADERPROC				glDetachShader;
PFNGLDELETESHADERPROC				glDeleteShader;
PFNGLCREATEPROGRAMPROC				glCreateProgram;
PFNGLLINKPROGRAMPROC				glLinkProgram;
PFNGLVALIDATEPROGRAMPROC			glValidateProgram;
PFNGLGETPROGRAMIVPROC				glGetProgramiv;
PFNGLGETPROGRAMINFOLOGPROC			glGetProgramInfoLog;
PFNGLUSEPROGRAMPROC					glUseProgram;
PFNGLDELETEPROGRAMPROC				glDeleteProgram;
PFNGLGETUNIFORMLOCATIONPROC			glGetUniformLocation;
PFNGLUNIFORM1IPROC					glUniform1i;
PFNGLUNIFORM1FPROC					glUniform1f;
PFNGLUNIFORM3FVPROC					glUniform3fv;
PFNGLUNIFORM4FVPROC					glUniform4fv;
PFNGLUNIFORMMATRIX3FVPROC			glUniformMatrix3fv;
PFNGLUNIFORMMATRIX4FVPROC			glUniformMatrix4fv;
PFNGLBINDATTRIBLOCATIONPROC			glBindAttribLocation;
PFNGLVERTEXATTRIBPOINTERPROC		glVertexAttribPointer;
PFNGLENABLEVERTEXATTRIBARRAYPROC	glEnableVertexAttribArray;
PFNGLDISABLEVERTEXATTRIBARRAYPROC	glDisableVertexAttribArray;


////////////////////////////////////////////////////////////////////////////////
// Renderer construction

Renderer::Renderer()
{
	_skeletonShader	= NULL;
	_infoMesh		= NULL;
}

////////////////////////////////////////////////////////////////////////////////
// Renderer destruction

Renderer::~Renderer()
{
	while ( ! _shaders.empty() )
	{
		_shaders.front()->Release();
		_shaders.pop_front();
	}
	
	while ( ! _materials.empty() )
	{
		_materials.front()->Release();
		_materials.pop_front();
	}
	
	while ( ! _meshes.empty() )
	{
		_meshes.front()->Release();
		_meshes.pop_front();
	}
	
	while ( ! _characters.empty() )
	{
		_characters.front()->Release();
		_characters.pop_front();
	}
}

////////////////////////////////////////////////////////////////////////////////
// Renderer initialize

bool Renderer::Initialize()
{
	// Console
	
	PCSTR version = (PCSTR)glGetString( GL_VERSION );
	cout << "Initializing OpenGL..." << endl;
	cout << "Found OpenGL version " << version << ", need version 2.0+" << endl;
	
	// Initialize extensions
	
	(void*&)glGenBuffers				= GetGLProc( "glGenBuffers" );
	(void*&)glDeleteBuffers				= GetGLProc( "glDeleteBuffers" );
	(void*&)glBindBuffer				= GetGLProc( "glBindBuffer" );
	(void*&)glBufferData				= GetGLProc( "glBufferData" );
	
	(void*&)glCreateShader				= GetGLProc( "glCreateShader" );
	(void*&)glShaderSource				= GetGLProc( "glShaderSource" );
	(void*&)glCompileShader				= GetGLProc( "glCompileShader" );
	(void*&)glGetShaderiv				= GetGLProc( "glGetShaderiv" );
	(void*&)glGetShaderInfoLog			= GetGLProc( "glGetShaderInfoLog" );
	(void*&)glAttachShader				= GetGLProc( "glAttachShader" );
	(void*&)glDetachShader				= GetGLProc( "glDetachShader" );
	(void*&)glDeleteShader				= GetGLProc( "glDeleteShader" );
	(void*&)glCreateProgram				= GetGLProc( "glCreateProgram" );
	(void*&)glLinkProgram				= GetGLProc( "glLinkProgram" );
	(void*&)glValidateProgram			= GetGLProc( "glValidateProgram" );
	(void*&)glGetProgramiv				= GetGLProc( "glGetProgramiv" );
	(void*&)glGetProgramInfoLog			= GetGLProc( "glGetProgramInfoLog" );
	(void*&)glUseProgram				= GetGLProc( "glUseProgram" );
	(void*&)glDeleteProgram				= GetGLProc( "glDeleteProgram" );
	(void*&)glGetUniformLocation		= GetGLProc( "glGetUniformLocation" );
	(void*&)glUniform1i					= GetGLProc( "glUniform1i" );
	(void*&)glUniform1f					= GetGLProc( "glUniform1f" );
	(void*&)glUniform3fv				= GetGLProc( "glUniform3fv" );
	(void*&)glUniform4fv				= GetGLProc( "glUniform4fv" );
	(void*&)glUniformMatrix3fv			= GetGLProc( "glUniformMatrix3fv" );
	(void*&)glUniformMatrix4fv			= GetGLProc( "glUniformMatrix4fv" );
	(void*&)glBindAttribLocation		= GetGLProc( "glBindAttribLocation" );
	(void*&)glVertexAttribPointer		= GetGLProc( "glVertexAttribPointer" );
	(void*&)glEnableVertexAttribArray	= GetGLProc( "glEnableVertexAttribArray" );
	(void*&)glDisableVertexAttribArray	= GetGLProc( "glDisableVertexAttribArray" );
	
	if ( glGenBuffers && glDeleteBuffers && glBindBuffer && glBufferData )
	{
		cout << "Using ARB_vertex_buffer_object" << endl;
	}
	else
	{
		cerr << "OpenGL does not support ARB_vertex_buffer_object" << endl;
		return false;
	}
	
	if (	glCreateShader && glShaderSource && glCompileShader && glGetShaderiv && glGetShaderInfoLog && glAttachShader &&
			glDetachShader && glDeleteShader && glCreateProgram && glLinkProgram && glValidateProgram && glGetProgramiv &&
            glGetProgramInfoLog && glUseProgram && glDeleteProgram && glGetUniformLocation && glUniform1i && glUniform1f && glUniform3fv && glUniform4fv &&
			glBindAttribLocation && glVertexAttribPointer && glEnableVertexAttribArray && glDisableVertexAttribArray )
	{
		cout << "Using ARB_vertex_shader and ARB_fragment_shader" << endl;
	}
	else
	{
		cerr << "OpenGL does not support ARB_vertex_shader and ARB_fragment_shader" << endl;
		return false;
	}
	
	if ( _skeletonShader = LoadShader( "skeleton" ) )
	{
		_skeletonShader->Release();
	}
	else
	{
		cerr << "Required shader missing." << endl;
		return false;
	}
	
	if ( _infoMesh = LoadMesh( "Root/system/info.obj", "Root/system", 4 ) )
	{
		_infoMesh->Release();
	}
	else
	{
		cerr << "Required mesh missing." << endl;
		return false;
	}
	
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// Renderer prepare a new context

bool Renderer::PrepareContext()
{
	// Setup shading
	
	glClearColor( 0.8f, 0.8f, 1, 1 );				// Clear to black
	glShadeModel( GL_SMOOTH );						// Smooth shading
	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );	// Shade both sides
	glLineWidth( 2 );								// Thicker lines
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	
	// Setup Z buffer
	
	glDepthRange( 0, 1 );							// Map to full range 0..1
	glDepthMask( GL_TRUE );							// Enable depth writing
	glDepthFunc( GL_LESS );							// Less than function
	glEnable( GL_DEPTH_TEST );						// Enable depth testing
	
	// Setup face culling
	
	glFrontFace( GL_CW );							// Front is clockwise
	glCullFace( GL_BACK );							// Cull the back face
	glEnable( GL_CULL_FACE );						// Enable culling
	
	// Setup a default light
	
	scalar white[ 4 ]	= { 1, 1, 1, 1 };			// White
	scalar amb[ 4 ]		= { 0.1f, 0.1f, 0.1f, 1 };	// Gray
	
	glEnable( GL_LIGHTING );						// Enable lighting
	glEnable( GL_LIGHT0 );							// Enable one light
	glLightfv( GL_LIGHT0, GL_DIFFUSE, white );		// Set diffuse
	glLightfv( GL_LIGHT0, GL_SPECULAR, white );		// Set specular
	glLightModelfv( GL_LIGHT_MODEL_AMBIENT, amb );	// Ambient colour
	
	// Upload
	
	for ( list<Shader*>::iterator it = _shaders.begin() ; it != _shaders.end() ; ++ it )
	{
		if ( ! ( *it )->Upload() ) return false;
	}
	
	for ( list<Material*>::iterator it = _materials.begin() ; it != _materials.end() ; ++ it )
	{
		( *it )->Upload();
	}
	
	for ( list<Mesh*>::iterator it = _meshes.begin() ; it != _meshes.end() ; ++ it )
	{
		( *it )->Upload();
	}
	
	for ( list<Character*>::iterator it = _characters.begin() ; it != _characters.end() ; ++ it )
	{
		( *it )->Upload();
	}
	
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// Renderer get a procedure address in OpenGL

void* Renderer::GetGLProc( PCSTR name )
{
#ifdef WITH_SDL
	return SDL_GL_GetProcAddress( name );
#else
#ifdef _WIN32
	return (void*)wglGetProcAddress( name );
#else
#error implement_me
#endif
#endif
}

////////////////////////////////////////////////////////////////////////////////
// Renderer load and cache a shader

Shader* Renderer::LoadShader( PCSTR name )
{
	for ( list<Shader*>::iterator it = _shaders.begin() ; it != _shaders.end() ; ++ it )
	{
		if ( ( * it )->IsNamed( name ) )
		{
			( *it )->AddRef();
			return *it;
		}
	}
	
	Shader* shader = new Shader( name );
	
	if ( shader->Load() )
	{
		_shaders.push_back( shader );
		shader->AddRef();
		return shader;
	}
	else
	{
		shader->Release();
		return NULL;
	}
}

////////////////////////////////////////////////////////////////////////////////
// Renderer load and cache a material

Material* Renderer::LoadMaterial( PCSTR name )
{
	for ( list<Material*>::iterator it = _materials.begin() ; it != _materials.end() ; ++ it )
	{
		if ( ( * it )->IsNamed( name ) )
		{
			( *it )->AddRef();
			return *it;
		}
	}
	
	Material* material = new Material( name );
	
	if ( material->Load() )
	{
		_materials.push_back( material );
		material->AddRef();
		return material;
	}
	else
	{
		material->Release();
		return NULL;
	}
}

////////////////////////////////////////////////////////////////////////////////
// Renderer load and cache a mesh

Mesh* Renderer::LoadMesh( PCSTR name, PCSTR materialPath, scalar scale )
{
	for ( list<Mesh*>::iterator it = _meshes.begin() ; it != _meshes.end() ; ++ it )
	{
		if ( ( * it )->IsNamed( name ) )
		{
			( *it )->AddRef();
			return *it;
		}
	}
	
	Mesh* mesh = new Mesh( name, materialPath, scale );
	
	if ( mesh->Load() )
	{
		_meshes.push_back( mesh );
		mesh->AddRef();
		return mesh;
	}
	else
	{
		mesh->Release();
		return NULL;
	}
}

////////////////////////////////////////////////////////////////////////////////
// Renderer load and cache a character

Character* Renderer::LoadCharacter( PCSTR name )
{
	for ( list<Character*>::iterator it = _characters.begin() ; it != _characters.end() ; ++ it )
	{
		if ( ( * it )->IsNamed( name ) )
		{
			( *it )->AddRef();
			return *it;
		}
	}
	
	Character* character = new Character( name );
	
	if ( character->Load() )
	{
		_characters.push_back( character );
		character->AddRef();
		return character;
	}
	else
	{
		cerr << "Unable to load character model '" << name << "'." << endl;
		character->Release();
		return NULL;
	}
}

////////////////////////////////////////////////////////////////////////////////
// Renderer render a cylinder

void Renderer::RenderCylinder( scalar radius, scalar height )
{
	glBegin( GL_QUADS );
	
	vec3 bottom( 0, -1000, 0 ), top( 0, height, 0 );
	
	for ( int step = 0, steps = 64 ; step < steps ; step ++ )
	{
		scalar a0 = M_PI * 2.0f * (scalar)step / (scalar)steps;
		scalar a1 = M_PI * 2.0f * (scalar)( step + 1 ) / (scalar)steps;
		
		vec3 r0( cosf( a0 ) * radius, 0, sinf( a0 ) * radius );
		vec3 r1( cosf( a1 ) * radius, 0, sinf( a1 ) * radius );
		
		glVertex3fv( bottom + r0 );
		glVertex3fv( top + r0 );
		glVertex3fv( top + r1 );
		glVertex3fv( bottom + r1 );
	}
	
	glEnd();
}
