//
// Shader.cpp
// Copyright Michael J Stokes, 2008.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "Renderer.h"
#include "Shader.h"


///////////////////////////////////////////////////////////////////////////////
// Shader - construction

Shader::Shader( PCSTR name ) : Resource( name )
{
#ifndef WITH_VRJUGGLER
	_program = 0;
#endif
	
	_boneMatrix = -1;
}

///////////////////////////////////////////////////////////////////////////////
// Shader - destruction

Shader::~Shader()
{
}

///////////////////////////////////////////////////////////////////////////////
// Shader - load

bool Shader::Load()
{
	_vertex		= Load( _name + "_vertex" );
	_fragment	= Load( _name + "_fragment" );
	
	if ( ( ! _vertex.empty() ) || ( ! _fragment.empty() ) )
	{
		return true;
	}
	else
	{
		cerr << "Unable to load shader '" << _name << "'." << endl;
		return false;
	}
}

///////////////////////////////////////////////////////////////////////////////
// Shader - load shader source

string Shader::Load( const string & name )
{
	std::ifstream stream( ( "shaders/" + name ).c_str() );
	string source;
	
	if ( stream.is_open() )
	{
		for ( ; ; )
		{
			int c = stream.get();
			if ( 0 == c ) break;
			source += c;
		}
	}
	else
	{
		PCSTR iname = name.c_str();
		
		for ( PCSTR* internal = INTERNAL ; *internal ; internal += 2 )
		{
			if ( 0 == strcmp( *internal, iname ) )
			{
				source = internal[ 1 ];
				return source;
			}
		}
	}
	
	return source;
}

///////////////////////////////////////////////////////////////////////////////
// Shader - upload to graphics hardware

bool Shader::Upload()
{
	GLuint program = glCreateProgram();
	
	if ( 0 == program )
	{
		cerr << "Unable to create shader program object." << endl;
		return false;
	}
	
	vector<GLuint> shaders;
	
	if ( ! _vertex.empty() )
	{
		shaders.push_back( Compile( GL_VERTEX_SHADER, _vertex ) );
	}
	
	if ( ! _fragment.empty() )
	{
		shaders.push_back( Compile( GL_FRAGMENT_SHADER, _fragment ) );
	}
	
	for ( vector<GLuint>::iterator it = shaders.begin() ; it != shaders.end() ; ++ it )
	{
		if ( 0 == *it )
		{
			for ( it = shaders.begin() ; it != shaders.end() ; ++ it )
			{
				if ( 0 != *it ) glDeleteShader( *it );
			}
			
			glDeleteProgram( program );
			
			return false;
		}
	}
	
	for ( vector<GLuint>::iterator it = shaders.begin() ; it != shaders.end() ; ++ it )
	{
		glAttachShader( program, *it );
		glDeleteShader( *it );
	}
	
	glBindAttribLocation( program, BONE_INDEX_ATTRIB, "bone_index" );
	glBindAttribLocation( program, BONE_WEIGHT_ATTRIB, "bone_weight" );
	
	for ( int step = 0 ; step < 2 ; step ++ )
	{
		GLint status = 0;
		
		if ( 0 == step )
		{
			glLinkProgram( program );
			glGetProgramiv( program, GL_LINK_STATUS, &status );
		}
		else
		{
			glValidateProgram( program );
			glGetProgramiv( program, GL_VALIDATE_STATUS, &status );
		}
		
		if ( GL_TRUE != status )
		{
			GLsizei length = 0;
			glGetProgramiv( program, GL_INFO_LOG_LENGTH, &length );
			
			char* message = new char[ length + 1 ];
			glGetProgramInfoLog( program, length, &length, message );
			message[ length ] = 0;
			
			cerr << "An error occurred " << ( step ? "validating" : "linking" ) << " shader program " << _name.c_str() << ":" << endl;
			cerr << message << endl;
			
			delete [] message;
			
			glDeleteProgram( program );
			
			return false;
		}
	}
	
	GLint current = 0;
	glGetIntegerv( GL_CURRENT_PROGRAM, &current );
	glUseProgram( program );
	
	_boneMatrix = glGetUniformLocation( program, "bone_matrix" );
	
	int uniform = glGetUniformLocation( program, "texture0" );
	if ( uniform >= 0 ) glUniform1i( uniform, 0 );
	uniform = glGetUniformLocation( program, "texture1" );
	if ( uniform >= 0 ) glUniform1i( uniform, 1 );
	
	glUseProgram( current );
	
	cout << "Shader program '" << _name.c_str() << "' compiled and linked successfully." << endl;
	
#ifdef WITH_VRJUGGLER
	*_program = program;
#else
	_program = program;
#endif
	
	return true;
}

///////////////////////////////////////////////////////////////////////////////
// Shader - compile a shader

GLuint Shader::Compile( GLenum type, const string & source )
{
	PCSTR typeString = ( GL_VERTEX_SHADER == type ) ? "vertex" : "fragment";
	
	GLuint shader = glCreateShader( type );
	
	if ( 0 == shader )
	{
		cerr << "Unable to create " << typeString << " shader." << endl;
		return 0;
	}
	
	int len		= (int)source.length();
	PSTR ptr	= new char[ len + 16 ];
	
	memset( ptr, 0, len + 16 );
	memcpy( ptr, source.c_str(), len );
	
	PCSTR cptr = ptr;
	glShaderSource( shader, 1, &cptr, &len );
	
	delete [] ptr;
	
	glCompileShader( shader );
	
	GLint status = 0;
	glGetShaderiv( shader, GL_COMPILE_STATUS, &status );
	if ( GL_TRUE == status ) return shader;
	
	GLsizei length = 0;
	glGetShaderiv( shader, GL_INFO_LOG_LENGTH, &length );
	
	char* message = new char[ length + 1 ];
	glGetShaderInfoLog( shader, length, &length, message );
	message[ length ] = 0;
	
	cerr << "Error compiling " << typeString << " shader '" << _name.c_str() << "':" << endl;
	cerr << message;
	
	delete [] message;
	
	glDeleteShader( shader );
	
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
// Shader - erase from graphics hardware

void Shader::Erase()
{
#ifdef WITH_VRJUGGLER
	if ( 0 != *_program )
	{
		glDeleteProgram( *_program );
		*_program = 0;
	}
#else
	if ( 0 != _program )
	{
		glDeleteProgram( _program );
		_program = 0;
	}
#endif
}

///////////////////////////////////////////////////////////////////////////////
// Shader - begin rendering

void Shader::Begin()
{
#ifdef WITH_VRJUGGLER
	glUseProgram( *_program );
#else
	glUseProgram( _program );
#endif
}

///////////////////////////////////////////////////////////////////////////////
// Shader - end rendering

void Shader::End()
{
	glUseProgram( 0 );
}

///////////////////////////////////////////////////////////////////////////////
// Shader - find the index of a uniform variable

int Shader::FindUniform( PCSTR name )
{
#ifdef WITH_VRJUGGLER
	GLuint program = *_program;
#else
	GLuint program = _program;
#endif
	if ( 0 == program ) return -1;
	return (int)glGetUniformLocation( program, name );
}

///////////////////////////////////////////////////////////////////////////////
// Shader - set bone matrices

void Shader::SetBoneMatrices( const matrix4x4* transform, int count )
{
	if ( _boneMatrix >= 0 )
	{
		glUniformMatrix4fv( _boneMatrix, count, GL_FALSE, *transform );
	}
}

///////////////////////////////////////////////////////////////////////////////
// Shader - internal shaders

PCSTR Shader::INTERNAL[] =
{
	"skeleton_vertex",
		"uniform mat4	bone_matrix[ 32 ];\n"
		"attribute vec4	bone_index;\n"
		"attribute vec4	bone_weight;\n"
		"\n"
		"void main()\n"
		"{\n"
		"vec4 vertex	= bone_weight.x * bone_matrix[ int( bone_index.x ) ] * gl_Vertex\n"
		"				+ bone_weight.y * bone_matrix[ int( bone_index.y ) ] * gl_Vertex\n"
		"				+ bone_weight.z * bone_matrix[ int( bone_index.z ) ] * gl_Vertex\n"
		"				+ bone_weight.w * bone_matrix[ int( bone_index.w ) ] * gl_Vertex;\n"
		"\n"
		"vec3 normal	= bone_weight.x * mat3( bone_matrix[ int( bone_index.x ) ] ) * gl_Normal\n"
		"				+ bone_weight.y * mat3( bone_matrix[ int( bone_index.y ) ] ) * gl_Normal\n"
		"				+ bone_weight.z * mat3( bone_matrix[ int( bone_index.z ) ] ) * gl_Normal\n"
		"				+ bone_weight.w * mat3( bone_matrix[ int( bone_index.w ) ] ) * gl_Normal;\n"
		"\n"
		"gl_Position	= gl_ModelViewProjectionMatrix * vertex;\n"
		"gl_TexCoord[0]	= gl_MultiTexCoord0;\n"
		"gl_FrontColor	= gl_LightSource[ 0 ].diffuse * dot( gl_LightSource[ 0 ].position.xyz, gl_NormalMatrix * normal );\n"
		"}\n",
	
	"skeleton_fragment",
		"uniform sampler2D texture0;\n"
		"\n"
		"void main()\n"
		"{\n"
		"gl_FragColor = texture2D( texture0, gl_TexCoord[0].st ) /* * gl_Color */;\n"
		"}\n",
	
	NULL
};
