//
// SkeletonTranslator.cpp
// Copyright Michael J Stokes, 2008.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "SkeletonTranslator.h"

#define MAP_TO_X -v.x
#define MAP_TO_Y v.y
#define MAP_TO_Z v.z
#define MAP_TO_V1 "v3"
#define MAP_TO_V2 "v2"
#define MAP_TO_V3 "v1"


////////////////////////////////////////////////////////////////////////////////
// SkeletonTranslator - translate from OGRE to native

bool SkeletonTranslator::Translate( PCSTR name, scalar scale, PCSTR material )
{
	SkeletonTranslator st( name, scale, material );
	return st.DoSkeleton() && st.DoMesh();
}

////////////////////////////////////////////////////////////////////////////////
// SkeletonTranslator - do the skeleton

bool SkeletonTranslator::DoSkeleton()
{
	cout << "Translating skeleton '" << _name << "'..." << endl;
	
	string sourcePath( "characters/" );
	sourcePath += _name;
	sourcePath += "/source/";
	sourcePath += _name;
	sourcePath += "-Armature.skeleton.xml";
	
	XMLElement* source = XMLElement::FromFile( sourcePath.c_str() );
	XMLAutoDelete adSource( source );
	
	if ( ! source->IsNamed( "skeleton" ) )
	{
		cerr << "Unable to open source XML skeleton '" << sourcePath << "'." << endl;
		return false;
	}
	
	vector<XMLElement*> bones;
	
	if ( XMLElement* bonesContainer = source->GetElement( "bones" ) )
	{
		for ( XMLElementIterator it = bonesContainer->GetElements() ; it ; ++ it )
		{
			if ( it->IsNamed( "bone" ) )
			{
				if ( XMLElement* position = it->GetElement( "position" ) )
				{
					vec3 v;
					v.x = position->GetAttributeFloat( "x" );
					v.y = position->GetAttributeFloat( "y" );
					v.z = position->GetAttributeFloat( "z" );
					position->AddAttributeFloat( "x", MAP_TO_X * _scale );
					position->AddAttributeFloat( "y", MAP_TO_Y * _scale );
					position->AddAttributeFloat( "z", MAP_TO_Z * _scale );
				}
				
				if ( XMLElement* rotation = it->GetElement( "rotation" ) )
				{
					rotation->SetName( "orientation" );
					
					if ( XMLElement* axis = rotation->GetElement( "axis" ) )
					{
						vec3 v;
						v.x = axis->GetAttributeFloat( "x" );
						v.y = axis->GetAttributeFloat( "y" );
						v.z = axis->GetAttributeFloat( "z" );
						rotation->AddAttributeFloat( "x", MAP_TO_X );
						rotation->AddAttributeFloat( "y", MAP_TO_Y );
						rotation->AddAttributeFloat( "z", MAP_TO_Z );
						axis->Delete();
					}
				}
				
				UINT id = it->GetAttributeInt( "id" );
				if ( id >= _map.size() ) _map.resize( id + 1 );
				_map[ id ] = (UINT)bones.size();
				bones.push_back( *it );
				_bones.push_back( it->GetAttributeString( "name" ) );
				
				if ( XMLAttribute* attr = it->GetAttribute( "id" ) ) attr->Delete();
			}
		}
	}
	
	if ( bones.empty() )
	{
		cerr << "Source XML skeleton '" << sourcePath << "' has no bones." << endl;
		return false;
	}
	
	if ( XMLElement* hierarchy = source->GetElement( "bonehierarchy" ) )
	{
		for ( XMLElementIterator it = hierarchy->GetElements() ; it ; ++ it )
		{
			if ( it->IsNamed( "boneparent" ) )
			{
				PCSTR name		= it->GetAttributeString( "bone" );
				PCSTR parent	= it->GetAttributeString( "parent" );
				
				for ( vector<XMLElement*>::iterator bit = bones.begin() ; bit != bones.end() ; ++ bit )
				{
					if ( 0 == strcmp( ( *bit )->GetAttributeString( "name" ), name ) )
					{
						( *bit )->AddAttribute( "parent", parent );
						break;
					}
				}
			}
		}
	}
	
	XMLElement target( "skeleton" );
	target.AddAttribute( "xmlns", "urn:no.ntnu.idi.vgs:character-skeleton:1.0" );
	
	target.AddAttributeFloat( "scale", _scale );
	
	if ( XMLElement* boneContainer = target.AddElement( "bones" ) )
	{
		for ( vector<XMLElement*>::iterator bit = bones.begin() ; bit != bones.end() ; ++ bit )
		{
			XMLElement* bone = *bit;
			bone->Detach();
			boneContainer->Add( bone );
		}
	}
	
	string targetPath( "characters/" );
	targetPath += _name;
	targetPath += "/skeleton.xml";
	
	if ( ! target.ToFile( targetPath.c_str(), true ) )
	{
		cerr << "Unable to open target XML skeleton '" << targetPath << "'." << endl;
		return false;
	}
	
	if ( XMLElement* animations = source->GetElement( "animations" ) )
	{
		for ( XMLElementIterator ait = animations->GetElements() ; ait ; ++ ait )
		{
			if ( ait->IsNamed( "animation" ) )
			{
				if ( ! DoAnimation( *ait ) ) return false;
			}
		}
	}
	
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// SkeletonTranslator - do the mesh

bool SkeletonTranslator::DoMesh()
{
	string sourcePath( "characters/" );
	sourcePath += _name;
	sourcePath += "/source/";
	sourcePath += _name;
	sourcePath += ".mesh.xml";
	
	XMLElement* source = XMLElement::FromFile( sourcePath.c_str() );
	XMLAutoDelete adSource( source );
	
	XMLElement* sub = source->IsNamed( "mesh" ) ? source->GetElement( "submeshes" ) : NULL;
	
	if ( NULL == sub )
	{
		cerr << "Unable to open source XML mesh '" << sourcePath << "'." << endl;
		return false;
	}
	
	XMLElement target( "mesh" );
	target.AddAttribute( "xmlns", "urn:no.ntnu.idi.vgs:character-mesh:1.0" );
	
	if ( XMLElement* bones = target.AddElement( "bones" ) )
	{
		for ( vector<UINT>::iterator it = _map.begin() ; it != _map.end() ; ++ it )
		{
			bones->AddElement( "bone" )->AddAttribute( "name", _bones[ *it ].c_str() );
		}
	}
	
	XMLElement* vertices	= target.AddElement( "vertices" );
	XMLElement* groups		= target.AddElement( "groups" );
	
	UINT vcount = 0;
	
	if ( XMLElement* geometry = source->GetElement( "sharedgeometry" ) )
	{
		vector<XMLElement*> vref;
		
		for ( XMLElementIterator vbi = geometry->GetElements() ; vbi ; ++ vbi )
		{
			if ( vbi->IsNamed( "vertexbuffer" ) )
			{
				if ( vbi->GetAttribute( "positions" ) )
				{
					for ( XMLElementIterator vi = vbi->GetElements() ; vi ; ++ vi )
					{
						if ( vi->IsNamed( "vertex" ) )
						{
							if ( XMLElement* position = vi->GetElement( "position" ) )
							{
								vec3 v;
								v.x = position->GetAttributeFloat( "x" );
								v.y = position->GetAttributeFloat( "y" );
								v.z = position->GetAttributeFloat( "z" );
								position->AddAttributeFloat( "x", MAP_TO_X * _scale );
								position->AddAttributeFloat( "y", MAP_TO_Y * _scale );
								position->AddAttributeFloat( "z", MAP_TO_Z * _scale );
							}
							
							if ( XMLElement* normal = vi->GetElement( "normal" ) )
							{
								vec3 v;
								v.x = normal->GetAttributeFloat( "x" );
								v.y = normal->GetAttributeFloat( "y" );
								v.z = normal->GetAttributeFloat( "z" );
								normal->AddAttributeFloat( "x", MAP_TO_X );
								normal->AddAttributeFloat( "y", MAP_TO_Y );
								normal->AddAttributeFloat( "z", MAP_TO_Z );
							}
							
							if ( XMLElement* tc = vi->GetElement( "texcoord" ) )
							{
								tc->SetName( "material" );
								if ( XMLAttribute* u = tc->GetAttribute( "u" ) ) u->SetName( "s" );
								if ( XMLAttribute* v = tc->GetAttribute( "v" ) ) v->SetName( "t" );
							}
							
							vref.push_back( *vi );
						}
					}
				}
				else
				{
					UINT i = 0;
					
					for ( XMLElementIterator vi = vbi->GetElements() ; vi ; ++ vi )
					{
						if ( vi->IsNamed( "vertex" ) )
						{
							for ( XMLElementIterator tci = vi->GetElements() ; tci ; ++ tci )
							{
								if ( tci->IsNamed( "texcoord" ) )
								{
									XMLElement* ntc = vref[ i ]->AddElement( "material" );
									if ( XMLAttribute* u = tci->GetAttribute( "u" ) ) ntc->AddAttribute( "s", u->GetValue().c_str() );
									if ( XMLAttribute* v = tci->GetAttribute( "v" ) ) ntc->AddAttribute( "t", v->GetValue().c_str() );
									break;
								}
							}
							
							i ++;
						}
					}
				}
			}
		}
		
		if ( XMLElement* bones = source->GetElement( "boneassignments" ) )
		{
			for ( XMLElementIterator vbai = bones->GetElements() ; vbai ; ++ vbai )
			{
				if ( vbai->IsNamed( "vertexboneassignment" ) )
				{
					XMLElement* bone = vref[ vbai->GetAttributeInt( "vertexindex" ) ]->AddElement( "bone" );
					bone->AddAttribute( "id", vbai->GetAttributeString( "boneindex" ) );
					bone->AddAttribute( "weight", vbai->GetAttributeString( "weight" ) );
				}
			}
		}
		
		for ( vector<XMLElement*>::iterator it = vref.begin() ; it != vref.end() ; ++ it )
		{
			XMLElement* vertex = *it;
			vertex->Detach();
			vertices->Add( vertex );
		}
	}
	
	for ( XMLElementIterator smi = sub->GetElements() ; smi ; ++ smi )
	{
		vector<XMLElement*> vref;
		
		XMLElement* group = groups->AddElement( "group" );
		
		if ( _material.empty() )
		{
			group->AddAttribute( "material", smi->GetAttributeString( "material" ) );
		}
		else if ( _material == "!" )
		{
			group->AddAttribute( "material", ( _name + ".png" ).c_str() );
		}
		else
		{
			group->AddAttribute( "material", _material.c_str() );
		}
		
		if ( XMLElement* geometry = smi->GetElement( "geometry" ) )
		{
			for ( XMLElementIterator vbi = geometry->GetElements() ; vbi ; ++ vbi )
			{
				if ( vbi->IsNamed( "vertexbuffer" ) )
				{
					for ( XMLElementIterator vi = vbi->GetElements() ; vi ; ++ vi )
					{
						if ( vi->IsNamed( "vertex" ) )
						{
							if ( XMLElement* position = vi->GetElement( "position" ) )
							{
								vec3 v;
								v.x = position->GetAttributeFloat( "x" );
								v.y = position->GetAttributeFloat( "y" );
								v.z = position->GetAttributeFloat( "z" );
								position->AddAttributeFloat( "x", MAP_TO_X * _scale );
								position->AddAttributeFloat( "y", MAP_TO_Y * _scale );
								position->AddAttributeFloat( "z", MAP_TO_Z * _scale );
							}
							
							if ( XMLElement* normal = vi->GetElement( "normal" ) )
							{
								vec3 v;
								v.x = normal->GetAttributeFloat( "x" );
								v.y = normal->GetAttributeFloat( "y" );
								v.z = normal->GetAttributeFloat( "z" );
								normal->AddAttributeFloat( "x", MAP_TO_X );
								normal->AddAttributeFloat( "y", MAP_TO_Y );
								normal->AddAttributeFloat( "z", MAP_TO_Z );
							}
							
							if ( XMLElement* tc = vi->GetElement( "texcoord" ) )
							{
								tc->SetName( "material" );
								if ( XMLAttribute* u = tc->GetAttribute( "u" ) ) u->SetName( "s" );
								if ( XMLAttribute* v = tc->GetAttribute( "v" ) ) v->SetName( "t" );
							}
							
							vref.push_back( *vi );
						}
					}
				}
			}
		}
		
		if ( XMLElement* bones = smi->GetElement( "boneassignments" ) )
		{
			for ( XMLElementIterator vbai = bones->GetElements() ; vbai ; ++ vbai )
			{
				if ( vbai->IsNamed( "vertexboneassignment" ) )
				{
					XMLElement* bone = vref[ vbai->GetAttributeInt( "vertexindex" ) ]->AddElement( "bone" );
					bone->AddAttribute( "id", vbai->GetAttributeString( "boneindex" ) );
					bone->AddAttribute( "weight", vbai->GetAttributeString( "weight" ) );
				}
			}
		}
		
		if ( XMLElement* faces = smi->GetElement( "faces" ) )
		{
			for ( XMLElementIterator fi = faces->GetElements() ; fi ; ++ fi )
			{
				if ( fi->IsNamed( "face" ) )
				{
					UINT v1 = vcount + fi->GetAttributeInt( MAP_TO_V1 );
					UINT v2 = vcount + fi->GetAttributeInt( MAP_TO_V2 );
					UINT v3 = vcount + fi->GetAttributeInt( MAP_TO_V3 );
					
					XMLElement* face = group->AddElement( "face" );
					char temp[ 32 ];
					
					sprintf( temp, "%d", v3 );
					face->AddAttribute( "v3", temp );
					sprintf( temp, "%d", v2 );
					face->AddAttribute( "v2", temp );
					sprintf( temp, "%d", v1 );
					face->AddAttribute( "v1", temp );
				}
			}
		}
		
		for ( vector<XMLElement*>::iterator it = vref.begin() ; it != vref.end() ; ++ it )
		{
			XMLElement* vertex = *it;
			vertex->Detach();
			vertices->Add( vertex );
			vcount ++;
		}
	}
	
	string targetPath( "characters/" );
	targetPath += _name;
	targetPath += "/mesh.xml";
	
	if ( ! target.ToFile( targetPath.c_str(), true ) )
	{
		cerr << "Unable to open target XML mesh '" << targetPath << "'." << endl;
		return false;
	}
	
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// SkeletonTranslator - do an animation

bool SkeletonTranslator::DoAnimation( XMLElement* source )
{
	XMLElement target( "animation" );
	target.AddAttribute( "xmlns", "urn:no.ntnu.idi.vgs:character-animation:1.0" );
	
	XMLElement* bones	= target.AddElement( "bones" );
	XMLElement* frames	= target.AddElement( "frames" );
	
	PCSTR name = source->GetAttributeString( "name" );
	if ( NULL == name ) return false;
	
	XMLElement* tracks = source->GetElement( "tracks" );
	if ( NULL == tracks ) return false;
	
	for ( XMLElementIterator ti = tracks->GetElements() ; ti ; ++ ti )
	{
		if ( ti->IsNamed( "track" ) )
		{
			PCSTR bname = ti->GetAttributeString( "bone" );
			if ( NULL == bname ) return false;
			bones->AddElement( "bone" )->AddAttribute( "name", bname );
			
			XMLElement* keyframes = ti->GetElement( "keyframes" );
			if ( NULL == keyframes ) return false;
			
			for ( XMLElementIterator ki = keyframes->GetElements() ; ki ; ++ ki )
			{
				if ( ki->IsNamed( "keyframe" ) )
				{
					PCSTR time = ki->GetAttributeString( "time" );
					if ( NULL == time ) return false;
					
					XMLElement* frame = NULL;
					
					for ( XMLElementIterator fi = frames->GetElements() ; fi ; ++ fi )
					{
						if ( 0 == strcmp( fi->GetAttributeString( "time" ), time ) )
						{
							frame = *fi;
							break;
						}
					}
					
					if ( NULL == frame )
					{
						frame = frames->AddElement( "frame" );
						frame->AddAttribute( "time", time );
						
						while ( frame->GetChildCount() < bones->GetChildCount() )
						{
							frame->AddElement( "bone" );
						}
					}
					
					XMLElement* bone = frame->AddElement( "bone" );
					
					if ( XMLElement* translate = ki->GetElement( "translate" ) )
					{
						translate->Detach();
						bone->Add( translate );
						vec3 v;
						v.x = translate->GetAttributeFloat( "x" );
						v.y = translate->GetAttributeFloat( "y" );
						v.z = translate->GetAttributeFloat( "z" );
						translate->AddAttributeFloat( "x", MAP_TO_X * _scale );
						translate->AddAttributeFloat( "y", MAP_TO_Y * _scale );
						translate->AddAttributeFloat( "z", MAP_TO_Z * _scale );
					}
					
					if ( XMLElement* rotate = ki->GetElement( "rotate" ) )
					{
						rotate->Detach();
						bone->Add( rotate );
						
						if ( XMLElement* axis = rotate->GetElement( "axis" ) )
						{
							vec3 v;
							v.x = axis->GetAttributeFloat( "x" );
							v.y = axis->GetAttributeFloat( "y" );
							v.z = axis->GetAttributeFloat( "z" );
							rotate->AddAttributeFloat( "x", MAP_TO_X );
							rotate->AddAttributeFloat( "y", MAP_TO_Y );
							rotate->AddAttributeFloat( "z", MAP_TO_Z );
							axis->Delete();
						}
					}
				}
			}
			
			for ( XMLElementIterator fi = frames->GetElements() ; fi ; ++ fi )
			{
				while ( fi->GetChildCount() < bones->GetChildCount() + 1 )
				{
					fi->AddElement( "bone" );
				}
			}
		}
	}
	
	string targetPath( "characters/" );
	targetPath += _name;
	targetPath += "/animations/";
	targetPath += name;
	targetPath += ".xml";
	
	if ( ! target.ToFile( targetPath.c_str(), true ) )
	{
		cerr << "Unable to open target XML animation '" << targetPath << "'." << endl;
		return false;
	}
	
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// SkeletonTranslator - convert OGRE XML to SKM

bool SkeletonTranslator::SKM( PCSTR input )
{
	string base( input );
	
	XMLElement* source = XMLElement::FromFile( ( base + "/Cube-Armature.skeleton.xml" ).c_str() );
	if ( NULL == source ) return false;
	
	XMLAutoDelete adSource( source );
	
	XMLElement* animations = source->GetElement( "animations" );
	if ( NULL == animations ) return false;
	XMLElement* animation = animations->GetElement( "animation" );
	if ( NULL == animation ) return false;
	XMLElement* tracks = animation->GetElement( "tracks" );
	if ( NULL == tracks ) return false;
	XMLElement* track = tracks->GetElement( "track" );
	if ( NULL == track ) return false;
	XMLElement* keyframes = track->GetElement( "keyframes" );
	if ( NULL == keyframes ) return false;
	
	FILE* output = fopen( ( base + "/Walk.skm" ).c_str(), "wb" );
	if ( NULL == output ) return false;
	
	fprintf( output, "SkMotion\r\n\r\nname \"%s\"\r\n\r\n", animation->GetAttributeString( "name" ) );
	
	int count = 0;
	
	fprintf( output, "channels %i\r\n", tracks->GetChildCount() + 2 );
	
	for ( XMLElementIterator itt = tracks->GetElements() ; itt ; ++ itt, count ++ )
	{
		PCSTR name = itt->GetAttributeString( "bone" );
		
		if ( 0 == count )
		{
			fprintf( output, "%s XPos\r\n%s YPos\r\n%s ZPos\r\n", name, name, name );
		}
		else
		{
			fprintf( output, "%s Quat\r\n", name );
		}
	}
	
	XMLElementIterator kfi = keyframes->GetElements();
	int frames = keyframes->GetChildCount();
	
	fprintf( output, "\r\nframes %i\r\n", frames );
	
	for ( int frame = 0 ; frame < frames ; frame ++, ++ kfi )
	{
		printf( "frame %i\n", frame );
		fprintf( output, "kt %f fr", kfi->GetAttributeFloat( "time" ) );
		
		count = 0;
		
		for ( XMLElementIterator itt = tracks->GetElements() ; itt ; ++ itt, count ++ )
		{
			XMLElement* track = *itt;
			keyframes = track->GetElement( "keyframes" );
			
			XMLElementIterator itk = keyframes->GetElements();
			for ( int i = 0 ; itk && i < frame ; ++ itk, i ++ );
			
			if ( ! itk )
			{
				printf( "Unable to find keyframe %i in bone track '%s'.\n", frame, track->GetAttributeString( "bone" ) );
				fclose( output );
				return false;
			}
			
			if ( 0 == count )
			{
				if ( XMLElement* translate = itk->GetElement( "translate" ) )
				{
					vec3 v( translate->GetAttributeFloat( "x" ), translate->GetAttributeFloat( "z" ), translate->GetAttributeFloat( "y" ) );
					v *= 0.1f;
					fprintf( output, " %f %f %f", v.x, v.y, v.z );
				}
				else
				{
					printf( "Missing translation in frame %i in bone track '%s'.\n", frame, track->GetAttributeString( "bone" ) );
					fclose( output );
					return false;
				}
			}
			else
			{
				if ( XMLElement* rotate = itk->GetElement( "rotate" ) )
				{
					XMLElement* ax = rotate->GetElement( "axis" );
					
					vec3 axis( ax->GetAttributeFloat( "x" ), ax->GetAttributeFloat( "y" ), ax->GetAttributeFloat( "z" ) );
					axis *= rotate->GetAttributeFloat( "angle" );
					
					fprintf( output, " %f %f %f", axis.x, axis.y, axis.z );
				}
				else
				{
					printf( "Missing rotation in frame %i in bone track '%s'.\n", frame, track->GetAttributeString( "bone" ) );
					fclose( output );
					return false;
				}
			}
		}
		
		fprintf( output, "\r\n" );
	}
	
	fclose( output );
	
	return true;
}
