//
// Sound.cpp
// Copyright Michael J Stokes, 2008.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "Sound.h"
#include "stb_vorbis.h"


///////////////////////////////////////////////////////////////////////////////
// Sound - construction

Sound::Sound( PCSTR name )
{
	_name		= name;
	_duration	= 0;
	_samples	= NULL;
	_vorbis		= false;
}

///////////////////////////////////////////////////////////////////////////////
// Sound - destruction

Sound::~Sound()
{
	if ( NULL != _samples ) delete [] _samples;
}

///////////////////////////////////////////////////////////////////////////////
// Sound - load

bool Sound::Load()
{
	if ( _name.find( ".wav" ) != string::npos )
	{
		return LoadWave();
	}
	else if ( _name.find( ".oga" ) != string::npos )
	{
		return LoadVorbis();
	}
	else if ( _name.find( ".ogg" ) != string::npos )
	{
		return LoadVorbis();
	}
	else
	{
		cerr << "Unknown sound file type for sound '" << _name << "'." << endl;
		return false;
	}
}

///////////////////////////////////////////////////////////////////////////////
// Sound - create instance

Sound::Instance* Sound::CreateInstance()
{
	if ( _vorbis )
	{
		return new VorbisInstance( this );
	}
	else
	{
		return new WaveInstance( this );
	}
}

///////////////////////////////////////////////////////////////////////////////
// Sound - load a WAVE file

#pragma pack(2)
typedef struct
{
	UINT16	nFormatTag;
	UINT16	nChannels;
	UINT32	nSamplesPerSec;
	UINT32	nAvgBytesPerSec;
	UINT16	nBlockAlign;
	UINT16	nBitsPerSample;
} WAVE_FMT;
#pragma pack()

bool Sound::LoadWave()
{
	std::ifstream stream( _name.c_str(), std::ios_base::binary );
	if ( ! stream.is_open() ) return false;
	
	BYTE tag[ 4 ];
	
	memset( tag, 0, 4 );
	stream.read( (PSTR)&tag, 4 );
	if ( 0 != memcmp( tag, "RIFF", 4 ) ) return false;
	stream.ignore( 4 );
	memset( tag, 0, 4 );
	stream.read( (PSTR)&tag, 4 );
	if ( 0 != memcmp( tag, "WAVE", 4 ) ) return false;
	
	bool haveFormat = false;
	WAVE_FMT format;
	
	while ( ! stream.eof() )
	{
		UINT32 size = 0;
		stream.read( (PSTR)&tag, 4 );
		stream.read( (PSTR)&size, 4 );
		if ( stream.eof() ) break;
		
		std::streampos pos = stream.tellg();
		
		if ( ( 0 == memcmp( tag, "fmt ", 4 ) ) && ( size >= 16 ) )
		{
			if ( haveFormat ) return false;
			stream.read( (PSTR)&format, 16 );
			if ( stream.eof() ) return false;
			
			if ( format.nFormatTag != 1 )
			{
				cerr << "Sound '" << _name << "' is not PCM" << endl;
				return false;
			}
			else if ( format.nBitsPerSample != 16 )
			{
				cerr << "Sound '" << _name << "' is not 16-bit" << endl;
				return false;
			}
			else if ( format.nSamplesPerSec != 44100 )
			{
				cerr << "Sound '" << _name << "' is not 44100 Hz" << endl;
				return false;
			}
			else if ( ( format.nChannels != 1 ) && ( format.nChannels != 2 ) )
			{
				cerr << "Sound '" << _name << "' is not mono or stereo" << endl;;
				return false;
			}
			else if ( ( format.nBlockAlign != ( 2 * format.nChannels ) ) || ( format.nAvgBytesPerSec != format.nBlockAlign * format.nSamplesPerSec ) )
			{
				cerr << "Sound '" << _name << "' is aligned in a non-standard manner" << endl;
				return false;
			}
			
			haveFormat = true;
		}
		else if ( 0 == memcmp( tag, "data", 4 ) )
		{
			if ( ! haveFormat ) return false;
			
			if ( 2 == format.nChannels )
			{
				_duration	= ( size / 4 );
				_samples	= new short[ _duration * 2 ];
				stream.read( (PSTR)_samples, _duration * 4 );
				if ( stream.eof() ) return false;
			}
			else
			{
				_duration	= ( size / 2 );
				_samples	= new short[ _duration * 2 ];
				
				short* sample = _samples;
				
				for ( UINT count = _duration ; count -- ; )
				{
					short v;
					stream.read( (PSTR)&v, sizeof( v ) );
					if ( stream.eof() ) return false;
					
					*sample ++		= v;
					*sample ++		= v;
				}
			}
			
			break;
		}
		
		if ( size & 1 ) size ++;
		pos += (std::streampos)size;
		stream.seekg( pos );
		if ( stream.tellg() != pos ) break;
	}
	
	return ( NULL != _samples );
}

///////////////////////////////////////////////////////////////////////////////
// Sound - load a Vorbis file

bool Sound::LoadVorbis()
{
	int err;
	
	if ( stb_vorbis* vorbis = stb_vorbis_open_filename( (char*)_name.c_str(), &err, NULL ) )
	{
		stb_vorbis_info info = stb_vorbis_get_info( vorbis );
		stb_vorbis_close( vorbis );
		
		if ( info.sample_rate != 44100 )
		{
			cerr << "Sound '" << _name << "' is not 44100 Hz" << endl;
		}
		else
		{
			_duration	= info.max_frame_size;
			_samples	= new short[ _duration * 4 ];
			_vorbis		= true;
			return true;
		}
	}
	
	return false;
}

///////////////////////////////////////////////////////////////////////////////
// Sound::Instance - construction

Sound::Instance::Instance( Sound* sound )
{
	_sound = sound;
	_sound->AddRef();
}

///////////////////////////////////////////////////////////////////////////////
// Sound::Instance - destruction

Sound::Instance::~Instance()
{
	_sound->Release();
}

///////////////////////////////////////////////////////////////////////////////
// Sound::WaveInstance - construction

Sound::WaveInstance::WaveInstance( Sound* sound ) : Instance( sound )
{
	_position = 0;
}

///////////////////////////////////////////////////////////////////////////////
// Sound::WaveInstance - mix

bool Sound::WaveInstance::Mix( int* output, UINT samples, int left, int right, bool loop )
{
	UINT mix = _sound->_duration - _position;
	if ( mix > samples ) mix = samples;
	
	if ( mix > 0 )
	{
		short* input = _sound->_samples + ( _position << 1 );
		
		_position += mix;
		samples -= mix;
		
		while ( mix -- )
		{
			*( output ++ ) += ( (int)( *input ++ ) * left  ) >> 7;
			*( output ++ ) += ( (int)( *input ++ ) * right ) >> 7;
		}
	}
	
	if ( _position < _sound->_duration ) return true;
	else if ( ! loop ) return false;
	
	_position = 0;
	mix = _sound->_duration;
	if ( mix > samples ) mix = samples;
	
	if ( mix > 0 )
	{
		short* input = _sound->_samples + ( _position << 1 );
		
		_position += mix;
		samples -= mix;
		
		while ( mix -- )
		{
			*( output ++ ) += ( (int)( *input ++ ) * left  ) >> 7;
			*( output ++ ) += ( (int)( *input ++ ) * right ) >> 7;
		}
	}
	
	return true;
}

///////////////////////////////////////////////////////////////////////////////
// Sound::VorbisInstance - construction

Sound::VorbisInstance::VorbisInstance( Sound* sound ) : Instance( sound )
{
	int err;
	stb_vorbis* vorbis = stb_vorbis_open_filename( (char*)sound->_name.c_str(), &err, NULL );
	_handle = vorbis;
}

///////////////////////////////////////////////////////////////////////////////
// Sound::VorbisInstance - destruction

Sound::VorbisInstance::~VorbisInstance()
{
	if ( stb_vorbis* vorbis = (stb_vorbis*)_handle ) stb_vorbis_close( vorbis );
}

///////////////////////////////////////////////////////////////////////////////
// Sound::VorbisInstance - mix

bool Sound::VorbisInstance::Mix( int* output, UINT samples, int left, int right, bool loop )
{
	stb_vorbis* vorbis = (stb_vorbis*)_handle;
	if ( NULL == vorbis ) return false;
	
	while ( samples > 0 )
	{
		short* input = _sound->_samples;
		int got = stb_vorbis_get_samples_short_interleaved( vorbis, 2, input, (int)samples * 2 );
		if ( got <= 0 ) return false;
		
		samples -= (UINT)got;
		
		while ( got -- )
		{
			*( output ++ ) += ( (int)( *input ++ ) * left  ) >> 7;
			*( output ++ ) += ( (int)( *input ++ ) * right ) >> 7;
		}
	}
	
	return true;
}
