//
// WinAudio.cpp
// Copyright Michael J Stokes, 2009.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "WinAudio.h"

#ifdef _WIN32

#include <process.h>

const IID IID_IDirectSoundNotify = { 0xB0210783, 0x89cd, 0x11d0, { 0xAF, 0x08, 0x00, 0xA0, 0xC9, 0x25, 0xCD, 0x16 } };

#define BUFFER_SIZE 2048
#include <dshow.h>


////////////////////////////////////////////////////////////////////////////////
// WinAudio - construction

WinAudio::WinAudio()
{
	_ds		= NULL;
	_dsb	= NULL;
	_thread	= false;
	_buffer	= NULL;
	
	InitializeCriticalSection( &_section );
}

////////////////////////////////////////////////////////////////////////////////
// WinAudio - destruction

WinAudio::~WinAudio()
{
	Terminate();
	DeleteCriticalSection( &_section );
}

////////////////////////////////////////////////////////////////////////////////
// WinAudio - initialize

bool WinAudio::Initialize()
{
	if ( NULL != _ds ) return true;
	
	IDirectSound* ds = NULL;
	DirectSoundCreate( NULL, &ds, NULL );
	
	if ( NULL == ds )
	{
		cerr << "Unable to initialize DirectSound." << endl;
		return false;
	}
	
	if ( FAILED( ds->SetCooperativeLevel( GetActiveWindow(), DSSCL_PRIORITY ) ) )
	{
		ds->Release();
		cerr << "Unable to set DirectSound cooperative level." << endl;
		return false;
	}
	
	WAVEFORMATEX wfx;
	memset( &wfx, 0, sizeof( wfx ) );
	
	wfx.wFormatTag		= WAVE_FORMAT_PCM;
	wfx.nChannels		= 2;
	wfx.nSamplesPerSec	= 44100;
	wfx.wBitsPerSample	= 16;
	wfx.nBlockAlign		= wfx.nChannels * wfx.wBitsPerSample / 8;
	wfx.nAvgBytesPerSec	= wfx.nSamplesPerSec * wfx.nBlockAlign;
	
	DSBUFFERDESC dsbd;
	memset( &dsbd, 0, sizeof( dsbd ) );
	
	dsbd.dwSize			= sizeof( dsbd );
	dsbd.dwFlags		= DSBCAPS_CTRLPOSITIONNOTIFY | DSBCAPS_GLOBALFOCUS;
	dsbd.dwBufferBytes	= 3 * BUFFER_SIZE * 4;
	dsbd.lpwfxFormat	= &wfx;
	
	IDirectSoundBuffer* dsb = NULL;
	ds->CreateSoundBuffer( &dsbd, &dsb, NULL );
	
	if ( NULL == dsb )
	{
		ds->Release();
		cerr << "Unable to create DirectSound buffer." << endl;
		return false;
	}
	
	IDirectSoundNotify* dsn = NULL;
	dsb->QueryInterface( IID_IDirectSoundNotify, (LPVOID*)&dsn );
	
	if ( NULL == dsn )
	{
		dsb->Release();
		ds->Release();
		cerr << "Unable to QI DirectSoundNotify." << endl;
		return false;
	}
	
	_events[ 0 ] = CreateEvent( NULL, true, false, NULL );
	_events[ 1 ] = CreateEvent( NULL, false, false, NULL );
	_events[ 2 ] = CreateEvent( NULL, false, false, NULL );
	_events[ 3 ] = CreateEvent( NULL, false, false, NULL );
	_events[ 4 ] = CreateEvent( NULL, true, true, NULL );
	
	DSBPOSITIONNOTIFY positions[ 3 ] =
	{
		{ 2 * BUFFER_SIZE * 4, _events[ 1 ] },
		{ 0 * BUFFER_SIZE * 4, _events[ 2 ] },
		{ 1 * BUFFER_SIZE * 4, _events[ 3 ] }
	};
	
	dsn->SetNotificationPositions( 3, positions );
	dsn->Release();
	
	_ds		= ds;
	_dsb	= dsb;
	
	_buffer	= new int[ BUFFER_SIZE * 2 ];
	
	return Audio::Initialize();
}

////////////////////////////////////////////////////////////////////////////////
// WinAudio - terminate

void WinAudio::Terminate()
{
	if ( NULL != _ds )
	{
		Pause( true );
		
		CloseHandle( _events[ 4 ] );
		CloseHandle( _events[ 3 ] );
		CloseHandle( _events[ 2 ] );
		CloseHandle( _events[ 1 ] );
		CloseHandle( _events[ 0 ] );
		
		_dsb->Release();
		_dsb = NULL;
		
		_ds->Release();
		_ds = NULL;
		
		delete [] _buffer;
		_buffer = NULL;
		
		Audio::Terminate();
	}
}

////////////////////////////////////////////////////////////////////////////////
// WinAudio - lock

void WinAudio::Lock()
{
	if ( NULL != _ds ) EnterCriticalSection( &_section );
}

////////////////////////////////////////////////////////////////////////////////
// WinAudio - unlock

void WinAudio::Unlock()
{
	if ( NULL != _ds ) LeaveCriticalSection( &_section );
}

////////////////////////////////////////////////////////////////////////////////
// WinAudio - pause and resume

void WinAudio::Pause( bool pause )
{
	if ( NULL != _ds )
	{
		if ( pause )
		{
			if ( _thread )
			{
				SetEvent( _events[ 0 ] );
				_dsb->Stop();
				WaitForSingleObject( _events[ 4 ], INFINITE );
				_thread = false;
			}
		}
		else
		{
			if ( ! _thread )
			{
				ResetEvent( _events[ 0 ] );
				ResetEvent( _events[ 4 ] );
				_thread = _start = true;
				_beginthread( ThreadEntry, 0, this );
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
// WinAudio - thread entry

void WinAudio::ThreadEntry( void* self )
{
	( (WinAudio*)self )->OnThread();
}

////////////////////////////////////////////////////////////////////////////////
// WinAudio - thread

void WinAudio::OnThread()
{
	SetThreadPriority( GetCurrentThread(), THREAD_PRIORITY_ABOVE_NORMAL );
	
	for ( ; ; )
	{
		DWORD target = _start ? 1 : WaitForMultipleObjects( 4, _events, false, INFINITE );
		if ( 0 == target ) break;
		
		target = ( target - 1 ) * BUFFER_SIZE * 4;
		
		Lock();
		Process( _buffer, BUFFER_SIZE );
		Unlock();
		
		LPVOID target1, target2;
		DWORD length1 = 0, length2 = 0;
		
		HRESULT hr = _dsb->Lock( target, BUFFER_SIZE * 4, &target1, &length1, &target2, &length2, 0 );
		
		if ( DSERR_BUFFERLOST == hr )
		{
			if ( FAILED( _dsb->Restore() ) )
			{
				Sleep( 100 );
				continue;
			}
			
			hr = _dsb->Lock( target, BUFFER_SIZE * 4, &target1, &length1, &target2, &length2, 0 );
		}
		
		if ( FAILED( hr ) )
		{
			Sleep( 100 );
			continue;
		}
		
		int*	input	= _buffer;
		short*	output	= (short*)target1;
		
		for ( UINT count = length1 / 2 ; count -- ; )
		{
			int sample = *input ++;
			
			if ( sample < -32760 )
				*output ++ = -32760;
			else if ( sample > 32760 )
				*output ++ = 32760;
			else
				*output ++ = (short)sample;
		}
		
		output	= (short*)target2;
		
		for ( UINT count = length2 / 2 ; count -- ; )
		{
			int sample = *input ++;
			
			if ( sample < -32760 )
				*output ++ = -32760;
			else if ( sample > 32760 )
				*output ++ = 32760;
			else
				*output ++ = (short)sample;
		}
		
		_dsb->Unlock( target1, length1, target2, length2 );
		
		if ( _start )
		{
			_dsb->SetCurrentPosition( 0 );
			_dsb->Play( 0, 0, DSBPLAY_LOOPING );
			_start = false;
		}
	}
	
	SetEvent( _events[ 4 ] );
}

#endif
