//
// XML.cpp
// Copyright Michael J Stokes, 2008.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "XML.h"
#include "XMLParser.h"


////////////////////////////////////////////////////////////////////////////////
// XMLElement to stream

void XMLElement::ToStream( std::ostream & stream, bool header, bool newlines ) const
{
	if ( header )
	{
		stream << "<?xml version=\"1.0\" encoding=\"utf-8\"?>";
		if ( newlines ) stream << "\r\n";
	}
	
	stream << "<";
	
	if ( ! _space.empty() )
	{
		stream << _space;
		stream << ":";
	}
	
	stream << _name;
	
	bool closed = false;
	
	for ( list<XMLNode*>::const_iterator it = _children.begin() ; it != _children.end() ; ++ it )
	{
		XMLNode* child = *it;
		
		if ( XMLAttribute* attribute = child->IsAttribute() )
		{
			stream << " ";
			
			if ( ! attribute->GetNamespace().empty() )
			{
				stream << attribute->GetNamespace();
				stream << ":";
			}
			
			stream << attribute->GetName();
			stream << "=\"";
			ValueToStream( stream, attribute->GetValue() );
			stream << "\"";
		}
		else if ( XMLElement* element = child->IsElement() )
		{
			if ( ! closed )
			{
				stream << ( newlines ? ">\r\n" : ">" );
				closed = true;
			}
			
			element->ToStream( stream );
			if ( newlines ) stream << "\r\n";
		}
		else if ( XMLBody* body = child->IsBody() )
		{
			if ( ! closed )
			{
				stream << ">";
				closed = true;
			}
			
			ValueToStream( stream, body->GetValue() );
		}
	}
	
	if ( closed )
	{
		stream << "</";
		
		if ( ! _space.empty() )
		{
			stream << _space;
			stream << ":";
		}
		
		stream << _name;
		stream << ">";
	}
	else
	{
		stream << "/>";
	}
}

////////////////////////////////////////////////////////////////////////////////
// XMLElement to string

string XMLElement::ToString( bool header, bool newlines ) const
{
	std::ostringstream stream( std::ostringstream::out );
	ToStream( stream, header, newlines );
	return stream.str();
}

////////////////////////////////////////////////////////////////////////////////
// XMLElement to file

bool XMLElement::ToFile( PCSTR name, bool header, bool newlines ) const
{
	std::ofstream stream( name );
	if ( ! stream.is_open() ) return false;
	ToStream( stream, header, newlines );
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// XMLElement write a string value to a stream with escaping

void XMLElement::ValueToStream( std::ostream& stream, const string & value )
{
	for ( string::size_type start = 0 ; start < value.length() ; )
	{
		string::size_type next = value.find_first_of( "<>\"'&", start );
		
		if ( string::npos == next )
		{
			stream.write( value.c_str() + start, (std::streamsize)( value.length() - start ) );
			break;
		}
		else if ( start < next )
		{
			stream.write( value.c_str() + start, (std::streamsize)( next - start ) );
		}
		
		switch ( value[ next ] )
		{
		case '<':
			stream << "&lt;";
			break;
		case '>':
			stream << "&gt;";
			break;
		case '\"':
			stream << "&quot;";
			break;
		case '\'':
			stream << "&apos;";
			break;
		case '&':
			stream << "&amp;";
			break;
		}
		
		start = next + 1;
	}
}

////////////////////////////////////////////////////////////////////////////////
// XMLElement create a new XML tree from a stream

XMLElement*	XMLElement::FromStream( std::istream & stream )
{
	XMLParser parser( stream );
	parser.GetIgnored();
	
	XMLElement* element = new XMLElement();
	
	if ( element->Parse( parser ) )
	{
		return element;
	}
	else
	{
		delete element;
		return NULL;
	}
}

////////////////////////////////////////////////////////////////////////////////
// XMLElement create a new XML tree from a string

XMLElement* XMLElement::FromString( const string & input )
{
	std::istringstream in( input, std::istringstream::in );
	return FromStream( in );
}

////////////////////////////////////////////////////////////////////////////////
// XMLElement create a new XML tree from a string

XMLElement* XMLElement::FromString( PCSTR input )
{
	string str( input );
	return FromString( str );
}

////////////////////////////////////////////////////////////////////////////////
// XMLElement create a new XML tree from a file

XMLElement* XMLElement::FromFile( PCSTR name )
{
	std::ifstream stream( name );
	if ( ! stream.is_open() ) return NULL;
	return FromStream( stream );
}

////////////////////////////////////////////////////////////////////////////////
// XMLElement parse an XML element

bool XMLElement::Parse( XMLParser & parser )
{
	if ( ! parser.GetMatch( "<", true ) ) return false;
	if ( ! parser.GetIdentifier( _name, true ) ) return false;
	
	if ( parser.GetMatch( ":" ) )
	{
		_space = _name;
		_name.clear();
		if ( ! parser.GetIdentifier( _name, true ) ) return false;
	}
	
	for ( ; ; )
	{
		if ( parser.IsEOF() )
		{
			return false;
		}
		else if ( parser.GetMatch( ">" ) )
		{
			break;
		}
		else if ( parser.GetMatch( "/" ) )
		{
			return parser.GetMatch( ">", true );
		}
		
		string space, name;
		
		if ( ! parser.GetIdentifier( name, true ) )
		{
			return false;
		}
		
		if ( parser.GetMatch( ":" ) )
		{
			space = name;
			if ( ! parser.GetIdentifier( name, true ) ) return false;
		}
		
		if ( ! parser.GetMatch( "=", true ) ) return false;
		
		char delim[ 2 ] = { 0, 0 };
		
		if ( parser.GetMatch( "\"" ) )
		{
			delim[ 0 ] = '\"';
		}
		else if ( parser.GetMatch( "\'" ) )
		{
			delim[ 0 ] = '\'';
		}
		else
		{
			parser.GetMatch( "\"", true );
			return false;
		}
		
		int len = parser.GetLength( delim );
		
		if ( len < 0 )
		{
			parser.GetMatch( delim, true );
			return false;
		}
		
		XMLAttribute* attribute = (XMLAttribute*)Add( new XMLAttribute( space.c_str(), name.c_str(), "" ) );
		if ( ! parser.GetValue( attribute->_value, len ) ) return false;
		if ( ! parser.GetMatch( delim, true ) ) return false;
	}
	
	for ( ; ; )
	{
		parser.GetIgnored();
		
		int len = parser.GetLength( "<" );
		if ( len < 0 ) return false;
		
		if ( len > 0 )
		{
			XMLBody* body = (XMLBody*)Add( new XMLBody() );
			if ( ! parser.GetValue( body->_value, len ) ) return false;
		}
		else if ( parser.GetMatch( "</" ) )
		{
			if ( ! _space.empty() )
			{
				if ( ! parser.GetMatch( _space.c_str(), true ) ) return false;
				if ( ! parser.GetMatch( ":", true ) ) return false;
			}
			
			if ( ! parser.GetMatch( _name.c_str(), true ) ) return false;
			if ( ! parser.GetMatch( ">", true ) ) return false;
			break;
		}
		else
		{
			XMLElement* element = new XMLElement();
			
			if ( element->Parse( parser ) )
			{
				Add( element );
			}
			else
			{
				delete element;
				return false;
			}
		}
	}
	
	return true;
}
