//
// XML.h
// Copyright Michael J Stokes, 2008.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#pragma once

class XMLNode;
class XMLNamedNode;
class XMLAttribute;
class XMLBody;
class XMLIterator;
class XMLElement;
class XMLParser;


class XMLNode
{
// Construction
protected:
	inline XMLNode() { _parent = NULL; }
	inline XMLNode( const XMLNode& copy ) { _parent = NULL; }
	inline ~XMLNode() {};
	
// Enumeration
public:
	enum Type
	{
		ATTRIBUTE	= 1,
		ELEMENT		= 2,
		BODY		= 4
	};
	
// Operations
public:
	inline XMLElement* GetParent()
	{
		return _parent;
	}
	
	inline Type GetType() const
	{
		return _type;
	}
	
	inline XMLAttribute* IsAttribute() const
	{
		return ( ( NULL != this ) && ( ATTRIBUTE == _type ) ) ? (XMLAttribute*)this : NULL;
	}
	
	inline XMLElement* IsElement() const
	{
		return ( ( NULL != this ) && ( ELEMENT == _type ) ) ? (XMLElement*)this : NULL;
	}
	
	inline XMLBody* IsBody() const
	{
		return ( ( NULL != this ) && ( BODY == _type ) ) ? (XMLBody*)this : NULL;
	}
	
	inline void Detach();
	inline void Delete();
	
// Data
protected:
	XMLElement*	_parent;
	Type		_type;
	
	friend class XMLElement;
};


class XMLNamedNode : public XMLNode
{
// Construction
protected:
	inline XMLNamedNode() {};
	
	inline XMLNamedNode( const XMLNamedNode& copy ) : XMLNode( copy )
	{
		_space	= copy._space;
		_name	= copy._name;
	}
	
	inline ~XMLNamedNode() {};
	
// Operations
public:
	inline const string & GetNamespace() const
	{
		return _space;
	}
	
	inline void SetNamespace( const string & space )
	{
		_space = space;
	}
	
	inline void SetNamespace( PCSTR space )
	{
		_space = space;
	}
	
	inline const string & GetName() const
	{
		return _name;
	}
	
	inline void SetName( const string & name )
	{
		_name = name;
	}
	
	inline void SetName( PCSTR name )
	{
		_name = name;
	}
	
	inline bool IsNamed( PCSTR space, PCSTR name )
	{
		if ( NULL == this )
		{
			return false;
		}
		else if ( NULL != space )
		{
			return ( _space == space ) && ( _name == name );
		}
		else
		{
			return _space.empty() && ( _name == name );
		}
	}
	
	inline bool IsNamed( PCSTR name )
	{
		return ( NULL != this ) && ( _name == name );
	}
	
// Data
protected:
	string	_space;
	string	_name;
};


class XMLAttribute : public XMLNamedNode
{
// Construction
public:
	inline XMLAttribute()
	{
		_type = ATTRIBUTE;
	}
	
	inline XMLAttribute( PCSTR name, PCSTR value )
	{
		_type	= ATTRIBUTE;
		_name	= name;
		_value	= value;
	}
	
	inline XMLAttribute( PCSTR space, PCSTR name, PCSTR value )
	{
		_type	= ATTRIBUTE;
		_space	= space;
		_name	= name;
		_value	= value;
	}
	
	inline XMLAttribute( const XMLAttribute& copy ) : XMLNamedNode( copy )
	{
		_type	= ATTRIBUTE;
		_value	= copy._value;
	}
	
	inline ~XMLAttribute() {}
	
// Operations
public:
	inline const string& GetValue() const
	{
		return _value;
	}
	
	inline void SetValue( const string & value )
	{
		_value = value;
	}
	
	inline void SetValue( PCSTR value )
	{
		_value = value;
	}
	
// Data
protected:
	string	_value;
	
	friend class XMLElement;
};


class XMLBody : public XMLNode
{
// Construction
public:
	inline XMLBody()
	{
		_type = BODY;
	}
	
	inline XMLBody( PCSTR value )
	{
		_type	= BODY;
		_value	= value;
	}
	
	inline XMLBody( const XMLBody& copy ) : XMLNode( copy )
	{
		_type	= BODY;
		_value	= copy._value;
	}
	
	inline ~XMLBody() {}
	
// Operations
public:
	inline const string& GetValue() const
	{
		return _value;
	}
	
	inline void SetValue( const string & value )
	{
		_value = value;
	}
	
	inline void SetValue( PCSTR value )
	{
		_value = value;
	}
	
// Data
protected:
	string	_value;
	
	friend class XMLElement;
};


class XMLIterator
{
// Construction
public:
	inline XMLIterator( list<XMLNode*>& list, XMLNode::Type type ) : _list( list )
	{
		_type	= type;
		_it		= list.begin();
		Next();
	}
	
	inline XMLIterator( const XMLIterator& copy ) : _list( copy._list )
	{
		_type	= copy._type;
		_it		= copy._it;
		_next	= copy._next;
	}
	
	inline ~XMLIterator() {};
	
// Operators
public:
	inline operator bool () const
	{
		return ( NULL != _next );
	}
	
	inline bool operator ! () const
	{
		return ( NULL == _next );
	}
	
	inline XMLNode* operator * ()
	{
		return _next;
	}
	
	inline XMLNode* operator -> ()
	{
		return _next;
	}
	
	inline XMLIterator & operator ++ ()
	{
		Next();
		return *this;
	}
	
// Access operations
public:
	inline XMLAttribute* GetAttribute()
	{
		return _next->IsAttribute();
	}
	
	inline XMLElement* GetElement()
	{
		return _next->IsElement();
	}
	
	inline XMLBody* GetBody()
	{
		return _next->IsBody();
	}
	
// Implementation
protected:
	inline void Next()
	{
		while ( _it != _list.end() )
		{
			_next = * _it ++;
			if ( (int)_next->GetType() & (int)_type ) return;
			if ( XMLNode::ATTRIBUTE == _type ) _it = _list.end();
		}
		
		_next = NULL;
	}
	
// Data
protected:
	list<XMLNode*>&				_list;
	XMLNode::Type				_type;
	list<XMLNode*>::iterator	_it;
	XMLNode*					_next;
};


class XMLElementIterator : public XMLIterator
{
// Construction
public:
	inline XMLElementIterator( list<XMLNode*>& list ) : XMLIterator( list, XMLNode::ELEMENT ) {};
	inline XMLElementIterator( const XMLElementIterator& copy ) : XMLIterator( copy ) {};
	
// Operators
public:
	inline XMLElement* operator * ()
	{
		return (XMLElement*)_next;
	}
	
	inline XMLElement* operator -> ()
	{
		return (XMLElement*)_next;
	}
};


class XMLElement : public XMLNamedNode
{
// Construction
public:
	inline XMLElement()
	{
		_type = ELEMENT;
	}
	
	inline XMLElement( PCSTR name )
	{
		_type	= ELEMENT;
		_name	= name;
	}
	
	inline XMLElement( PCSTR space, PCSTR name )
	{
		_type	= ELEMENT;
		_space	= space;
		_name	= name;
	}
	
	inline XMLElement( const XMLElement& copy ) : XMLNamedNode( copy )
	{
		_type = ELEMENT;
		
		for ( list<XMLNode*>::const_iterator it = copy._children.begin() ; it != copy._children.end() ; ++ it )
		{
			XMLNode* node;
			
			if ( XMLAttribute* attribute = ( *it )->IsAttribute() )
			{
				node = new XMLAttribute( *attribute );
			}
			else if ( XMLElement* element = ( *it )->IsElement() )
			{
				node = new XMLElement( *element );
			}
			else if ( XMLBody* body = ( *it )->IsBody() )
			{
				node = new XMLBody( *body );
			}
			else
			{
				continue;
			}
			
			_children.push_back( node );
			node->_parent = this;
		}
	}
	
	inline ~XMLElement()
	{
		Empty();
	}
	
// Operations
public:
	inline XMLElement* Clone()
	{
		return new XMLElement( *this );
	}
	
	inline XMLNode* Add( XMLNode* child )
	{
		if ( child->GetParent() == this ) return child;
		child->Detach();
		
		if ( XMLAttribute* attribute = child->IsAttribute() )
		{
			if ( XMLAttribute* existing = GetAttribute( attribute->GetNamespace().c_str(), attribute->GetName().c_str() ) )
			{
				existing->Delete();
			}
			
			_children.push_front( child );
		}
		else
		{
			_children.push_back( child );
		}
		
		child->_parent = this;
		return child;
	}
	
	inline void Remove( XMLNode* child )
	{
		if ( child->GetParent() == this )
		{
			_children.remove( child );
			child->_parent = NULL;
		}
	}
	
	inline void Empty()
	{
		while ( ! _children.empty() )
		{
			XMLNode* node = _children.back();
			_children.pop_back();
			node->_parent = NULL;
			node->Delete();
		}
	}
	
	inline XMLAttribute* GetAttribute( PCSTR space, PCSTR name )
	{
		for ( list<XMLNode*>::iterator it = _children.begin() ; it != _children.end() ; ++ it )
		{
			XMLAttribute* child = ( *it )->IsAttribute();
			if ( NULL == child ) break;
			if ( child->IsNamed( space, name ) ) return child;
		}
		
		return NULL;
	}
	
	inline XMLAttribute* GetAttribute( PCSTR name )
	{
		return GetAttribute( NULL, name );
	}

	inline const PCSTR GetAttributeString( PCSTR name )
	{
		if ( XMLAttribute* attribute = GetAttribute( NULL, name ) ) return attribute->GetValue().c_str(); else return NULL;
	}
	
	inline int GetAttributeInt( PCSTR name, int def = 0 )
	{
		if ( PCSTR str = GetAttributeString( name ) ) sscanf( str, "%d", &def );
		return def;
	}
	
	inline float GetAttributeFloat( PCSTR name, float def = 0 )
	{
		if ( PCSTR str = GetAttributeString( name ) ) sscanf( str, "%f", &def );
		return def;
	}
	
	inline XMLElement* GetElement( PCSTR space, PCSTR name )
	{
		for ( list<XMLNode*>::reverse_iterator it = _children.rbegin() ; it != _children.rend() ; ++ it )
		{
			XMLElement* child = (XMLElement*)*it;
			if ( child->IsAttribute() ) break;
			if ( child->IsElement() && child->IsNamed( space, name ) ) return child;
		}
		
		return NULL;
	}
	
	inline XMLElement* GetElement( PCSTR name )
	{
		for ( list<XMLNode*>::reverse_iterator it = _children.rbegin() ; it != _children.rend() ; ++ it )
		{
			XMLElement* child = (XMLElement*)*it;
			if ( child->IsAttribute() ) break;
			if ( child->IsElement() && child->IsNamed( name ) ) return child;
		}
		
		return NULL;
	}
	
	inline PCSTR GetSimpleValue() const
	{
		for ( list<XMLNode*>::const_iterator it = _children.begin() ; it != _children.end() ; ++ it )
		{
			if ( ( *it )->IsBody() ) return ( (XMLBody*)( *it ) )->GetValue().c_str();
		}
		
		return NULL;
	}
	
	inline void SetSimpleValue( PCSTR value )
	{
		for ( list<XMLNode*>::reverse_iterator it = _children.rbegin() ; it != _children.rend() ; ++ it )
		{
			XMLNode* child = *it;
			if ( child->IsAttribute() ) break;
			if ( child->IsBody() ) child->Delete();
		}
		
		if ( ( NULL != value ) && ( 0 != *value ) )
		{
			Add( new XMLBody( value ) );
		}
	}
	
	inline void SetSimpleValue( const string & value )
	{
		SetSimpleValue( value.c_str() );
	}
	
	inline XMLElement* AddElement( PCSTR name )
	{
		return (XMLElement*)Add( new XMLElement( name ) );
	}
	
	inline XMLElement* AddElement( PCSTR space, PCSTR name )
	{
		return (XMLElement*)Add( new XMLElement( space, name ) );
	}
	
	inline XMLAttribute* AddAttribute( PCSTR name, PCSTR value )
	{
		return (XMLAttribute*)Add( new XMLAttribute( name, value ) );
	}
	
	inline XMLAttribute* AddAttributeInt( PCSTR name, int value )
	{
		char temp[ 64 ];
		sprintf( temp, "%i", value );
		return AddAttribute( name, temp );
	}
	
	inline XMLAttribute* AddAttributeFloat( PCSTR name, float value )
	{
		char temp[ 64 ];
		sprintf( temp, "%f", value );
		return AddAttribute( name, temp );
	}
	
	inline UINT GetChildCount() const
	{
		return (UINT)_children.size();
	}
	
	inline XMLIterator GetChildren( int type = ATTRIBUTE | ELEMENT | BODY )
	{
		return XMLIterator( _children, (Type)type );
	}
	
	inline XMLIterator GetAttributes()
	{
		return XMLIterator( _children, ATTRIBUTE );
	}
	
	inline XMLElementIterator GetElements()
	{
		return XMLElementIterator( _children );
	}
	
	inline XMLIterator GetBodyNodes()
	{
		return XMLIterator( _children, BODY );
	}
	
	inline XMLIterator GetContents()
	{
		return XMLIterator( _children, (Type)( ELEMENT | BODY ) );
	}
	
// Encoding
public:
	void				ToStream( std::ostream & stream, bool header = false, bool newlines = true ) const;
	string				ToString( bool header = false, bool newlines = true ) const;
	bool				ToFile( PCSTR name, bool header = false, bool newlines = true ) const;
	
// Decoding
public:
	static XMLElement*	FromStream( std::istream & stream );
	static XMLElement*	FromString( const string & input );
	static XMLElement*	FromString( PCSTR input );
	static XMLElement*	FromFile( PCSTR name );
	
// Implementation
protected:
	static void			ValueToStream( std::ostream& stream, const string & value );
	bool				Parse( XMLParser & parser );
	
// Data
protected:
	list<XMLNode*>	_children;
};


class XMLAutoDelete
{
// Construction
public:
	inline XMLAutoDelete( XMLElement* element )
	{
		_element = element;
	}
	
	inline ~XMLAutoDelete()
	{
		if ( NULL != _element ) delete _element;
	}
	
	inline XMLElement* Detach()
	{
		XMLElement* element = _element;
		_element = NULL;
		return element;
	}
	
// Data
protected:
	XMLElement*	_element;
};

inline void XMLNode::Detach()
{
	if ( NULL != _parent )
	{
		_parent->Remove( this );
	}
}

inline void XMLNode::Delete()
{
	if ( NULL != this )
	{
		Detach();
		
		switch ( _type )
		{
		case ATTRIBUTE:
			delete (XMLAttribute*)this;
			break;
		case ELEMENT:
			delete (XMLElement*)this;
			break;
		case BODY:
			delete (XMLBody*)this;
			break;
		}
	}
}
