//
// XMLParser.h
// Copyright Michael J Stokes, 2008.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#pragma once


class XMLParser
{
// Construction
public:
	inline XMLParser( std::istream & stream ) : _stream( stream )
	{
	}
	
// Methods
public:
	inline bool IsEOF()
	{
		return _stream.eof();
	}
	
	inline void Skip( int len )
	{
		_stream.ignore( (std::streamsize)len );
	}
	
	inline bool GetWhitespace()
	{
		for ( int ws = 0 ; ; )
		{
			int c = _stream.get();
			
			if ( ( c > 0 ) && ( c <= 32 ) )
			{
				ws ++;
			}
			else if ( 0 == c )
			{
				return ( ws > 0 );
			}
			else
			{
				_stream.unget();
				return ( ws > 0 );
			}
		}
	}
	
	inline bool GetMatch( PCSTR value, bool error = false, bool eatWhitespace = true )
	{
		std::streampos pos = _stream.tellg();
		if ( eatWhitespace ) GetWhitespace();
		
		for ( PCSTR val = value ; *val ; val ++ )
		{
			int c = _stream.get();
			if ( c == *val ) continue;
			else if ( 0 != c ) _stream.seekg( pos );
			if ( error ) cerr << "XMLParser error: expected '" << value << "'" << endl;
			return false;
		}
		
		return true;
	}
	
	inline bool GetIdentifier( string & value, bool error = false, bool eatWhitespace = true )
	{
		std::streampos pos = _stream.tellg();
		if ( eatWhitespace ) GetWhitespace();
		
		for ( ; ; )
		{
			int c = _stream.get();
			
			if ( ( ( c < 0 ) || ( c > 32 ) ) && ( NULL == strchr( "<:>/=;", c ) ) )
			{
				value += c;
			}
			else if ( 0 == c )
			{
				if ( error ) cerr << "XMLParser error: expected identifier" << endl;
				return false;
			}
			else if ( value.empty() )
			{
				if ( error ) cerr << "XMLParser error: expected identifier" << endl;
				_stream.seekg( pos );
				return false;
			}
			else
			{
				_stream.unget();
				return true;
			}
		}
	}
	
	inline int GetLength( PCSTR delim )
	{
		int len = 0;
		
		for ( std::streampos pos = _stream.tellg() ; ; len ++ )
		{
			int c = _stream.get();
			if ( c <= 0 ) return -1;
			
			if ( NULL != strchr( delim, c ) )
			{
				_stream.seekg( pos );
				return (int)len;
			}
		}
	}
	
	inline bool GetIgnored()
	{
		for ( bool ws = false ; ; )
		{
			bool s = GetWhitespace();
			if ( s ) ws = true;
			
			if ( GetMatch( "<!--", false, false ) || GetMatch( "<?", false, false ) )
			{
				int len = GetLength( ">" );
				if ( len < 0 ) return ws;
				Skip( len + 1 );
			}
			
			if ( ( ! s ) || IsEOF() ) return ws;
		}
	}
	
	inline bool GetValue( string & value, int len )
	{
		value.reserve( value.length() + len );
		
		for ( ; len > 0 ; len -- )
		{
			int c = _stream.get();
			if ( 0 == c ) return false;
			
			if ( '&' != c )
			{
				value += c;
				continue;
			}
			
			if ( GetMatch( "lt;", false, false ) )
			{
				value += "<";
				len -= 3;
			}
			else if ( GetMatch( "gt;", false, false ) )
			{
				value += ">";
				len -= 3;
			}
			else if ( GetMatch( "amp;", false, false ) )
			{
				value += "&";
				len -= 4;
			}
			else if ( GetMatch( "quot;", false, false ) )
			{
				value += "\"";
				len -= 5;
			}
			else if ( GetMatch( "apos;", false, false ) )
			{
				value += "\'";
				len -= 5;
			}
			else if ( GetMatch( "#x", false, false ) )
			{
				string num;
				if ( ! GetIdentifier( num, true, false ) ) return false;
				if ( ! GetMatch( ";", true, false ) ) return false;
				if ( 1 != sscanf( num.c_str(), "%x", &c ) ) return false;
				value += c;
				len -= 3 + (int)num.length();
			}
			else if ( GetMatch( "#", false, false ) )
			{
				string num;
				if ( ! GetIdentifier( num, true, false ) ) return false;
				if ( ! GetMatch( ";", true, false ) ) return false;
				if ( 1 != sscanf( num.c_str(), "%d", &c ) ) return false;
				value += c;
				len -= 2 + (int)num.length();
			}
			else
			{
				value += "&";
			}
		}
		
		return true;
	}
	
// Data
protected:
	std::istream & _stream;
};
