// stbi-0.97 - public domain JPEG/PNG reader - http://nothings.org/stb_image.c

// Limitations:
//    - no progressive/interlaced support (jpeg, png)
//    - 8-bit samples only (jpeg, png)
//    - not threadsafe
//    - channel subsampling of at most 2 in each dimension (jpeg)
//    - no delayed line count (jpeg) -- IJG doesn't support either
//
// Basic usage:
//    int x,y,n;
//    unsigned char *data = stbi_load(file, &x, &y, &n, 0);
//    // ... process data if not NULL ... 
//    // ... x = width, y = height, n = # 8-bit components per pixel ...
//    // ... replace '0' with '1'..'4' to force that many components per pixel
//    stbi_image_free(data)
//
// Standard parameters:
//    int *x       -- outputs image width in pixels
//    int *y       -- outputs image height in pixels
//    int *comp    -- outputs # of image components in image file
//    int req_comp -- if non-zero, # of image components requested in result
//
// The return value from an image loader is an 'unsigned char *' which points
// to the pixel data. The pixel data consists of *y scanlines of *x pixels,
// with each pixel consisting of N interleaved 8-bit components; the first
// pixel pointed to is top-left-most in the image. There is no padding between
// image scanlines or between pixels, regardless of format. The number of
// components N is 'req_comp' if req_comp is non-zero, or *comp otherwise.
// If req_comp is non-zero, *comp has the number of components that _would_
// have been output otherwise. E.g. if you set req_comp to 4, you will always
// get RGBA output, but you can check *comp to easily see if it's opaque.
//
// An output image with N components has the following components interleaved
// in this order in each pixel:
//
//     N=#comp     components
//       1           grey
//       2           grey, alpha
//       3           red, green, blue
//       4           red, green, blue, alpha
//
// If image loading fails for any reason, the return value will be NULL,
// and *x, *y, *comp will be unchanged. The function stbi_failure_reason()
// can be queried for an extremely brief, end-user unfriendly explanation
// of why the load failed. Define STBI_NO_FAILURE_STRINGS to avoid
// compiling these strings at all, and STBI_FAILURE_USERMSG to get slightly
// more user-friendly ones.
//
// Paletted PNG and BMP images are automatically depalettized.

enum
{
   STBI_default = 0, // only used for req_comp

   STBI_grey       = 1,
   STBI_grey_alpha = 2,
   STBI_rgb        = 3,
   STBI_rgb_alpha  = 4,
};


// PRIMARY API - works on images of any type

// load image by filename, open file, or memory buffer
BYTE *stbi_load_from_file  (std::istream *f, int *x, int *y, int *comp, int req_comp);
BYTE *stbi_load_from_memory(BYTE *buffer, int len, int *x, int *y, int *comp, int req_comp);
// for stbi_load_from_file, std::istream pointer is left pointing immediately after image

// ZLIB client - used by PNG, available for other purposes

char *stbi_zlib_decode_malloc_guesssize(int initial_size, int *outlen);
char *stbi_zlib_decode_malloc(char *buffer, int len, int *outlen);
int   stbi_zlib_decode_buffer(char *obuffer, int olen, char *ibuffer, int ilen);
int   stbi_zlib_decode_buffer_no_header(char *obuffer, int olen, char *ibuffer, int ilen);


// TYPE-SPECIFIC ACCESS

// is it a jpeg?
int      stbi_jpeg_test_file       (std::istream *f);
int      stbi_jpeg_test_memory     (BYTE *buffer, int len);

BYTE *stbi_jpeg_load_from_file  (std::istream *f, int *x, int *y, int *comp, int req_comp);
BYTE *stbi_jpeg_load_from_memory(BYTE *buffer, int len, int *x, int *y, int *comp, int req_comp);

// is it a png?
int      stbi_png_test_file        (std::istream *f);
int      stbi_png_test_memory      (BYTE *buffer, int len);

BYTE *stbi_png_load_from_file   (std::istream *f, int *x, int *y, int *comp, int req_comp);
BYTE *stbi_png_load_from_memory (BYTE *buffer, int len, int *x, int *y, int *comp, int req_comp);

// is it a bmp?
int      stbi_bmp_test_file        (std::istream *f);
int      stbi_bmp_test_memory      (BYTE *buffer, int len);

BYTE *stbi_bmp_load_from_file   (std::istream *f, int *x, int *y, int *comp, int req_comp);
BYTE *stbi_bmp_load_from_memory (BYTE *buffer, int len, int *x, int *y, int *comp, int req_comp);
