//
// ConsoleCtrl.cpp
// Copyright Michael J Stokes, 2009.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "ConsoleCtrl.h"


IMPLEMENT_ABSTRACT_CLASS( ConsoleCtrl, wxListCtrl )

BEGIN_EVENT_TABLE( ConsoleCtrl, wxListCtrl )
	EVT_SIZE( ConsoleCtrl::OnSize )
END_EVENT_TABLE()


//////////////////////////////////////////////////////////////////////////////
// ConsoleCtrl - construction

ConsoleCtrl::ConsoleCtrl( wxWindow* parent ) :
	wxListCtrl( parent, wxID_ANY, wxDefaultPosition, wxSize( 0, 0 ), wxLC_REPORT | wxLC_NO_HEADER | wxBORDER_NONE )
{
	InsertColumn( 0, T("Message"), wxLIST_FORMAT_LEFT, 0 );
	
	_cout = new Stream( this, cout );
	_cerr = new Stream( this, cerr );
	
	cout << "Console initialized." << endl;
}

//////////////////////////////////////////////////////////////////////////////
// ConsoleCtrl - destruction

ConsoleCtrl::~ConsoleCtrl()
{
	delete _cerr;
	delete _cout;
}

//////////////////////////////////////////////////////////////////////////////
// ConsoleCtrl - deliver a line

void ConsoleCtrl::Deliver( Stream* stream, const string& line )
{
	wxListItem item;
	item.SetId( GetItemCount() );
	item.SetImage( 0 );
	item.SetText( line.c_str() );
	
	if ( stream == _cerr )
	{
		item.SetTextColour( *wxRED );
	}
	
	item.SetId( InsertItem( item ) );

	EnsureVisible( GetItemCount() - 1 );
	
#ifdef _WIN32
	RedrawWindow( (HWND)GetHandle(), NULL, NULL, RDW_UPDATENOW );
#endif
}

//////////////////////////////////////////////////////////////////////////////
// ConsoleCtrl - resize

void ConsoleCtrl::OnSize( wxSizeEvent& event )
{
	SetColumnWidth( 0, GetClientSize().x - 20 );
	event.Skip();
}


//////////////////////////////////////////////////////////////////////////////
// ConsoleCtrl::Stream - construction

ConsoleCtrl::Stream::Stream( ConsoleCtrl* console, std::ostream& stream ) : _console( console ), _stream( stream )
{
	_old = stream.rdbuf();
	stream.rdbuf( this );
}

//////////////////////////////////////////////////////////////////////////////
// ConsoleCtrl::Stream - destruction

ConsoleCtrl::Stream::~Stream()
{
	_stream.rdbuf( _old );
}

//////////////////////////////////////////////////////////////////////////////
// ConsoleCtrl::Stream - overflow implementation

ConsoleCtrl::Stream::int_type ConsoleCtrl::Stream::overflow( int_type v )
{
	if ( v == '\n' )
	{
		_console->Deliver( this, _string );
		_string.clear();
	}
	else
	{
		_string += v;
	}
	
	return v;
}

//////////////////////////////////////////////////////////////////////////////
// ConsoleCtrl::Stream - xsputn implementation

std::streamsize	ConsoleCtrl::Stream::xsputn( const char* p, std::streamsize n )
{
	_string.append( p, p + n );
	
	for ( ; ; )
	{
		int pos = (int)_string.find( '\n' );
		if ( pos == string::npos ) break;
		
		string line( _string.begin(), _string.begin() + pos );
		_console->Deliver( this, line );
		_string.erase( _string.begin(), _string.begin() + pos + 1 );
	}
	
	return n;
}
