//
// FeatureDialog.cpp
// Copyright Michael J Stokes, 2008.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "Editor.h"
#include "Application.h"
#include "Features.h"
#include "Feature.h"
#include "FeatureDialog.h"
#include "ScriptDialog.h"

#ifdef _WIN32
#pragma warning( disable : 4312 )
#endif

IMPLEMENT_ABSTRACT_CLASS( FeatureDialog, wxDialog )

BEGIN_EVENT_TABLE( FeatureDialog, wxDialog )
	EVT_TEXT( 1, FeatureDialog::OnNameChange )  
	EVT_RADIOBUTTON( 2, FeatureDialog::OnRadio ) 
	EVT_RADIOBUTTON( 3, FeatureDialog::OnRadio ) 
	EVT_BUTTON( ID_FEATURE_NEW, FeatureDialog::OnBrowse )
	EVT_LIST_ITEM_SELECTED( 4, FeatureDialog::OnListSelected )
	EVT_LIST_ITEM_DESELECTED( 4, FeatureDialog::OnListSelected )
	EVT_LIST_ITEM_ACTIVATED( 4, FeatureDialog::OnListActivated )
	EVT_BUTTON( wxID_ADD, FeatureDialog::OnAdd )
	EVT_BUTTON( wxID_EDIT, FeatureDialog::OnEdit )
	EVT_BUTTON( wxID_DELETE, FeatureDialog::OnDelete )
	EVT_BUTTON( wxID_OK, FeatureDialog::OnOK )
END_EVENT_TABLE()


//////////////////////////////////////////////////////////////////////////////
// FeatureDialog construction

FeatureDialog::FeatureDialog( wxWindow* parent ) : wxDialog( parent, wxID_ANY, T("Feature"), wxDefaultPosition )
{
	Feature* feature = Application::GetInstance()->GetFeatures()->GetSelection().feature;
	
	wxFlexGridSizer* lines = new wxFlexGridSizer( 2, 8, 8 );
	lines->Add( new wxStaticText( this, wxID_ANY, T("&Name:") ), 0, wxALIGN_LEFT | wxALIGN_CENTER_VERTICAL );
	lines->Add( _name = new wxTextCtrl( this, 1 ), 1, wxEXPAND | wxALIGN_CENTER_VERTICAL );
	
	_optSimple = new wxRadioButton( this, 2, "Default behaviour; &play the following OGG sound file:", wxDefaultPosition, wxDefaultSize, wxRB_GROUP );
	_optScript = new wxRadioButton( this, 3, "Execute the following &BML script:" );
	
	wxBoxSizer* simple = new wxBoxSizer( wxHORIZONTAL );
	simple->AddSpacer( 20 );
	simple->Add( _simplePath = new wxStaticText( this, wxID_ANY, "" ), 1, wxALIGN_CENTER_VERTICAL );
	simple->AddSpacer( 10 );
	simple->Add( _simpleButton = new wxButton( this, ID_FEATURE_NEW, "Open Folder" ), 0, wxEXPAND | wxALIGN_CENTER_VERTICAL );
	
	wxBoxSizer* script = new wxBoxSizer( wxHORIZONTAL );
	script->AddSpacer( 20 );
	script->Add( _list = new wxListCtrl( this, 4, wxDefaultPosition, wxDefaultSize, wxLC_REPORT ), 1, wxEXPAND );
	script->AddSpacer( 10 );
	wxBoxSizer* buttons = new wxBoxSizer( wxVERTICAL );
	buttons->Add( _add = new wxButton( this, wxID_ADD, "&Add..." ) );
	buttons->AddSpacer( 8 );
	buttons->Add( _edit = new wxButton( this, wxID_EDIT, "&Edit..." ) );
	buttons->AddSpacer( 8 );
	buttons->Add( _delete = new wxButton( this, wxID_DELETE, "&Delete" ) );
	script->Add( buttons, 0 );
	
	wxBoxSizer* root = new wxBoxSizer( wxVERTICAL );
	root->Add( lines, 0, wxEXPAND | wxLEFT | wxTOP | wxRIGHT, 10 );
	root->AddSpacer( 16 );
	root->Add( _optSimple, 0, wxEXPAND | wxLEFT | wxRIGHT, 10 );
	root->AddSpacer( 10 );
	root->Add( simple, 0, wxEXPAND | wxLEFT | wxRIGHT, 10 );
	root->AddSpacer( 16 );
	root->Add( _optScript, 0, wxEXPAND | wxLEFT | wxRIGHT, 10 );
	root->AddSpacer( 10 );
	root->Add( script, 0, wxEXPAND | wxLEFT | wxRIGHT, 10 );
	root->AddSpacer( 16 );
	root->Add( CreateButtonSizer( wxOK | wxCANCEL ), 0, wxALIGN_CENTER | wxLEFT | wxBOTTOM | wxRIGHT, 10 );
	
	_name->SetValue( feature->GetName() );
	_name->SetMinSize( wxSize( 360, _name->GetMinSize().y ) );
	_name->SetFocus();
	_list->SetMinSize( wxSize( 360, 120 ) );
	
	_optSimple->SetValue( NULL == feature->GetScript() );
	_optScript->SetValue( NULL != feature->GetScript() );
	
	wxCommandEvent e;
	OnNameChange( e );
	OnRadio( e );
	
	_list->InsertColumn( 0, "Time", wxLIST_FORMAT_LEFT, 40 );
	_list->InsertColumn( 1, "Actor", wxLIST_FORMAT_LEFT, 60 );
	_list->InsertColumn( 2, "Script", wxLIST_FORMAT_LEFT, 240 );
	
	if ( XMLElement* script = feature->GetScript() )
	{
		for ( XMLElementIterator it = script->GetElements() ; it ; ++ it )
		{
			if ( it->IsNamed( "block" ) )
			{
				int time = (int)it->GetAttributeFloat( "time" );
				
				wxListItem item;
				item.SetText( wxString::Format( "%.2i:%.2i", time / 60, time % 60 ) );
				item.SetData( it->Clone() );
				item.SetId( _list->InsertItem( item ) );
				
				PCSTR actor = it->GetAttributeString( "actor" );
				item.SetColumn( 1 );
				item.SetText( actor ? actor : "Unknown" );
				_list->SetItem( item );
				
				wxString value;
				
				if ( XMLElementIterator first = it->GetElements() )
				{
					value = first->ToString( false, false );
				}
				
				if ( value.Length() > 128 )
				{
					value.Truncate( 128 );
					value += "...";
				}
				
				item.SetColumn( 2 );
				item.SetText( value );
				_list->SetItem( item );
			}
		}
		
		_list->SortItems( ListCompare, 0 );
	}
	
	SetSizerAndFit( root );
}

//////////////////////////////////////////////////////////////////////////////
// FeatureDialog destruction

FeatureDialog::~FeatureDialog()
{
	for ( int i = 0 ; i < _list->GetItemCount() ; i ++ )
	{
		( (XMLElement*)_list->GetItemData( i ) )->Delete();
	}
}

//////////////////////////////////////////////////////////////////////////////
// FeatureDialog list sort

int FeatureDialog::ListCompare( long item1, long item2, long sortData )
{
	XMLElement* xml1 = (XMLElement*)item1;
	XMLElement* xml2 = (XMLElement*)item2;
	
	scalar time1 = xml1->GetAttributeFloat( "time" );
	scalar time2 = xml2->GetAttributeFloat( "time" );
	
	if ( time1 == time2 )
		return 0;
	else if ( time1 < time2 )
		return -1;
	else
		return 1;
}

//////////////////////////////////////////////////////////////////////////////
// FeatureDialog name change

void FeatureDialog::OnNameChange( wxCommandEvent& event )
{
	wxFileName name( "features/" + _name->GetValue() + "/voice.ogg" );
	name.Normalize();
	_simplePath->SetLabel( name.GetFullPath() );
}

//////////////////////////////////////////////////////////////////////////////
// FeatureDialog radio button

void FeatureDialog::OnRadio( wxCommandEvent& event )
{
	_simplePath->Enable( _optSimple->GetValue() );
	_simpleButton->Enable( _optSimple->GetValue() );
	_list->Enable( _optScript->GetValue() );
	_add->Enable( _optScript->GetValue() );
	_edit->Enable( _optScript->GetValue() && ( 1 == _list->GetSelectedItemCount() ) );
	_delete->Enable( _optScript->GetValue() && ( 0 < _list->GetSelectedItemCount() ) );
}

//////////////////////////////////////////////////////////////////////////////
// FeatureDialog browse button

void FeatureDialog::OnBrowse( wxCommandEvent& event )
{
#ifdef _WIN32
	wxFileName name( "features/" + _name->GetValue() );
	name.Normalize();
	ShellExecute( (HWND)GetHandle(), "open", name.GetFullPath(), NULL, NULL, SW_SHOWNORMAL );
#endif
}

//////////////////////////////////////////////////////////////////////////////
// FeatureDialog list selection change

void FeatureDialog::OnListSelected( wxListEvent& event )
{
	_edit->Enable( _optScript->GetValue() && ( 1 == _list->GetSelectedItemCount() ) );
	_delete->Enable( _optScript->GetValue() && ( 0 < _list->GetSelectedItemCount() ) );
}

//////////////////////////////////////////////////////////////////////////////
// FeatureDialog list activation

void FeatureDialog::OnListActivated( wxListEvent& event )
{
	if ( 1 == _list->GetSelectedItemCount() )
	{
		wxCommandEvent cmd;
		OnEdit( cmd );
	}
}

//////////////////////////////////////////////////////////////////////////////
// FeatureDialog add button

void FeatureDialog::OnAdd( wxCommandEvent& event )
{
	ScriptDialog dlg( this );
	if ( wxID_OK != dlg.ShowModal() ) return;
	
	XMLElement* block = dlg.GetBlockDetach();
	int time = (int)block->GetAttributeFloat( "time" );
	
	wxListItem item;
	item.SetText( wxString::Format( "%.2i:%.2i", time / 60, time % 60 ) );
	item.SetData( block );
	item.SetId( _list->InsertItem( item ) );
	
	PCSTR actor = block->GetAttributeString( "actor" );
	item.SetColumn( 1 );
	item.SetText( actor ? actor : "Unknown" );
	_list->SetItem( item );
	
	wxString value;
	
	if ( XMLElementIterator first = block->GetElements() )
	{
		value = first->ToString( false, false );
	}
	
	if ( value.Length() > 128 )
	{
		value.Truncate( 128 );
		value += "...";
	}
	
	item.SetColumn( 2 );
	item.SetText( value );
	_list->SetItem( item );
	
	_list->SortItems( ListCompare, 0 );
}

//////////////////////////////////////////////////////////////////////////////
// FeatureDialog edit button

void FeatureDialog::OnEdit( wxCommandEvent& event )
{
	long index = _list->GetNextItem( -1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED );
	if ( index < 0 ) return;
	
	ScriptDialog dlg( this, (XMLElement*)_list->GetItemData( index ) );
	
	if ( wxID_OK != dlg.ShowModal() )
	{
		dlg.GetBlockDetach();
		return;
	}
	
	XMLElement* block = dlg.GetBlockDetach();
	int time = (int)block->GetAttributeFloat( "time" );
	
	wxListItem item;
	item.SetId( index );
	item.SetText( wxString::Format( "%.2i:%.2i", time / 60, time % 60 ) );
	item.SetData( block );
	_list->SetItem( item );
	
	PCSTR actor = block->GetAttributeString( "actor" );
	item.SetColumn( 1 );
	item.SetText( actor ? actor : "Unknown" );
	_list->SetItem( item );
	
	wxString value;
	
	if ( XMLElementIterator first = block->GetElements() )
	{
		value = first->ToString( false, false );
	}
	
	if ( value.Length() > 128 )
	{
		value.Truncate( 128 );
		value += "...";
	}
	
	item.SetColumn( 2 );
	item.SetText( value );
	_list->SetItem( item );
	
	_list->SortItems( ListCompare, 0 );
}

//////////////////////////////////////////////////////////////////////////////
// FeatureDialog delete button

void FeatureDialog::OnDelete( wxCommandEvent& event )
{
	for ( long index = -1 ; ( index = _list->GetNextItem( index, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED ) ) >= 0 ; )
	{
		delete (XMLElement*)_list->GetItemData( index );
		_list->DeleteItem( index );
		index --;
	}
}

//////////////////////////////////////////////////////////////////////////////
// FeatureDialog OK button

void FeatureDialog::OnOK( wxCommandEvent& event )
{
	wxString name( _name->GetValue() );
	
	for ( UINT i = 0 ; i < name.Length() ; i ++ )
	{
		if ( ( name[ i ] >= 0 ) && ( name[ i ] <= 32 ) )
		{
			wxMessageBox( T("There are invalid characters in the feature name that you specified.\n\nRemember that you cannot use spaces in feature names."), T("Feature Name") );
			return;
		}
	}
	
	Feature* feature = Application::GetInstance()->GetFeatures()->GetSelection().feature;
	
	if ( feature->GetName() != name )
	{
		if ( ! feature->Rename( name ) )
		{
			wxMessageBox( T("Unable to rename the feature.") );
			return;
		}
	}
	
	if ( _optScript->GetValue() )
	{
		XMLElement* script = new XMLElement( "script" );
		
		while ( _list->GetItemCount() > 0 )
		{
			XMLElement* block = (XMLElement*)_list->GetItemData( 0 );
			_list->DeleteItem( 0 );
			script->Add( block );
		}
		
		feature->SetScript( script );
	}
	else
	{
		feature->SetScript( NULL );
	}
	
	feature->Save();
	
	EndDialog( wxID_OK );
}
