//
// GLView.cpp
// Copyright Michael J Stokes, 2008.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "Editor.h"
#include "Application.h"
#include "Renderer.h"
#include "Features.h"
#include "Feature.h"
#include "Actors.h"
#include "GLView.h"
#include "FeatureDialog.h"
#include "BMLFrame.h"


IMPLEMENT_ABSTRACT_CLASS( GLView, wxGLCanvas )

BEGIN_EVENT_TABLE( GLView, wxGLCanvas )
	EVT_IDLE( GLView::OnIdle )
	EVT_SIZE( GLView::OnSize )
	EVT_ERASE_BACKGROUND( GLView::OnErase )
	EVT_PAINT( GLView::OnPaint )
	EVT_MOTION( GLView::OnMouseMove )
	EVT_LEFT_DOWN( GLView::OnLeftDown )
	EVT_LEFT_UP( GLView::OnLeftUp )
	EVT_RIGHT_DOWN( GLView::OnRightDown )
	EVT_RIGHT_UP( GLView::OnRightUp )
	EVT_LEFT_DCLICK( GLView::OnDoubleClick )
	EVT_MOUSE_CAPTURE_LOST( GLView::OnCaptureLost ) 
	EVT_KEY_DOWN( GLView::OnKeyDown )
	EVT_UPDATE_UI( ID_FEATURE_NEW, GLView::OnUpdateOpen )
	EVT_MENU( ID_FEATURE_NEW, GLView::OnFeatureNew )
	EVT_BUTTON( ID_FEATURE_NEW, GLView::OnFeatureNew )
	EVT_UPDATE_UI( ID_FEATURE_EDIT, GLView::OnUpdateFeature )
	EVT_MENU( ID_FEATURE_EDIT, GLView::OnFeatureEdit )
	EVT_BUTTON( ID_FEATURE_EDIT, GLView::OnFeatureEdit )
	EVT_UPDATE_UI( ID_FEATURE_DELETE, GLView::OnUpdateFeature )
	EVT_MENU( ID_FEATURE_DELETE, GLView::OnFeatureDelete )
	EVT_BUTTON( ID_FEATURE_DELETE, GLView::OnFeatureDelete )
	EVT_UPDATE_UI( ID_POINT_ACTIVATION, GLView::OnUpdateFeature )
	EVT_MENU( ID_POINT_ACTIVATION, GLView::OnPointActivation )
	EVT_BUTTON( ID_POINT_ACTIVATION, GLView::OnPointActivation )
	EVT_UPDATE_UI( ID_POINT_LOCATION, GLView::OnUpdateFeature )
	EVT_MENU( ID_POINT_LOCATION, GLView::OnPointLocation )
	EVT_BUTTON( ID_POINT_LOCATION, GLView::OnPointLocation )
	EVT_UPDATE_UI( ID_POINT_GUIDE, GLView::OnUpdateFeature )
	EVT_MENU( ID_POINT_GUIDE, GLView::OnPointGuide )
	EVT_BUTTON( ID_POINT_GUIDE, GLView::OnPointGuide )
	EVT_UPDATE_UI( ID_POINT_DELETE, GLView::OnUpdatePointDelete )
	EVT_UPDATE_UI( ID_POINT_DELETE_WITH_CAPTION, GLView::OnUpdatePointDelete )
	EVT_MENU( ID_POINT_DELETE, GLView::OnPointDelete )
	EVT_BUTTON( ID_POINT_DELETE, GLView::OnPointDelete )
	EVT_MENU( ID_POINT_DELETE_WITH_CAPTION, GLView::OnPointDelete )
END_EVENT_TABLE()


//////////////////////////////////////////////////////////////////////////////
// GLView - construction

static int GL_CANVAS_ATTRIBUTES[] = { WX_GL_RGBA, WX_GL_DOUBLEBUFFER, WX_GL_STENCIL_SIZE, 1, 0 };

GLView::GLView( wxWindow* parent ) :
	wxGLCanvas( parent, wxID_ANY, wxDefaultPosition, wxSize( 0, 0 ), wxWS_EX_PROCESS_IDLE, T("ViewCtrl"), GL_CANVAS_ATTRIBUTES )
{
	_app		= NULL;
	_time		= 0;
	_turning	= false;
	_dragging	= false;
	_sizing		= false;
}

//////////////////////////////////////////////////////////////////////////////
// GLView - destruction

GLView::~GLView()
{
	Application::Destroy();
}

//////////////////////////////////////////////////////////////////////////////
// GLView - idle

void GLView::OnIdle( wxIdleEvent& event )
{
	wxTopLevelWindow* parent = (wxTopLevelWindow*)wxGetApp().GetTopWindow();
	
	if ( ( NULL != parent ) && parent->IsKindOf( CLASSINFO( wxTopLevelWindow ) ) && parent->IsActive() )
	{
		Refresh();
		event.RequestMore();
	}
	else if ( BMLFrame* frame = BMLFrame::GetInstance() )
	{
		if ( frame->IsActive() )
		{
			Refresh();
			event.RequestMore();
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
// GLView - size

void GLView::OnSize( wxSizeEvent& event )
{
	if ( NULL != Application::GetInstance() )
	{
		wxSize size( GetClientSize() );
		glViewport( 0, 0, size.GetWidth(), size.GetHeight() );
		
		vec2 fov( 90, 0 );
		fov.y = atan( (scalar)size.GetHeight() / ( (scalar)size.GetWidth() / tan( fov.x * M_PI / 360.0f ) ) ) * 360.0f / M_PI;
		
		glMatrixMode( GL_PROJECTION );
		glLoadIdentity();
		gluPerspective( fov.y, (scalar)size.GetWidth() / (scalar)size.GetHeight(), 0.1f, 256.0f );
		glScalef( 1, 1, -1 );
		glMatrixMode( GL_MODELVIEW );
		
		Refresh();
	}
}

//////////////////////////////////////////////////////////////////////////////
// GLView - paint

void GLView::OnErase( wxEraseEvent& event )
{
}

void GLView::OnPaint( wxPaintEvent& event )
{
	wxPaintDC dc( this );
	
	SetCurrent();
	
	Application* app = Application::GetInstance();
	
	if ( NULL == app )
	{
		glClearColor( 0.25f, 0.25f, 0.25f, 0 );
		glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
		SwapBuffers();
		return;
	}
	else if ( _app != app )
	{
		_app = app;
		wxSizeEvent event;
		OnSize( event );
	}
	
	glClearColor( 0.6f, 0.6f, 1.0f, 1.0f );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glLoadIdentity();
	
	matrix4x4 rotation( _orientation.Inverted() );
	glLoadMatrixf( rotation );
	
	if ( NULL == app->GetEgo() )
	{
		glTranslatef( -_position.x, -_position.y, -_position.z );
		
		vec3 dir( _orientation.Direction() );
		dir.Set( -dir.x, 1, -dir.z );
		dir.Normalize();
		scalar v[ 4 ] = { dir.x, dir.y, dir.z, 0 };
		glLightfv( GL_LIGHT0, GL_POSITION, v );
		
		vec3 pos( _position );
		pos.y -= 0.75f;
		app->GetActors()->RegisterPosition( "ego", pos );
	}
	else
	{
		glTranslatef( 0, -1.5f, 0 );
	}
	
	wxTopLevelWindow* parent = (wxTopLevelWindow*)wxGetApp().GetTopWindow();
	bool forward = ( NULL != parent ) && parent->IsKindOf( CLASSINFO( wxTopLevelWindow ) ) && parent->IsActive() && wxGetMouseState().LeftDown();
	
	DWORD now = wxGetLocalTimeMillis().GetLo();
	if ( 0 == _time ) _time = now;
	app->Update( (scalar)( now - _time ) / 1000.0f, _orientation.Direction(), forward ? Application::ACTION_FORWARD : 0 );
	
	app->Render();
	_query = app->GetDepthQuery();
	
	if ( NULL == app->GetEgo() )
	{
		if ( !! _target )
		{
			Feature::RenderBox( _target, vec3( 1, 1, 0 ), true, ( now & 0x100 ) ? 0.03f : 0.02f );
		}
		
		Features::Selection& sel = app->GetFeatures()->GetSelection();
		
		if ( ( NULL != sel.feature ) && ( 0 == sel.type ) )
		{
			scalar inner, outer;
			vec3 position;
			sel.feature->GetActivationRadii( sel.index, inner, outer, position );
			
			glPushMatrix();
			glTranslatef( position.x, position.y, position.z );
			glDisable( GL_LIGHTING );
			glEnable( GL_BLEND );
			
			glCullFace( GL_FRONT );
			glColor4f( 1, 0, 0, 0.2f );
			Renderer::RenderCylinder( outer, 1 );
			glColor4f( 0, 1, 0, 0.2f );
			Renderer::RenderCylinder( inner, 2 );
			glCullFace( GL_BACK );
			Renderer::RenderCylinder( inner, 2 );
			glColor4f( 1, 0, 0, 0.2f );
			Renderer::RenderCylinder( outer, 1 );
			glColor3f( 1, 1, 1 );
			
			glDisable( GL_BLEND );
			glEnable( GL_LIGHTING );
			glPopMatrix();
		}
	}
	
	SwapBuffers();
}

//////////////////////////////////////////////////////////////////////////////
// GLView - mouse

void GLView::OnMouseMove( wxMouseEvent& event )
{
	Application* application = Application::GetInstance();
	if ( NULL == application ) return;
	
	wxSize size( GetClientSize() );
	wxPoint iposition( event.GetPosition() );
	wxPoint idelta( iposition - _lastPoint );
	
	vec2 position(	(scalar)( iposition.x - size.GetWidth() / 2 ) / (scalar)( size.GetWidth() / 2 ),
					(scalar)( size.GetHeight() / 2 - iposition.y ) / (scalar)( size.GetHeight() / 2 ) );
	
	vec2 delta(	(scalar)idelta.x / (scalar)( size.GetWidth() / 2 ),
				(scalar)idelta.y / (scalar)( size.GetHeight() / 2 ) );
	
	if ( _turning )
	{
		if ( ( 0 == idelta.x ) && ( 0 == idelta.y ) ) return;
		WarpPointer( _lastPoint.x, _lastPoint.y );
		
		if ( _moving )
		{
			if ( NULL == Application::GetInstance()->GetEgo() )
			{
				delta *= 20;
				_position += matrix3x3( _orientation ) * vec3( delta.x, 0, -delta.y );
			}
		}
		else
		{
			_angles.yaw += delta.x * ( M_PI / -4 );
			_angles.pitch += delta.y * ( M_PI / -4 );
			
			while ( _angles.yaw < 0 ) _angles.yaw += M_PI * 2;
			while ( _angles.yaw >= M_PI * 2 ) _angles.yaw -= M_PI * 2;
			
			_angles.pitch = max( _angles.pitch, -M_PI * 0.45f );
			_angles.pitch = min( _angles.pitch, M_PI * 0.5f );
			
			_orientation = quaternion( _angles.yaw, vec3( 0, 1, 0 ) ) * quaternion( _angles.pitch, vec3( 1, 0, 0 ) );
			_orientation.Normalize();
		}
	}
	else if ( _dragging )
	{
		Features::Selection& sel = application->GetFeatures()->GetSelection();
		
		if ( NULL != sel.feature )
		{
			sel.feature->MovePoint( sel.type, sel.index, _query );
		}
	}
	else if ( _sizing )
	{
		Features::Selection& sel = application->GetFeatures()->GetSelection();
		
		if ( NULL != sel.feature )
		{
			scalar inner, outer;
			vec3 discard;
			
			if ( sel.feature->GetActivationRadii( sel.index, inner, outer, discard ) )
			{
				scalar& value = ( event.ShiftDown() ? inner : outer );
				value -= delta.y * 2.0f;
				
				if ( value < 0.1f ) value = 0.1f;
				if ( inner + 0.1f >= outer ) outer = inner + 0.1f;
				
				sel.feature->SetActivationRadii( sel.index, inner, outer );
			}
		}
		
		_lastPoint = iposition;
	}
	
	application->SetDepthQuery( iposition.x, size.GetHeight() - iposition.y );
	
	Refresh();
}

void GLView::OnLeftDown( wxMouseEvent& event )
{
	SetFocus();
	
	Application* app = Application::GetInstance();
	if ( NULL == app ) return;
	if ( NULL != app->GetEgo() ) return;
	
	if ( _turning )
	{
		_moving = true;
		return;
	}
	else if ( _dragging || _sizing )
	{
		return;
	}
	
	wxSize size( GetClientSize() );
	wxPoint iposition( event.GetPosition() );
	
	vec2 position(	(scalar)( iposition.x - size.GetWidth() / 2 ) / (scalar)( size.GetWidth() / 2 ),
					(scalar)( size.GetHeight() / 2 - iposition.y ) / (scalar)( size.GetHeight() / 2 ) );
	
	vec3 right, up, forward;
	_orientation.Axes( &right, &up, &forward );
	
	vec3 ray( forward );
	ray += right * position.x;
	ray += up * position.y * ( (scalar)size.GetHeight() / (scalar)size.GetWidth() );
	ray.Normalize();
	
	scalar distance = 100;
	
	Features::Selection hit = app->GetFeatures()->HitTest( _position, ray, distance );
	Features::Selection& sel = app->GetFeatures()->GetSelection();
	
	if ( distance > ( _query - _position ).Length() ) hit.feature = NULL;
	
	if ( NULL != hit.feature )
	{
		if ( sel != hit )
		{
			sel = hit;
			
			wxCommandEvent update( wxEVT_COMMAND_MENU_SELECTED, ID_UPDATED );
			GetGrandParent()->AddPendingEvent( update );
		}
		else if ( NULL != sel.feature )
		{
			if ( event.ControlDown() )
			{
				if ( 0 == sel.type )
				{
					_sizing = true;
					CaptureMouse();
				}
			}
			else
			{
				_dragging = true;
				CaptureMouse();
			}
		}
	}
	
	if ( NULL == hit.feature )
	{
		_target = _query;
		_target.y += 0.05f;
	}
	
	Refresh();
}

void GLView::OnLeftUp( wxMouseEvent& event )
{
	if ( _turning )
	{
		_moving = false;
	}
	else if ( _dragging || _sizing )
	{
		_dragging = _sizing = false;
		
		if ( HasCapture() ) ReleaseMouse();
		
		Features::Selection& sel = Application::GetInstance()->GetFeatures()->GetSelection();
		if ( NULL != sel.feature ) sel.feature->Save();
		
		Refresh();
	}
}

void GLView::OnRightDown( wxMouseEvent& event )
{
	SetFocus();
	
	if ( _turning || _dragging || _sizing ) return;
	
	_turning	= true;
	_moving		= false;
	_lastPoint	= event.GetPosition();
	
	CaptureMouse();
#ifdef _WIN32
	ShowCursor( FALSE );
#endif
}

void GLView::OnRightUp( wxMouseEvent& event )
{
	if ( _turning )
	{
		_turning = false;
		
		if ( HasCapture() ) ReleaseMouse();
#ifdef _WIN32
		ShowCursor( TRUE );
#endif
	}
}

void GLView::OnDoubleClick( wxMouseEvent& event )
{
	if ( ! _turning )
	{
		wxCommandEvent command( wxEVT_COMMAND_MENU_SELECTED, ID_FEATURE_EDIT );
		OnFeatureEdit( command );
	}
}

//////////////////////////////////////////////////////////////////////////////
// GLView - mouse capture lost

void GLView::OnCaptureLost( wxMouseCaptureLostEvent& event )
{
	if ( _turning )
	{
		wxMouseEvent ev;
		OnRightUp( ev );
	}
	else if ( _dragging || _sizing )
	{
		wxMouseEvent ev;
		OnLeftUp( ev );
	}
}

//////////////////////////////////////////////////////////////////////////////
// GLView - key down

void GLView::OnKeyDown( wxKeyEvent& event )
{
	if ( Application* app = Application::GetInstance() )
	{
		switch ( event.GetKeyCode() )
		{
		case WXK_ESCAPE:
			if ( NULL != app->GetEgo() )
			{
				wxCommandEvent evt( wxEVT_COMMAND_MENU_SELECTED, ID_RUN );
				GetGrandParent()->AddPendingEvent( evt );
			}
			else if ( NULL != app->GetFeatures()->GetSelection().feature )
			{
				app->GetFeatures()->GetSelection().feature = NULL;
				
				Refresh();
				
				wxCommandEvent update( wxEVT_COMMAND_MENU_SELECTED, ID_UPDATED );
				GetGrandParent()->AddPendingEvent( update );
			}
			return;
		}
	}
	
	event.Skip();
}

//////////////////////////////////////////////////////////////////////////////
// GLView - commands

void GLView::OnUpdateOpen( wxUpdateUIEvent& event )
{
	event.Enable( ( NULL != Application::GetInstance() ) && ( NULL == Application::GetInstance()->GetEgo() ) );
}

void GLView::OnUpdateFeature( wxUpdateUIEvent& event )
{
	if ( Application* application = Application::GetInstance() )
	{
		event.Enable( ( NULL == application->GetEgo() ) && ( NULL != application->GetFeatures()->GetSelection().feature ) );
	}
	else
	{
		event.Enable( false );
	}
}

void GLView::OnFeatureNew( wxCommandEvent& event )
{
	if ( ! _target )
	{
		wxMessageBox( "First position the yellow target cursor.", "Target" );
	}
	else if ( Application* application = Application::GetInstance() )
	{
		Features* features	= application->GetFeatures();
		Feature* feature	= features->AddFeature();
		
		feature->AddPoint( 0, _target );
		feature->Save();
		_target.Zero();
		
		FeatureDialog dlg( GetGrandParent() );
		
		if ( wxID_OK != dlg.ShowModal() )
		{
			features->DeleteFeature( feature );
		}
		
		Refresh();
		
		wxCommandEvent update( wxEVT_COMMAND_MENU_SELECTED, ID_UPDATED );
		GetGrandParent()->AddPendingEvent( update );
	}
}

void GLView::OnFeatureEdit( wxCommandEvent& event )
{
	if ( Application* application = Application::GetInstance() )
	{
		if ( NULL == application->GetEgo() )
		{
			if ( NULL != application->GetFeatures()->GetSelection().feature )
			{
				FeatureDialog dlg( GetGrandParent() );
				dlg.ShowModal();
				Refresh();
				
				wxCommandEvent update( wxEVT_COMMAND_MENU_SELECTED, ID_UPDATED );
				GetGrandParent()->AddPendingEvent( update );
			}
		}
	}
}

void GLView::OnFeatureDelete( wxCommandEvent& event )
{
	if ( Application* application = Application::GetInstance() )
	{
		if ( Feature* feature = application->GetFeatures()->GetSelection().feature )
		{
			if ( wxYES == wxMessageBox(	T("Are you sure you want to delete the feature \"") + feature->GetName() + T("\"?"),
										T("Delete Feature"), wxYES_NO ) )
			{
				feature->Delete();
				application->GetFeatures()->GetSelection().feature = NULL;
				Refresh();
				
				wxCommandEvent update( wxEVT_COMMAND_MENU_SELECTED, ID_UPDATED );
				GetGrandParent()->AddPendingEvent( update );
			}
		}
	}
}

void GLView::OnPointActivation( wxCommandEvent& event )
{
	if ( ! _target )
	{
		wxMessageBox( "First position the yellow target cursor.", "Target" );
	}
	else if ( Application* application = Application::GetInstance() )
	{
		Features::Selection& sel = application->GetFeatures()->GetSelection();
		
		if ( Feature* feature = sel.feature )
		{
			sel.type	= 0;
			sel.index	= feature->AddPoint( sel.type, _target );
			
			_target.Zero();
			
			Refresh();
			
			wxCommandEvent update( wxEVT_COMMAND_MENU_SELECTED, ID_UPDATED );
			GetGrandParent()->AddPendingEvent( update );
		}
	}
}

void GLView::OnPointLocation( wxCommandEvent& event )
{
	if ( ! _target )
	{
		wxMessageBox( "First position the yellow target cursor.", "Target" );
	}
	else if ( Application* application = Application::GetInstance() )
	{
		Features::Selection& sel = application->GetFeatures()->GetSelection();
		
		if ( Feature* feature = sel.feature )
		{
			sel.type	= 1;
			sel.index	= feature->AddPoint( sel.type, _target );
			
			_target.Zero();
			
			Refresh();
			
			wxCommandEvent update( wxEVT_COMMAND_MENU_SELECTED, ID_UPDATED );
			GetGrandParent()->AddPendingEvent( update );
		}
	}
}

void GLView::OnPointGuide( wxCommandEvent& event )
{
	if ( ! _target )
	{
		wxMessageBox( "First position the yellow target cursor.", "Target" );
	}
	else if ( Application* application = Application::GetInstance() )
	{
		Features::Selection& sel = application->GetFeatures()->GetSelection();
		
		if ( Feature* feature = sel.feature )
		{
			sel.type	= 2;
			sel.index	= feature->AddPoint( sel.type, _target );
			
			_target.Zero();
			
			Refresh();
			
			wxCommandEvent update( wxEVT_COMMAND_MENU_SELECTED, ID_UPDATED );
			GetGrandParent()->AddPendingEvent( update );
		}
	}
}

void GLView::OnUpdatePointDelete( wxUpdateUIEvent& event )
{
	if ( Application* application = Application::GetInstance() )
	{
		if ( NULL == application->GetEgo() )
		{
			Features::Selection& sel = application->GetFeatures()->GetSelection();
			event.Enable( NULL != sel.feature );
			
			if ( event.GetId() == ID_POINT_DELETE_WITH_CAPTION )
			{
				switch ( sel.type )
				{
				case 0:
					event.SetText( T("Delete Activation &Point\tDelete") );
					return;
				case 1:
					event.SetText( T("Delete Location &Point\tDelete") );
					return;
				case 2:
					event.SetText( T("Delete Guide &Point\tDelete") );
					return;
				default:
					event.SetText( T("Delete Selected &Point\tDelete") );
					return;
				}
			}
			
			return;
		}
	}
	
	event.Enable( false );
	
	if ( event.GetId() == ID_POINT_DELETE_WITH_CAPTION )
	{
		event.SetText( T("Delete Selected &Point\tDelete") );
	}
}

void GLView::OnPointDelete( wxCommandEvent& event )
{
	if ( Application* application = Application::GetInstance() )
	{
		Features::Selection& sel = application->GetFeatures()->GetSelection();
		
		if ( Feature* feature = sel.feature )
		{
			if ( feature->RemovePoint( sel.type, sel.index ) )
			{
				sel.type = 0;
				sel.index = 0;
				Refresh();
			}
			else
			{
				OnFeatureDelete( event );
			}
			
			wxCommandEvent update( wxEVT_COMMAND_MENU_SELECTED, ID_UPDATED );
			GetGrandParent()->AddPendingEvent( update );
		}
	}
}
