//
// MainFrame.cpp
// Copyright Michael J Stokes, 2009.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "Editor.h"
#include "Application.h"
#include "Actors.h"
#include "MainFrame.h"
#include "XPToolBar.h"
#include "Sidebar.h"
#include "GLView.h"
#include "ConsoleCtrl.h"
#include "MeshDialog.h"
#include "BMLFrame.h"


IMPLEMENT_ABSTRACT_CLASS( MainFrame, wxFrame )

BEGIN_EVENT_TABLE( MainFrame, wxFrame )
	EVT_CLOSE( MainFrame::OnClose )
	EVT_MENU( wxID_EXIT, MainFrame::OnExit )
	EVT_MENU( wxID_NEW, MainFrame::OnNew )
	EVT_MENU( wxID_OPEN, MainFrame::OnOpen )
	EVT_UPDATE_UI( ID_RUN, MainFrame::OnUpdateRun )
	EVT_MENU( ID_RUN, MainFrame::OnRun )
	EVT_UPDATE_UI( ID_MESHES, MainFrame::OnUpdateOpen )
	EVT_MENU( ID_MESHES, MainFrame::OnMeshes )
	EVT_UPDATE_UI( ID_BML_CONSOLE, MainFrame::OnUpdateOpen )
	EVT_MENU( ID_BML_CONSOLE, MainFrame::OnBMLConsole )
	EVT_UPDATE_UI( ID_SMARTBODY, MainFrame::OnUpdateSmartbody )
	EVT_MENU( ID_SMARTBODY, MainFrame::OnSmartbody )
END_EVENT_TABLE()


//////////////////////////////////////////////////////////////////////////////
// MainFrame construction

MainFrame::MainFrame() : wxFrame( NULL, -1, T("Virtual Guide System") )
{
	_sidebar	= NULL;
	_view		= NULL;
	_console	= NULL;
	
	wxTheApp->SetTopWindow( this );
	
	_horizontal	= new wxSplitterWindow( this, wxID_ANY, wxDefaultPosition, wxSize( 0, 0 ) );
	_vertical	= new wxSplitterWindow( _horizontal, wxID_ANY, wxDefaultPosition, wxSize( 0, 0 ) );
	_sidebar	= new Sidebar( _horizontal );
	_view		= new GLView( _vertical );
	_console	= new ConsoleCtrl( _vertical );
	
	_horizontal->SetMinimumPaneSize( 32 );
	_horizontal->SetSashGravity( 1 );
	_horizontal->SplitVertically( _vertical, _sidebar, 200 );
	_vertical->SetMinimumPaneSize( 32 );
	_vertical->SetSashGravity( 1 );
	_vertical->SplitHorizontally( _view, _console, 40 );
	
	wxMenuBar* menus = new wxMenuBar( wxBORDER_NONE );
	
	wxMenu* file = new wxMenu();
	menus->Append( file, "&File" );
	file->Append( wxID_NEW, T("&New Project...") );
	file->Append( wxID_OPEN, T("&Open Project...\tCtrl+O") );
	file->AppendSeparator();
	file->Append( wxID_EXIT, T("E&xit") );
	
	wxMenu* edit = new wxMenu();
	menus->Append( edit, T("&Edit") );
	edit->Append( ID_FEATURE_NEW, T("&New Feature...\tCtrl+F") );
	edit->Append( ID_FEATURE_EDIT, T("&Edit Feature...\tSpace") );
	edit->Append( ID_FEATURE_DELETE, T("&Delete Selected Feature") );
	edit->AppendSeparator();
	edit->Append( ID_POINT_ACTIVATION, T("Add &Activation Point\tCtrl+A") );
	edit->Append( ID_POINT_GUIDE, T("Add &Guide Point\tCtrl+G") );
	edit->Append( ID_POINT_LOCATION, T("Add &Location Point\tCtrl+L") );
	edit->Append( ID_POINT_DELETE_WITH_CAPTION, T("Delete Selected &Point\tDel") );
	
	wxMenu* tools = new wxMenu();
	menus->Append( tools, "&Tools" );
	tools->Append( ID_MESHES, T("&Mesh Manager...\tCtrl+M") );
	tools->Append( ID_BML_CONSOLE, T("&BML Console...\tCtrl+B") );
	tools->AppendSeparator();
	tools->AppendCheckItem( ID_SMARTBODY, T("&Start SmartBody") );
	tools->AppendSeparator();
	tools->AppendCheckItem( ID_RUN, T("&Run Mode\tF5") );
	
	wxMenu* help = new wxMenu();
	menus->Append( help, T("&Help") );
	help->Append( wxID_ABOUT, T("&About...") );
	
	SetMenuBar( menus );
	
	wxToolBar* toolbar = XPToolBar::Create( this );
	toolbar->AddTool( wxID_OPEN, T("Open"), wxBitmap( Open_xpm ) );
	toolbar->AddSeparator();
	toolbar->AddTool( ID_FEATURE_NEW, T("New Feature"), wxBitmap( Star_xpm ) );
	toolbar->AddSeparator();
	toolbar->AddTool( ID_POINT_ACTIVATION, T("Activation Point"), wxBitmap( Activation_xpm ) );
	toolbar->AddTool( ID_POINT_LOCATION, T("Location Point"), wxBitmap( Location_xpm ) );
	toolbar->AddTool( ID_POINT_GUIDE, T("Guide Point"), wxBitmap( Guide_xpm ) );
	toolbar->AddTool( ID_POINT_DELETE, T("Delete Point"), wxArtProvider::GetBitmap( wxART_DELETE, wxART_TOOLBAR ) );
	toolbar->AddSeparator();
	toolbar->AddTool( ID_MESHES, T("Meshes"), wxBitmap( Meshes_xpm ) );
	toolbar->AddTool( ID_BML_CONSOLE, T("BML Console"), wxBitmap( Meshes_xpm ) );
	toolbar->AddCheckTool( ID_RUN, T("Run Mode"), wxBitmap( Run_xpm ) );
	toolbar->Realize();
	SetToolBar( toolbar );
	
	SetSize( wxSize( 1024, 768 ) );
	SetMinSize( wxSize( 640, 300 ) );
	CenterOnScreen();
}

//////////////////////////////////////////////////////////////////////////////
// MainFrame destruction

MainFrame::~MainFrame()
{
	if ( BMLFrame* bml = BMLFrame::GetInstance() ) delete bml;
}

//////////////////////////////////////////////////////////////////////////////
// MainFrame event rerouting

bool MainFrame::ProcessEvent( wxEvent& event )
{
	if ( ( event.IsCommandEvent() ) && ( 0 != event.GetId() ) )
	{
		int prop = event.StopPropagation();
		bool result = false;
		
		if ( NULL != _sidebar ) result = result || _sidebar->ProcessEvent( event );
		if ( NULL != _view ) result = result || _view->ProcessEvent( event );
		
		event.ResumePropagation( prop );
		
		if ( result ) return true;
	}
	
	return wxFrame::ProcessEvent( event );
}

//////////////////////////////////////////////////////////////////////////////
// MainFrame command handlers

void MainFrame::OnClose( wxCloseEvent& event )
{
	Destroy();
}

void MainFrame::OnExit( wxCommandEvent& event )
{
	Close();
}

void MainFrame::OnNew( wxCommandEvent& event )
{
	wxDirDialog dlg( this, T("New Project"), wxEmptyString );
	
	if ( wxID_OK != dlg.ShowModal() ) return;
	
	string path( dlg.GetPath() );
	size_t pos = path.find_last_of( "/\\" );
	if ( pos != string::npos ) path.resize( pos );
	
	Application::Destroy();
	Application* application = new Application();
	
	if ( application->Create( path.c_str() ) )
	{
		application->PrepareContext();
	}
	else
	{
		Application::Destroy();
	}
	
	Refresh();
	
	wxCommandEvent update( wxEVT_COMMAND_MENU_SELECTED, ID_UPDATED );
	AddPendingEvent( update );
}

void MainFrame::OnOpen( wxCommandEvent& event )
{
	wxFileDialog dlg( this, T("Open Project"), wxGetCwd() + "/Root", wxEmptyString,
		T("Projects|root.xml"), wxFD_OPEN );

	if ( wxID_OK != dlg.ShowModal() ) return;
	
	string path( dlg.GetPath() );
	size_t pos = path.find_last_of( "/\\" );
	if ( pos != string::npos ) path.resize( pos );
	
	Application::Destroy();
	Application* application = new Application();
	
	if ( application->Load( path.c_str() ) )
	{
		application->PrepareContext();
	}
	else
	{
		Application::Destroy();
	}
	
	Refresh();
	
	wxCommandEvent update( wxEVT_COMMAND_MENU_SELECTED, ID_UPDATED );
	AddPendingEvent( update );
}

void MainFrame::OnUpdateOpen( wxUpdateUIEvent& event )
{
	event.Enable( NULL != Application::GetInstance() );
}

void MainFrame::OnUpdateRun( wxUpdateUIEvent& event )
{
	event.Enable( NULL != Application::GetInstance() );
	event.Check( NULL != Application::GetInstance()->GetEgo() );
}

void MainFrame::OnRun( wxCommandEvent& event )
{
	if ( Application* app = Application::GetInstance() )
	{
		if ( NULL == app->GetEgo() )
		{
			if ( app->GetActors()->IsConnected() )
			{
				app->Start();
			}
			else if ( wxYES == wxMessageBox( "SmartBody is not running.  You must start a SmartBody process before entering Run Mode.\n\nAttempt to launch SmartBody now?", "Run Mode", wxYES_NO ) )
			{
				OnSmartbody( event );
			}
		}
		else
		{
			app->Stop();
		}
		
		Refresh();
	}
}

void MainFrame::OnMeshes( wxCommandEvent& event )
{
	if ( NULL != Application::GetInstance() )
	{
		MeshDialog dlg( this );
		dlg.ShowModal();
		Refresh();
		
		wxCommandEvent update( wxEVT_COMMAND_MENU_SELECTED, ID_UPDATED );
		AddPendingEvent( update );
	}
}

void MainFrame::OnBMLConsole( wxCommandEvent& event )
{
	if ( Application* app = Application::GetInstance() )
	{
		if ( app->GetActors()->IsConnected() )
		{
			BMLFrame* bml = BMLFrame::GetInstance();
			
			if ( NULL == bml )
			{
				bml = new BMLFrame();
			}
			
			bml->Show( true );
			if ( bml->IsIconized() ) bml->Iconize( false );
			bml->SetFocus();
		}
		else if ( wxYES == wxMessageBox( "SmartBody is not running.  You must start a SmartBody process before entering Run Mode.\n\nAttempt to launch SmartBody now?", "Run Mode", wxYES_NO ) )
		{
			OnSmartbody( event );
		}
	}
}

void MainFrame::OnUpdateSmartbody( wxUpdateUIEvent& event )
{
	if ( Application* app = Application::GetInstance() )
	{
		event.Enable( true );
		event.Check( app->GetActors()->IsConnected() );
		event.SetText( app->GetActors()->IsConnected() ? "&SmartBody Running" : "&Start SmartBody" );
	}
	else
	{
		event.Enable( false );
		event.Check( false );
		event.SetText( "&Start SmartBody" );
	}
}

void MainFrame::OnSmartbody( wxCommandEvent& event )
{
	if ( Application* app = Application::GetInstance() )
	{
		if ( ! app->GetActors()->IsConnected() )
		{
#ifdef _WIN32
			char raw[ MAX_PATH ];
			GetModuleFileName( NULL, raw, sizeof( raw ) );
			string path( raw );
			
			size_t pos = path.find_last_of( "\\" );
			
			if ( string::npos != pos )
			{
				path.resize( pos );
			}
			
			if ( ! wxFileExists( path + "\\SBM\\SmartBody.bat" ) )
			{
				path += "\\..\\..";
			}
			
			path += "\\SBM";
			ShellExecute( (HWND)GetHandle(), NULL, ( path + "\\SmartBody.bat" ).c_str(), NULL, path.c_str(), SW_SHOWNORMAL );
#endif
		}
	}
}
