//
// MeshDialog.cpp
// Copyright Michael J Stokes, 2008.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "Editor.h"
#include "Application.h"
#include "Environment.h"
#include "EnvironmentMesh.h"
#include "MeshDialog.h"

IMPLEMENT_ABSTRACT_CLASS( MeshDialog, wxDialog )

BEGIN_EVENT_TABLE( MeshDialog, wxDialog )
	EVT_LIST_ITEM_SELECTED( ID_MESHES, MeshDialog::OnSelection )
	EVT_LIST_ITEM_DESELECTED( ID_MESHES, MeshDialog::OnSelection )
	EVT_CHECKBOX( ID_MESH_FLAGS, MeshDialog::OnFlags )
	EVT_BUTTON( ID_MESH_ADD, MeshDialog::OnAdd )
	EVT_BUTTON( ID_MESH_REMOVE, MeshDialog::OnRemove )
	EVT_BUTTON( wxID_OK, MeshDialog::OnOK )
END_EVENT_TABLE()

static PCSTR FLAGS[] = { "", "clip", "render", "clip+render", "depth", "clip+depth", "render", "clip+render" };


//////////////////////////////////////////////////////////////////////////////
// MeshDialog construction

MeshDialog::MeshDialog( wxWindow* parent ) : wxDialog( parent, wxID_ANY, T("Environment Meshes"), wxDefaultPosition )
{
	_list = new wxListCtrl( this, ID_MESHES, wxDefaultPosition, wxDefaultSize, wxLC_REPORT | wxLC_SINGLE_SEL );
	
	_list->InsertColumn( 0, T("Name"), wxLIST_FORMAT_LEFT, 200 );
	_list->InsertColumn( 1, T("Flags"), wxLIST_FORMAT_RIGHT, 80 );
	_list->SetMinSize( wxSize( 300, 0 ) );
	
	wxBoxSizer* controls = new wxBoxSizer( wxVERTICAL );
	controls->Add( new wxButton( this, ID_MESH_ADD, T("Add...") ), 0, wxALL, 2 );
	controls->Add( _remove = new wxButton( this, ID_MESH_REMOVE, T("Remove") ), 0, wxALL, 2 );
	controls->Add( _clip = new wxCheckBox( this, ID_MESH_FLAGS, T("Clip") ), 0, wxALL, 2 );
	controls->Add( _render = new wxCheckBox( this, ID_MESH_FLAGS, T("Render") ), 0, wxALL, 2 );
	controls->Add( _depth = new wxCheckBox( this, ID_MESH_FLAGS, T("Depth") ), 0, wxALL, 2 );
	
	wxBoxSizer* lr = new wxBoxSizer( wxHORIZONTAL );
	lr->Add( _list, 1, wxEXPAND | wxALL, 8 );
	lr->Add( controls, 0, wxALIGN_TOP | wxALL, 8 );
	
	wxBoxSizer* root = new wxBoxSizer( wxVERTICAL );
	root->Add( lr, 1, wxEXPAND | wxALL, 8 );
	root->Add( CreateButtonSizer( wxOK | wxCANCEL ), 0, wxALIGN_CENTER | wxBOTTOM, 10 );
	SetSizerAndFit( root );
	
	Environment* environment = Application::GetInstance()->GetEnvironment();
	const list<EnvironmentMesh*>& meshes = environment->GetMeshes();
	
	for ( list<EnvironmentMesh*>::const_iterator it = meshes.begin() ; it != meshes.end() ; ++ it )
	{
		EnvironmentMesh* mesh = *it;
		wxListItem item;
		
		item.SetText( mesh->GetName() );
		item.SetId( _list->InsertItem( item ) );
		item.SetColumn( 1 );
		item.SetText( FLAGS[ mesh->GetFlags() ] );
		_list->SetItem( item );
		_list->SetItemData( item.GetId(), mesh->GetFlags() );
	}
	
	_remove->Enable( false );
	_clip->Enable( false );
	_render->Enable( false );
	_depth->Enable( false );
}

//////////////////////////////////////////////////////////////////////////////
// MeshDialog destruction

MeshDialog::~MeshDialog()
{
}

//////////////////////////////////////////////////////////////////////////////
// MeshDialog command handlers

void MeshDialog::OnSelection( wxListEvent& event )
{
	bool enable	= ( 0 != _list->GetSelectedItemCount() );
	long index	= enable ? _list->GetNextItem( -1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED ) : -1;
	
	_remove->Enable( enable );
	_clip->Enable( enable );
	_render->Enable( enable );
	_depth->Enable( enable );
	
	if ( index >= 0 )
	{
		int flags = _list->GetItemData( index );
		_clip->SetValue( 0 != ( flags & EnvironmentMesh::CLIP ) );
		_render->SetValue( 0 != ( flags & EnvironmentMesh::RENDER ) );
		_depth->SetValue( 0 != ( flags & ( EnvironmentMesh::RENDER | EnvironmentMesh::DEPTH ) ) );
		if ( _render->GetValue() ) _depth->Enable( false );
	}
}

void MeshDialog::OnFlags( wxCommandEvent& event )
{
	long index	= _list->GetNextItem( -1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED );
	if ( index < 0 ) return;
	
	int flags = 0;
	if ( _clip->GetValue() ) flags |= EnvironmentMesh::CLIP;
	if ( _render->GetValue() ) flags |= EnvironmentMesh::RENDER | EnvironmentMesh::DEPTH;
	if ( _depth->GetValue() ) flags |= EnvironmentMesh::DEPTH;
	
	_list->SetItemData( index, flags );

	wxListItem item;
	item.SetId( index );
	item.SetColumn( 1 );
	item.SetText( FLAGS[ flags ] );
	_list->SetItem( item );
	
	_clip->SetValue( 0 != ( flags & EnvironmentMesh::CLIP ) );
	_render->SetValue( 0 != ( flags & EnvironmentMesh::RENDER ) );
	_depth->SetValue( 0 != ( flags & ( EnvironmentMesh::RENDER | EnvironmentMesh::DEPTH ) ) );
	_depth->Enable( ! _render->GetValue() );
}

void MeshDialog::OnAdd( wxCommandEvent& event )
{
	Application* application = Application::GetInstance();
	
	wxFileDialog dlg( this, T("Add Mesh"),
		application->GetPath() + "/environment/meshes", wxEmptyString,
		T("Wavefront OBJ|*.obj"), wxFD_OPEN );
	
	if ( wxID_OK != dlg.ShowModal() ) return;
	
	string path( dlg.GetPath() );
	size_t pos = path.find_last_of( "/\\" );
	if ( pos == string::npos ) return;
	
	string name( path.substr( pos + 1 ) );
	path.resize( pos );
	
#ifdef _WIN32
	for ( UINT fix = 0 ; fix < path.size() ; fix ++ )
	{
		if ( path[ fix ] == '/' ) path[ fix ] = '\\';
	}
#endif
	
	if ( 0 != path.compare( application->GetPath() + PATH_SLASH_STR "environment" PATH_SLASH_STR "meshes" ) )
	{
		if ( wxYES != wxMessageBox(
			T("You will need to copy the OBJ file to the /environment/meshes directory.  Would you like to do that now?"),
			T("Import Mesh"), wxYES_NO ) ) return;
		
		path = application->GetPath() + "/environment/meshes/" + name;
		
		if ( ! wxCopyFile( dlg.GetPath(), path, false ) )
		{
			wxMessageBox( T("Unable to copy '") + dlg.GetPath() + T("' to '") + path + T("'.") );
			return;
		}
	}
	
	int flags = EnvironmentMesh::CLIP | EnvironmentMesh::RENDER;
	wxListItem item;
	
	item.SetText( name );
	item.SetId( _list->InsertItem( item ) );
	item.SetColumn( 1 );
	item.SetText( FLAGS[ flags ] );
	_list->SetItem( item );
	_list->SetItemData( item.GetId(), flags );
}

void MeshDialog::OnRemove( wxCommandEvent& event )
{
	long index = _list->GetNextItem( -1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED );
	
	if ( index >= 0 )
	{
		_list->DeleteItem( index );
		_remove->Enable( false );
		_clip->Enable( false );
		_render->Enable( false );
		_depth->Enable( false );
	}
}

void MeshDialog::OnOK( wxCommandEvent& event )
{
	Environment* environment = Application::GetInstance()->GetEnvironment();
	environment->Empty();
	
	for ( long index = 0 ; index < _list->GetItemCount() ; index ++ )
	{
		string source( _list->GetItemText( index ) );
		int flags = _list->GetItemData( index );
		environment->AddMesh( source.c_str(), flags );
	}
	
	environment->Save();
	
	EndDialog( wxID_OK );
}
