//
// Sidebar.cpp
// Copyright Michael J Stokes, 2009.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "Editor.h"
#include "Application.h"
#include "Features.h"
#include "Feature.h"
#include "Sidebar.h"


IMPLEMENT_ABSTRACT_CLASS( Sidebar, wxPanel )

BEGIN_EVENT_TABLE( Sidebar, wxPanel )
	EVT_MENU( ID_UPDATED, Sidebar::Update )
	EVT_TREE_SEL_CHANGED( 1, Sidebar::OnSelect )
END_EVENT_TABLE()



//////////////////////////////////////////////////////////////////////////////
// Sidebar - construction

Sidebar::Sidebar( wxWindow* parent ) : wxPanel( parent ), _images( 16, 16 )
{
	_sizer = new wxBoxSizer( wxVERTICAL );

	_tree = new wxTreeCtrl( this, 1, wxPoint( 0, 0 ), wxSize( 160, 400 ), wxTR_HAS_BUTTONS | wxTR_LINES_AT_ROOT | wxTR_HIDE_ROOT | wxTR_NO_LINES );
	_sizer->Add( _tree, 1, wxEXPAND | wxALL, 8 );
	
	_images.Add( wxBitmap( Star_xpm ) );
	_images.Add( wxBitmap( Activation_xpm ) );
	_images.Add( wxBitmap( Location_xpm ) );
	_images.Add( wxBitmap( Guide_xpm ) );
	_tree->SetImageList( &_images );
	
	_sizer->Add( new wxButton( this, ID_FEATURE_NEW, "New Feature" ), 0, wxEXPAND | wxLEFT | wxRIGHT, 8 );
	_sizer->AddSpacer( 4 );
	_sizer->Add( new wxButton( this, ID_FEATURE_EDIT, "Edit Feature" ), 0, wxEXPAND | wxLEFT | wxRIGHT, 8 );
	_sizer->AddSpacer( 4 );
	_sizer->Add( new wxButton( this, ID_POINT_ACTIVATION, "New Activation Point" ), 0, wxEXPAND | wxLEFT | wxRIGHT, 8 );
	_sizer->AddSpacer( 4 );
	_sizer->Add( new wxButton( this, ID_POINT_LOCATION, "New Location Point" ), 0, wxEXPAND | wxLEFT | wxRIGHT, 8 );
	_sizer->AddSpacer( 4 );
	_sizer->Add( new wxButton( this, ID_POINT_GUIDE, "New Guide Point" ), 0, wxEXPAND | wxLEFT | wxRIGHT, 8 );
	_sizer->AddSpacer( 4 );
	_sizer->Add( new wxButton( this, ID_POINT_DELETE, "Delete Point" ), 0, wxEXPAND | wxLEFT | wxRIGHT | wxBOTTOM, 8 );
	
	SetSizer( _sizer );
}

//////////////////////////////////////////////////////////////////////////////
// Sidebar - destruction

Sidebar::~Sidebar()
{
}

//////////////////////////////////////////////////////////////////////////////
// Sidebar - update

void Sidebar::Update( wxCommandEvent& event )
{
	event.Skip();
	
	Application* app = Application::GetInstance();
	
	if ( NULL == app )
	{
		_tree->DeleteAllItems();
		return;
	}
	
	const list<Feature*>& features = app->GetFeatures()->GetFeatures();
	
	wxTreeItemId root = _tree->GetRootItem();
	if ( ! root.IsOk() ) root = _tree->AddRoot( T("Features") );
	
	wxTreeItemIdValue cookie;
	
	for ( wxTreeItemId search = _tree->GetFirstChild( root, cookie ) ; search.IsOk() ; search = _tree->GetNextChild( root, cookie ) )
	{
		bool found = false;
		
		if ( TID* tid = (TID*)_tree->GetItemData( search ) )
		{
			for ( list<Feature*>::const_iterator it = features.begin() ; it != features.end() ; ++ it )
			{
				if ( tid->feature == *it )
				{
					found = true;
					break;
				}
			}
		}
		
		if ( ! found ) _tree->Delete( search );
	}
	
	for ( list<Feature*>::const_iterator it = features.begin() ; it != features.end() ; ++ it )
	{
		Feature* feature = *it;
		wxTreeItemId search;
		bool found = false;
		
		for ( search = _tree->GetFirstChild( root, cookie ) ; search.IsOk() ; search = _tree->GetNextChild( root, cookie ) )
		{
			if ( TID* tid = (TID*)_tree->GetItemData( search ) )
			{
				if ( tid->feature == feature )
				{
					found = true;
					break;
				}
			}
		}
		
		if ( ! found )
		{
			search = _tree->AppendItem( root, feature->GetName(), 0, 0, new TID( feature ) );
		}
		else if ( ! _tree->GetItemText( search ).IsSameAs( wxString( feature->GetName() ) ) )
		{
			_tree->SetItemText( search, feature->GetName() );
		}
		
		if ( app->GetFeatures()->GetSelection().feature == feature )
		{
			UpdateNode( search, feature, true );
			_tree->Expand( search );
		}
		else
		{
			UpdateNode( search, feature, false );
		}
		
		_tree->SortChildren( search );
	}
	
	_tree->SortChildren( root );
}

//////////////////////////////////////////////////////////////////////////////
// Sidebar - update a feature node

void Sidebar::UpdateNode( wxTreeItemId parent, Feature* feature, bool selected )
{
	wxTreeItemIdValue cookie;
	
	for ( wxTreeItemId search = _tree->GetFirstChild( parent, cookie ) ; search.IsOk() ; search = _tree->GetNextChild( parent, cookie ) )
	{
		bool found = false;
		
		if ( TID* tid = (TID*)_tree->GetItemData( search ) )
		{
			if ( tid->feature == feature )
			{
				if ( feature->CheckPoint( tid->type, tid->index ) )
				{
					found = true;
				}
			}
		}
		
		if ( ! found ) _tree->Delete( search );
	}
	
	Features::Selection& sel = Application::GetInstance()->GetFeatures()->GetSelection();
	
	for ( UINT i = 0 ; i < feature->GetActivations().size() ; i ++ )
	{
		UpdateChild( parent, feature, 0, i, feature->GetName() + wxString::Format( "_a%i", i + 1 ),
			selected && ( sel.type == 0 ) && ( sel.index == i ) );
	}
	
	for ( UINT i = 0 ; i < feature->GetLocations().size() ; i ++ )
	{
		UpdateChild( parent, feature, 1, i, feature->GetName() + wxString::Format( "_l%i", i + 1 ),
			selected && ( sel.type == 1 ) && ( sel.index == i ) );
	}
	
	for ( UINT i = 0 ; i < feature->GetGuides().size() ; i ++ )
	{
		UpdateChild( parent, feature, 2, i, feature->GetName() + wxString::Format( "_g%i", i + 1 ),
			selected && ( sel.type == 2 ) && ( sel.index == i ) );
	}
}

//////////////////////////////////////////////////////////////////////////////
// Sidebar - update point nodes

void Sidebar::UpdateChild( const wxTreeItemId& parent, Feature* feature, int type, int index, const wxString& name, bool selected )
{
	wxTreeItemIdValue cookie;
	wxTreeItemId search;
	bool found = false;
	
	for ( search = _tree->GetFirstChild( parent, cookie ) ; search.IsOk() ; search = _tree->GetNextChild( parent, cookie ) )
	{
		if ( TID* tid = (TID*)_tree->GetItemData( search ) )
		{
			if ( ( tid->feature == feature ) && ( tid->type == type ) && ( tid->index == index ) )
			{
				found = true;
				break;
			}
		}
	}
	
	if ( ! found )
	{
		search = _tree->AppendItem( parent, name, 1 + type, 1 + type, new TID( feature, type, index ) );
	}
	else if ( ! _tree->GetItemText( search ).IsSameAs( name ) )
	{
		_tree->SetItemText( search, name );
	}
	
	if ( selected )
	{
		_tree->SelectItem( search );
	}
}

//////////////////////////////////////////////////////////////////////////////
// Sidebar - selection changed

void Sidebar::OnSelect( wxTreeEvent& event )
{
	Application* app = Application::GetInstance();
	if ( NULL == app ) return;
	
	wxTreeItemId id = _tree->GetSelection();
	if ( ! id.IsOk() ) return;
	
	TID* tid = (TID*)_tree->GetItemData( id );
	if ( NULL == tid ) return;
	
	if ( tid->type < 0 )
	{
		wxTreeItemIdValue cookie;
		
		for ( wxTreeItemId search = _tree->GetFirstChild( id, cookie ) ; search.IsOk() ; search = _tree->GetNextChild( id, cookie ) )
		{
			if ( TID* tid2 = (TID*)_tree->GetItemData( search ) )
			{
				if ( ( tid2->feature == tid->feature ) && ( tid2->type == 0 ) && ( tid2->index == 0 ) )
				{
					_tree->Expand( id );
					_tree->SelectItem( search );
					id = search;
					tid = tid2;
					break;
				}
			}
		}
		
		if ( tid->type < 0 ) return;
	}
	
	Features::Selection& sel = app->GetFeatures()->GetSelection();
	
	sel.feature	= tid->feature;
	sel.type	= tid->type;
	sel.index	= tid->index;
	
	GetGrandParent()->Refresh();
}
