//
// XPToolBar.cpp
// Copyright Michael J Stokes, 2009.
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Common.h"
#include "XPToolBar.h"

IMPLEMENT_ABSTRACT_CLASS( XPToolBar, wxControl )

BEGIN_EVENT_TABLE( XPToolBar, wxToolBarBase )
	EVT_ERASE_BACKGROUND( XPToolBar::OnEraseBackground )
	EVT_PAINT( XPToolBar::OnPaint )
	EVT_MOTION( XPToolBar::OnMouseMove )
	EVT_LEAVE_WINDOW( XPToolBar::OnMouseLeave )
	EVT_LEFT_DOWN( XPToolBar::OnMouseDown )
	EVT_LEFT_UP( XPToolBar::OnMouseUp )
	EVT_MOUSE_CAPTURE_LOST( XPToolBar::OnLostCapture )
END_EVENT_TABLE()


//////////////////////////////////////////////////////////////////////////////
// XPToolBar select and create the best toolbar for the platform

wxToolBar* XPToolBar::Create( wxWindow* parent )
{
#ifdef _WIN32
	return (wxToolBar*)new XPToolBar( parent );
#else
	return new wxToolBar( parent, wxID_ANY, wxDefaultPosition, wxDefaultSize, wxTB_FLAT | wxTB_HORZ_TEXT );
#endif
}

//////////////////////////////////////////////////////////////////////////////
// XPToolBar default construction

XPToolBar::XPToolBar()
{
	Init();
}

//////////////////////////////////////////////////////////////////////////////
// XPToolBar create construction

XPToolBar::XPToolBar( wxWindow *parent, wxWindowID id, const wxPoint& pos, const wxSize& size, long style, const wxString& name )
{
	Init();
	Create( parent, id, pos, size, style, name );
}

//////////////////////////////////////////////////////////////////////////////
// XPToolBar destruction

XPToolBar::~XPToolBar()
{
	if ( wxFrame *frame = wxDynamicCast(GetParent(), wxFrame) )
	{
		if ( ! frame->IsBeingDeleted() ) frame->SendSizeEvent();
	}
}

//////////////////////////////////////////////////////////////////////////////
// XPToolBar init

void XPToolBar::Init()
{
	m_xMargin			= 2;
	m_yMargin			= 1;
	m_toolSeparation	= 1;
	m_toolPacking		= 1;
	m_defaultWidth		= 16;
	m_defaultHeight		= 16;
	
	_backColour		= wxSystemSettings::GetColour( wxSYS_COLOUR_BTNFACE );
	_backPen		= wxThePenList->FindOrCreatePen( _backColour, 1, wxSOLID );
	_backBrush		= wxTheBrushList->FindOrCreateBrush( _backColour, wxSOLID );
	_sepPen			= wxThePenList->FindOrCreatePen( wxSystemSettings::GetColour( wxSYS_COLOUR_BTNSHADOW ), 1, wxSOLID );
	_selColour		= wxColour( 193, 210, 238 );
	_selPen			= wxThePenList->FindOrCreatePen( wxColour( 49, 106, 197 ), 1, wxSOLID );
	_selBrush		= wxTheBrushList->FindOrCreateBrush( _selColour, wxSOLID );
	_shadowColour	= wxColour( 144, 157, 178 );
	_shadowBrush	= wxTheBrushList->FindOrCreateBrush( _shadowColour, wxSOLID );
	_disabledColour	= wxSystemSettings::GetColour( wxSYS_COLOUR_BTNSHADOW );
	
	_hover			= NULL;
	_down			= NULL;
}

//////////////////////////////////////////////////////////////////////////////
// XPToolBar create

bool XPToolBar::Create( wxWindow *parent, wxWindowID id, const wxPoint& pos, const wxSize& size, long style, const wxString& name )
{
	FixupStyle();
	style &= ~wxBORDER_MASK;
	style |= wxBORDER_NONE;
	
    if ( ! wxWindow::Create( parent, id, pos, size, style, name ) ) return false;
	
	SetBackgroundStyle( wxBG_STYLE_CUSTOM );
	SetFont( wxSystemSettings::GetFont( wxSYS_DEFAULT_GUI_FONT ) );
	
	return true;
}

//////////////////////////////////////////////////////////////////////////////
// XPToolBar realize the toolbar

bool XPToolBar::Realize()
{
	SetSize( DoGetBestSize() );
	return true;
}

//////////////////////////////////////////////////////////////////////////////
// XPToolBar find the tool at the specified point

wxToolBarToolBase* XPToolBar::FindToolForPosition( wxCoord x, wxCoord y ) const
{
	if ( y < 0 ) return NULL;
	if ( y > m_yMargin * 2 + m_defaultHeight + 8 ) return NULL;
	
	wxPoint pt( m_xMargin, m_yMargin );
	wxWindowDC dc( (wxWindow*)this );
	
	dc.SetFont( GetFont() );
	
	for ( wxToolBarToolsList::compatibility_iterator node = m_tools.GetFirst() ; node ; node = node->GetNext() )
	{
		wxToolBarToolBase* tool = node->GetData();
		
		if ( tool->IsSeparator() )
		{
			pt.x += 4;
		}
		else
		{
			int width = m_defaultWidth + 7 + m_toolSeparation;
			
			if ( ! tool->GetLabel().IsEmpty() )
			{
				width += dc.GetTextExtent( tool->GetLabel() ).x + 3;
			}
			
			if ( ( x >= pt.x ) && ( x < pt.x + width ) ) return tool;
			
			pt.x += width;
		}
	}
	
	return NULL;
}

//////////////////////////////////////////////////////////////////////////////
// XPToolBar tool creation

wxToolBarToolBase* XPToolBar::CreateTool( int id, const wxString& label, const wxBitmap& bmpNormal, const wxBitmap& bmpDisabled, wxItemKind kind, wxObject* clientData, const wxString& shortHelp, const wxString& longHelp )
{
	if ( bmpNormal.IsOk() && ! bmpDisabled.IsOk() )
	{
		return new wxToolBarToolBase( this, id, label, bmpNormal, wxBitmap( bmpNormal.ConvertToImage().ConvertToGreyscale() ), kind, clientData, shortHelp, longHelp );
	}
	else
	{
		return new wxToolBarToolBase( this, id, label, bmpNormal, bmpDisabled, kind, clientData, shortHelp, longHelp );
	}
}

wxToolBarToolBase* XPToolBar::CreateTool( wxControl* control )
{
    return new wxToolBarToolBase( this, control );
}

//////////////////////////////////////////////////////////////////////////////
// XPToolBar tool insersion

bool XPToolBar::DoInsertTool( size_t pos, wxToolBarToolBase* tool )
{
	tool->Attach( this );
	InvalidateBestSize();
	return true;
}

//////////////////////////////////////////////////////////////////////////////
// XPToolBar tool deletion

bool XPToolBar::DoDeleteTool( size_t pos, wxToolBarToolBase* tool )
{
	if ( _hover == tool )
	{
		_hover = NULL;
		Refresh();
	}
	
	if ( _down == tool )
	{
		_down = NULL;
		Refresh();
		ReleaseMouse();
	}
	
	tool->Detach();
	InvalidateBestSize();
	return true;
}

//////////////////////////////////////////////////////////////////////////////
// XPToolBar tool state updates

void XPToolBar::DoEnableTool( wxToolBarToolBase* tool, bool enable )
{
	Refresh();
}

void XPToolBar::DoToggleTool( wxToolBarToolBase* tool, bool toggle )
{
    Refresh();
}

void XPToolBar::DoSetToggle( wxToolBarToolBase* tool, bool toggle )
{
	Refresh();
}

//////////////////////////////////////////////////////////////////////////////
// XPToolBar calculate the best size

wxSize XPToolBar::DoGetBestSize() const
{
	wxSize size( m_xMargin * 2, m_yMargin * 2 + m_defaultHeight + 7 );
	wxWindowDC dc( (wxWindow*)this );
	
	dc.SetFont( GetFont() );
	
	for ( wxToolBarToolsList::compatibility_iterator node = m_tools.GetFirst() ; node ; node = node->GetNext() )
	{
		wxToolBarToolBase* tool = node->GetData();
		
		if ( tool->IsSeparator() )
		{
			size.x += 4;
		}
		else
		{
			size.x += m_defaultWidth + 7 + m_toolSeparation;
			
			if ( ! tool->GetLabel().IsEmpty() )
			{
				size.x += dc.GetTextExtent( tool->GetLabel() ).x + 3;
			}
		}
	}
	
	CacheBestSize( size );
	return size;
}

//////////////////////////////////////////////////////////////////////////////
// XPToolBar erase the background

void XPToolBar::OnEraseBackground( wxEraseEvent& event )
{
}

//////////////////////////////////////////////////////////////////////////////
// XPToolBar paint the control

void XPToolBar::OnPaint( wxPaintEvent& event )
{
	wxAutoBufferedPaintDC dc( this );
	
	int width, height;
	GetClientSize( &width, &height );
	
	dc.SetPen( *_backPen );
	dc.SetBrush( *_backBrush );
	dc.DrawRectangle( 0, 0, width, height );
	dc.SetBackgroundMode( wxTRANSPARENT );
	dc.SetFont( GetFont() );
	
	wxPoint pt( m_xMargin, m_yMargin );
	int iof = ( m_defaultHeight + 7 ) / 2 - m_defaultHeight / 2;
	int tof = ( m_defaultHeight + 7 ) / 2 - dc.GetTextExtent( T("Xy") ).y / 2;
	
	for ( wxToolBarToolsList::compatibility_iterator node = m_tools.GetFirst() ; node ; node = node->GetNext() )
	{
		wxToolBarToolBase* tool = node->GetData();
		
		if ( tool->IsSeparator() )
		{
			dc.SetPen( *_sepPen );
			dc.DrawLine( pt.x + 1, pt.y, pt.x + 1, pt.y + m_defaultHeight + 7 );
			pt.x += 4;
			continue;
		}
		
		int wid = m_defaultWidth + 7;
		
		if ( ! tool->GetLabel().IsEmpty() )
		{
			wid += dc.GetTextExtent( tool->GetLabel() ).x + 3;
		}
		
		const wxBitmap& image = tool->GetNormalBitmap();
		
		if ( ! tool->IsEnabled() )
		{
			dc.SetTextForeground( _disabledColour );
			dc.SetTextBackground( _backColour );
			dc.DrawBitmap( tool->GetDisabledBitmap(), pt.x + 3, pt.y + iof, true );
		}
		else if ( ( tool == _down ) && ( tool == _hover ) )
		{
			dc.SetPen( *_selPen );
			dc.SetBrush( *_shadowBrush );
			dc.DrawRectangle( pt.x, pt.y, wid, m_defaultHeight + 7 );
			dc.DrawBitmap( image, pt.x + 3, pt.y + iof, true );
		}
		else if ( tool->IsToggled() )
		{
			bool sel = ( ( tool == _down ) || ( ( tool == _hover ) && ( NULL == _down ) ) );
			dc.SetPen( sel ? *_selPen : *_sepPen );
			dc.SetBrush( sel ? *_selBrush : *wxWHITE_BRUSH );
			dc.DrawRectangle( pt.x, pt.y, wid, m_defaultHeight + 7 );
			dc.DrawBitmap( image, pt.x + 3, pt.y + iof, true );
		}
		else if ( ( tool == _down ) || ( ( tool == _hover ) && ( NULL == _down ) ) )
		{
			wxBitmap mask( m_defaultWidth, m_defaultHeight, 1 );
			wxMemoryDC dcm( mask );
			
			dcm.SetPen( *wxWHITE_PEN );
			dcm.SetBrush( *wxWHITE_BRUSH );
			dcm.DrawRectangle( 0, 0, m_defaultWidth, m_defaultHeight );
			dcm.DrawBitmap( image, 0, 0, true );
			dcm.SelectObject( wxNullBitmap );
			
			dc.SetPen( *_selPen );
			dc.SetBrush( *_selBrush );
			dc.DrawRectangle( pt.x, pt.y, wid, m_defaultHeight + 7 );
			dc.SetTextForeground( _shadowColour );
			dc.SetTextBackground( _selColour );
			dc.DrawBitmap( mask, pt.x + 4, pt.y + iof + 1, false );
			dc.DrawBitmap( image, pt.x + 2, pt.y + iof - 1, true );
		}
		else
		{
			dc.DrawBitmap( image, pt.x + 3, pt.y + iof, true );
		}
		
		if ( ! tool->GetLabel().IsEmpty() )
		{
			if ( tool->IsEnabled() ) dc.SetTextForeground( *wxBLACK );
			dc.DrawText( tool->GetLabel(), pt.x + 6 + m_defaultWidth, pt.y + tof );
		}
		
		pt.x += wid + m_toolSeparation;
	}
}

//////////////////////////////////////////////////////////////////////////////
// XPToolBar mouse move event

void XPToolBar::OnMouseMove( wxMouseEvent& event )
{
	wxToolBarToolBase* tool = FindToolForPosition( event.GetX(), event.GetY() );
	
	if ( _hover != tool )
	{
		_hover = tool;
		Refresh();
	}
}

//////////////////////////////////////////////////////////////////////////////
// XPToolBar mouse leave event

void XPToolBar::OnMouseLeave( wxMouseEvent& event )
{
	if ( NULL != _hover )
	{
		_hover = NULL;
		Refresh();
	}
}

//////////////////////////////////////////////////////////////////////////////
// XPToolBar mouse down event

void XPToolBar::OnMouseDown( wxMouseEvent& event )
{
	if ( NULL != _down )
	{
		ReleaseMouse();
	}
	
	if ( _down = FindToolForPosition( event.GetX(), event.GetY() ) )
	{
		if ( _down->IsEnabled() )
		{
			Refresh();
			CaptureMouse();
		}
		else
		{
			_down = NULL;
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
// XPToolBar mouse down up

void XPToolBar::OnMouseUp( wxMouseEvent& event )
{
	if ( NULL != _down )
	{
		if ( FindToolForPosition( event.GetX(), event.GetY() ) == _down )
		{
			wxCommandEvent event( wxEVT_COMMAND_MENU_SELECTED, _down->GetId() );
			GetEventHandler()->AddPendingEvent( event );
		}
		
		_down = NULL;
		if ( HasCapture() ) ReleaseMouse();
		Refresh();
	}
}

//////////////////////////////////////////////////////////////////////////////
// XPToolBar mouse capture lost up

void XPToolBar::OnLostCapture( wxMouseCaptureLostEvent& event )
{
	if ( NULL != _down )
	{
		_down = NULL;
		Refresh();
	}
}
